/*****************************************************************************
 * NetGameDLL : BattleShips                                                  *
 *                                                                           *
 * Inlcude file for main code                                                *
 *****************************************************************************/
#ifndef BSHIPS_H
#define BSHIPS_H

#include "Images.h"              // Image handling
#include "Sprites.h"             // Sprite library
#include "Fonts.h"               // Font library, using the Sprite library
#include "FrameRegulation.h"     // Frame regulation library
#include "SoundFX.h"             // Sound effects library
#include "Config.h"              // Setup window processing and config save/load

#ifdef DEBUG_BUILD
//#define DEBUG_BUILD_SFXLOADMESSAGES
#endif

// Define tButton type:

#define BUTTONSTATE_DISABLED 0
#define BUTTONSTATE_ACTIVE 1
#define BUTTONSTATE_MOUSEOVER 2
#define BUTTONSTATE_PRESSED 3
typedef struct _tButton
{
  int iState;
  pSprite sprDisabled;
  pSprite sprActive;
  pSprite sprMouseOver;
  pSprite sprPressed;
  int XPos, YPos;
} tButton, *pButton;

// Define tTable type:

typedef struct _tTableDef
{
  long TableTopLeftCornerX, TableTopLeftCornerY;
  long TableXOneStepX, TableXOneStepY;
  long TableYOneStepX, TableYOneStepY;
  long TableOrientation;
} tTableDef, *pTableDef;

// We will play BattleShips, with 5 ships to sink:
#define NUM_OF_SHIPS 5

// Now the main thing: defining the Skin:
typedef struct _tSkin
{
  pImage ImageCache;       // The ImageCache is used to speed up loading.
                           // When a PCX image is to be used, it is expanded
                           // into this, and it remains there, so if it's
                           // the same PCX file next time, it won't be
                           // loaded/expanded again.
                           // This is used only at Skin loading, 
                           // otherwise it should be NULL!

  pImage BackgroundImage;  // Background image while positioning the ships
  tTableDef Table;         // The dimensions of table where to put the ships

  pImage BackgroundImage2; // Background image when playing (with 2 tables)
  tTableDef Table1;        // Definitions of the two tables
  tTableDef Table2;

  pSprite PositionCursor;  // The default table-cursor
  pSprite HitMarker;       // Small marker to mark hit shot
  pSprite HitBlowUp;       // Small animation to show us that we hit something
  pSprite MissMarker;      // Small marker to mark missed shot
  pSprite MissBlowUp;      // Small animation to show us that we missed the enemy

  pSprite Ship_Down[NUM_OF_SHIPS][6];   // Sprites for ships (from 0 to 5 hits, per ship!)
  pSprite Ship_Right[NUM_OF_SHIPS][6];

  long Ship_StartPositionX[NUM_OF_SHIPS]; // and their base positions
  long Ship_StartPositionY[NUM_OF_SHIPS];

  int Ship_StartPosition_DrawOrder[NUM_OF_SHIPS]; // Order of drawing of ships at start position. e.g. 0,1,2,3,4

  pFont Font1;             // Font 1 is used for Messageboard messages
  pFont Font2;             // Font 2 is used for displaying time and round...

  pButton QuitButton;      // The quit button
  pButton StartButton;     // The start button
  pButton SetupButton;     // The game setup button
  pButton VolumeButton;    // The button to bring up the volume setting dialog

  // Message Board position:
  int TextAreaX1, TextAreaY1, TextAreaX2, TextAreaY2;
  // Round info area:
  int RoundInfoAreaX1, RoundInfoAreaY1, RoundInfoAreaX2, RoundInfoAreaY2;
  // Time info area:
  int TimeInfoAreaX1, TimeInfoAreaY1, TimeInfoAreaX2, TimeInfoAreaY2;
  // Score info area:
  int ScoreInfoAreaX1, ScoreInfoAreaY1, ScoreInfoAreaX2, ScoreInfoAreaY2;

  // Message Board messages:
  char *msgWelcome;
  char *msgEnemyDisconnected;
  char *msgWaitingForEnemy;
  char *msgEnemyPositioned;
  char *msgEnemyPositionedTooSrv;
  char *msgEnemyPositionedTooCli;
  char *msgWePositionedTooSrv;
  char *msgWePositionedTooCli;
  char *msgEnemyShot;
  char *msgWeHit;
  char *msgWeMissed;
  char *msgGameWon;
  char *msgGameLost;
  char *msgConfigChanged;

  // Sound effects:
  PSFX_FX BackgroundMusic;  // Background music
  PSFX_FX WeHitSFX;         // Sound played when our ship is to be hit
  PSFX_FX WeMissSFX;        // Sound played when our ship is to be missed
  PSFX_FX EnemyHitSFX;      // Sound played when an enemy ship is to be hit
  PSFX_FX EnemyMissSFX;     // Sound played when the enemy's ship is to be missed
  PSFX_FX OverButtonSFX;    // Sound played when the cursor goes over a button
  PSFX_FX LeaveButtonSFX;   // Sound played when the cursor leaves a button
  PSFX_FX PressButtonSFX;   // Sound played when a button is pressed
  PSFX_FX GameStartSFX;     // Sound played when both players pressed the start button
  PSFX_FX GameWonSFX;       // Sound played when this player won the game
  PSFX_FX GameLostSFX;      // Sound played when this player lost the game

  // Animation wait times, for synchronization with sound effects:
  long WeHitAnimationWaitTime;
  long WeMissAnimationWaitTime;
  long EnemyHitAnimationWaitTime;
  long EnemyMissAnimationWaitTime;

  // Things describing the volume setting panel:
  pSprite VolumePanel;     // The volume setting panel background
  int VolumePanelX, VolumePanelY;

  pButton MainVolumeArm;   // The Slider arms on the volume setting panel
  pButton MusicVolumeArm;
  pButton SFXVolumeArm;
  pButton VolumePanelCloseBtn;

  // Description of the places, where these arms can be moved:
  int MainVolumeLineX, MainVolumeLineY, MainVolumeLineLength;
  int MusicVolumeLineX, MusicVolumeLineY, MusicVolumeLineLength;
  int SFXVolumeLineX, SFXVolumeLineY, SFXVolumeLineLength;

} tSkin, *pSkin;

#endif
