/*****************************************************************************
 * NetGameDLL : BattleShips                                                  *
 *                                                                           *
 * Main code                                                                 *
 *****************************************************************************/

#define INCL_TYPES
#define INCL_WIN
#define INCL_DOS
#define INCL_DOSERRORS
#define INCL_GPI
#include <os2.h>
#define INCL_MMIOOS2
#include <os2me.h>
#include <dive.h>
#include <fourcc.h>
#include <stdio.h>   // FILE usage
#include <string.h>

// Include GameDLL definitions
#define HAVE_SETUPGAME
#include "GameDLLdef.h"

// Include BShips structures, headers etc...
#include "BShips.h"

// Include resource definitions
#include "BSRes.h"

//---- Game Information ----

tGameInfoBuf GameInfo = { 0, 1, 0,                   // Game version: v0.1.0
                          1, 0,                      // NetComm version: v1.0
                          "BattleShips",             // Game name
			  "Find and sink all the ships of your opponent\n"
			  "on a 10x10 table!\n",     // Game description
                          GT_DUEL                    // Game type
                        };

//---- Helper functions from NetGame Client ----

fnSendMessage    *pSendMessage;
fnReceiveMessage *pReceiveMessage;

//---- Global variables -----------

BOOL  fWeAreTheServer  = FALSE;   // TRUE if we invited the other user, FALSE if we've been invited
BOOL  fFirstTimeRender = TRUE;    // TRUE, if it's the first time to build the output image (we make it in a different way, then next times)

ULONG Socket = 0;                 // TCP/IP socket, that can be used to communicate with the other side
CHAR *RemoteUserName = NULL;      // Name of opponent
CHAR *LocalUserName = NULL;       // Our name
HMODULE DLLHandle = NULLHANDLE;   // Handle of this instance of this DLL

ULONG srcBufferWidth = 0;         // Parameters of the source image that we want to
ULONG srcBufferHeight = 0;        //   show in the DIVE window: dimensions and
ULONG srcBufferFOURCC = 0;        //   pixel format
HMTX  hmtxUseDive = 0;            // Mutex semaphore to protect DIVE (only 1 thread will call into DIVE at a time)
HDIVE hDiveInst = 0;              // Handle for opened DIVE instance
ULONG DiveImageHandle = 0;        // Handle of Dive Image Buffer
BOOL  fVRNDisabled = TRUE;        // TRUE, if blitting is disabled for some reason (window is being moved, etc...)
ULONG ActFPS = 0;                 // Last measured FPS value
ULONG ulNumBlitted = 0;           // We count, how many images we can blit in a second in this variable
HEV   hevBlitNeededEvent = 0;     // Posted when the blitter should blit one image

HWND  hFrame = 0;                 // Window handle of frame
HWND  hClient = 0;                // Window handle of window client area
char *WindowClassName="BattleShipsDiveWindowClass";
char *AppTitle="BattleShips";
HPOINTER WindowIcon;              // Icon of window

BOOL  fShuttingDown = FALSE;      // Set to TRUE if the game is closing, so all the extra threads should exit and do cleanups
TID   FrameRendererThread = -1;   // Thread ID of frame renderer thread
TID   NetworkReceiverThread = -1; // Thread ID of the Network Receiver thread
TID   SetupThread = -1;           // Thread ID of setup window processing thread
char *Skin_BaseDir = "BShips.dat";// Base directory that contains BattleShips specific data (skins, skin datas, config file, etc...)
pSkin Skin = NULL;                // Current skin
HMTX  hmtxUseSkin = 0;            // Mutex semaphore to make sure that noone will use the skin and SFX engine while it is being changed!

char *ActualMessage = NULL;       // MessageBoard stuff: The actual message to show, or the next message to show
int   MessageBoard_NewMessage = 0;//                     Set to 1 if there is a new ActualMessage to scroll in
char *VisibleMessage = NULL;      //                     The message that is currently visible
int   MessagePos = 0;             //                     Y-position of visible message

BOOL  bVolumePanelActive = FALSE; // Flag
int   iVolumePanelState;          // Describing what is the actual state of volume panel
#define VOLUMEPANELSTATE_NORMAL     0
#define VOLUMEPANELSTATE_DRAGMAIN   1
#define VOLUMEPANELSTATE_DRAGMUSIC  2
#define VOLUMEPANELSTATE_DRAGSFX    3
#define VOLUMEPANELSTATE_PUSHCLOSE  4

//---- Defines and global variables, describing the actual state of the game:

//   When we're still positioning the ships:
#define GAMESTATE_POSITION_SELF 0
//   When we're still positioning the ships, but the enemy has finished its positioning:
#define GAMESTATE_POSITION_SELF_ENEMY_POSITIONED 1
//   When we finished positioning, and only waiting for the enemy to do so:
#define GAMESTATE_WAIT_FOR_POSITION_ENEMY 2
//   We're playing already, and it's our turn now:
#define GAMESTATE_SELF_ROUND 3
//   We're playing already, and it's the turn of the enemy now:
#define GAMESTATE_ENEMY_ROUND 4
//   All ships of the enemy has been destroyed, normal end of game:
#define GAMESTATE_END_OF_GAME_WON 5
//   All of our ships has been destroyed, normal end of game:
#define GAMESTATE_END_OF_GAME_LOST 6
//   The enemy has lost connection, or simply disconnected. Unexpected end of game:
#define GAMESTATE_PEER_DISCONNECTED 7

int iGameState = GAMESTATE_POSITION_SELF;  // Start with this gamestate

int iDragShip  = -1;  // Stores the ID of the ship currently being dragged, or -1.
int iDragShipDir = 0; // If there is a ship being dragged, it tells if it's dragged as a vertical (down-ship) (0) or horizontal (right-ship) (1)

int ShipPosition[NUM_OF_SHIPS][2]; // Position of the ships on the table, in table coordinates
int ShipDirection[NUM_OF_SHIPS];   // Orientation of ships on the table, 0=down (vertical), 1=right (horizontal) ship
int ShipVisibility[NUM_OF_SHIPS];  // Transparency of the ships (0..255)
int ShipNumOfHits[NUM_OF_SHIPS];   // Number of hits per ship (max Ship_Length[]!)

int Ship_Length[NUM_OF_SHIPS] = {2,3,3,4,5}; // Length of ships in units

// When a ship is hit, the table entry goes on the following change:
// Empty -> 
// Audio only (for some time, defined in skin) -> 
// Audio and Blowup Animation of ship too (until end of animation) -> 
// Final hit marker
#define TABLEENTRY_EMPTY               0
#define TABLEENTRY_HIT_AUDIOONLY       1
#define TABLEENTRY_HIT_AUDIOANDBLOWUP  2
#define TABLEENTRY_HIT                 3
#define TABLEENTRY_MISS_AUDIOONLY      4
#define TABLEENTRY_MISS_AUDIOANDBLOWUP 5
#define TABLEENTRY_MISS                6

// So the tables:
unsigned char LocalTable[10][10]; // Containing the shots of enemy
unsigned char EnemyTable[10][10]; // Containing our shots

// Other data for tables, based on LocalTable[][] value
// (mostly function as counters for audioonly waiting and blowup animation)
unsigned int LocalTableMiscData[10][10];
unsigned int EnemyTableMiscData[10][10];

// Other variables about the status of the game:
unsigned long ulRound = 0;  // Actual number of round
unsigned long ulScore = 0;  // Score (not used yet!)
unsigned long ulTime = 0;   // Elapsed time

// ID of timer that will be used to measure FPS value
#define TMR_FPSMEASURE 1

// Define special window message to notify main thread about skin/config change!
#define WM_CHANGECONFIG  (WM_USER + 123)

// Set image refresh rate to 30 FPS. (the valus of 32 will give us 30 in average (~28..32)
#define IMAGE_REFRESH_FPS 32.0

// Some "local" variables for the main thread:
HAB  Ab = 0;                         // Anchor block
HMQ  Mq = 0;                         // Message queue
int  MouseXPos = 0;                  // The current position of mouse, converted to
int  MouseYPos = 0;                  //   image coordinates
BOOL MouseLeftPressed = FALSE;       // The status of the mouse buttons
BOOL MouseLeftPressed_Prev = FALSE;
BOOL MouseRightPressed = FALSE;
BOOL MouseRightPressed_Prev = FALSE;

//                              ---- Code ----


// First some function heads, for forward declaration:
pSkin LoadSkin(char *skin_filename);
void CleanupSkin(pSkin *Skin);

//---------------------------------------------------------------------
// LibMain
//
// This gets called at DLL initialization and termination.
// We have nothing to do here.
//
// (it would be nice to implement cleanup code, so it would do
//  nice cleanup at crashes, but... hey, it won't crash anyway! ;) )
//---------------------------------------------------------------------

unsigned _System LibMain( unsigned hmod, unsigned termination )
{
  if (termination)
  {
    // Cleanup!
  } else
  {
    // Startup!
  }
  return 1;
}

//---------------------------------------------------------------------
// GetGameInfo
//
// NetGame Client calls this to get some information about this module.
// All we have to do here is to provide a GameInfoBuf, which describes
// this module. (See the declaration of that structure at the top of
// this file!)
//---------------------------------------------------------------------
BOOL FAR __syscall __export GetGameInfo( ULONG BufSize, pGameInfoBuf Buffer )
{
  // Check for valid buffer size
  if (BufSize!=sizeof(tGameInfoBuf)) return FALSE;

  memcpy(Buffer, &GameInfo, sizeof(GameInfo));
  return TRUE;
}

// ---- Now some local functions that are used by this module: ----

//---------------------------------------------------------------------
// WndProc
//
// This is the window procedure, handling events of the BattleShips
// window. It measures and shows FPS, configures DIVE when the visible
// region changes, and so on...
//---------------------------------------------------------------------
MRESULT EXPENTRY WndProc( HWND hwnd, USHORT msg, MPARAM mp1, MPARAM mp2 )
{
  HPS ps;
  SWP swp;
  RECTL rcl;
  HRGN rgn;
  RGNRECT rgnCtl;
  POINTL pl;
  RECTL rcls[50];
  SETUP_BLITTER SetupBlitter;
  char *NewName;

  switch (msg)
  {
    case WM_CREATE: // Called at the creation of the window
      {
	// Initialize counter:
	ulNumBlitted=0;
        // Start the FPS-measurer timer
	WinStartTimer(WinQueryAnchorBlock(hwnd), hwnd, TMR_FPSMEASURE, 1000);

	return (MRESULT) FALSE;
      }
    case WM_TIMER:  // Comes when the timer, created above, ticks
      {
	if (SHORT1FROMMP(mp1)==TMR_FPSMEASURE)
	{ // Ok, it's a message from our FPS measurer timer:
	  ActFPS=ulNumBlitted; // The actual FPS is the number of blitted images until now
	  ulNumBlitted=0;      // Restart counting

	  // Create a string like: BattleShips - 30 FPS
	  NewName=(char *)malloc(strlen(AppTitle)+80);
	  sprintf(NewName, "%s - %d FPS", AppTitle, ActFPS);
          // Show measured FPS in window text
	  WinSetWindowText(hFrame, NewName);
          free(NewName);
          // That's all!
	  return (MRESULT) FALSE;
        }
        break;
      }
    case WM_VRNDISABLED: // When blitting should be disabled
      {
	fVRNDisabled = TRUE;
	if (DosRequestMutexSem(hmtxUseDive, SEM_INDEFINITE_WAIT)==NO_ERROR)
	{
	  DiveSetupBlitter(hDiveInst, NULL); // Cancel blitting
	  DosReleaseMutexSem(hmtxUseDive);
	}
	return (MRESULT) FALSE;
      }
    case WM_VRNENABLED: // Blitting can be restarted, or the visible area
      {                 // of the window has changed
	if (DosRequestMutexSem(hmtxUseDive, SEM_INDEFINITE_WAIT)==NO_ERROR)
	{

	  ps = WinGetPS(hwnd);
	  rgn = GpiCreateRegion( ps, 0, NULL );
  
	  // Query the visible area of our window,
	  WinQueryVisibleRegion(hwnd, rgn);
  
	  rgnCtl.ircStart = 0;
	  rgnCtl.crc = 50;
	  rgnCtl.ulDirection = 1;
  
	  // in a form of maximum 50 OR'ed rectangles
  
	  GpiQueryRegionRects(ps, rgn, NULL, &rgnCtl, &rcls);
  
	  GpiDestroyRegion(ps, rgn);
  
	  WinReleasePS(ps);
  
	  // Now find the position and size, relative to parent
  
	  WinQueryWindowPos( hwnd, &swp );
  
	  // Convert the point to relative from lower-left corner of desktop
  
	  pl.x = swp.x;
	  pl.y = swp.y;
	  WinMapWindowPoints( hFrame, HWND_DESKTOP, &pl, 1);
  
	  // Tell the DIVE about the new settings
  
	  SetupBlitter.ulStructLen       = sizeof( SETUP_BLITTER );
	  SetupBlitter.fccSrcColorFormat = srcBufferFOURCC; // Pixel format of source image (RGB4, BGR4, R565, YUV8 etc...)
	  SetupBlitter.ulSrcWidth        = srcBufferWidth;  // Dimensions of source image
	  SetupBlitter.ulSrcHeight       = srcBufferHeight;
	  SetupBlitter.ulSrcPosX         = 0;
	  SetupBlitter.ulSrcPosY         = 0;
	  SetupBlitter.fInvert           = FALSE; // First line of buffer will be the top line of image
	  SetupBlitter.ulDitherType      = 1;     // Use 2x2 dithering if necessary
  
	  SetupBlitter.fccDstColorFormat = FOURCC_SCRN;   // Output format is as on the screen
	  SetupBlitter.lDstPosX          = 0;
	  SetupBlitter.lDstPosY          = 0;
  
	  SetupBlitter.ulDstWidth        = swp.cx;        // Destination size is as window allows
	  SetupBlitter.ulDstHeight       = swp.cy;

//        printf("Blit %d x %d -> %d x %d\n", srcBufferWidth, srcBufferHeight, swp.cx, swp.cy); fflush(stdout);

	  SetupBlitter.lScreenPosX       = pl.x;          // Destination position is relative to desktop
	  SetupBlitter.lScreenPosY       = pl.y;
  
	  SetupBlitter.ulNumDstRects     = rgnCtl.crcReturned;  // And the information about visible region
	  SetupBlitter.pVisDstRects      = &rcls;
  
          DiveSetupBlitter( hDiveInst, &SetupBlitter ); // So, setup blitter,
	  DosReleaseMutexSem(hmtxUseDive);
	  fVRNDisabled = FALSE;                         // and mark down that blitting can go
	}
	return (MRESULT) FALSE;
      }
    case WM_CHAR:      // When a key is pressed or released
      {
        // We do nothing here yet.
        // Will be used for chatting...
        break;
      }
    case WM_MOUSEMOVE: // When the mouse position is changed over out window
      {
        // Query the current size of our client area
        WinQueryWindowPos(hClient, &swp);
        // Convert mouse position to image position
	MouseXPos = (SHORT1FROMMP(mp1)*srcBufferWidth) / swp.cx;
        MouseYPos = ((swp.cy-SHORT2FROMMP(mp1)) * srcBufferHeight) / swp.cy;
        break;
      }
    case WM_BUTTON1DOWN:   // Mouse button pressed
      {
	MouseLeftPressed=TRUE;
        break;
      }
    case WM_BUTTON1UP:     // Mouse button released
      {
	MouseLeftPressed=FALSE;
        break;
      }
    case WM_BUTTON2DOWN:   // Mouse button pressed
      {
	MouseRightPressed=TRUE;
        break;
      }
    case WM_BUTTON2UP:     // Mouse button released
      {
	MouseRightPressed=FALSE;
        break;
      }
    case WM_CLOSE:         // The close button has been pressed
      {
	WinPostMsg( hwnd, WM_QUIT, 0, 0);
	return (MRESULT) FALSE;
      }
    default:
        break;
  }
  // If it wasn't handled, let the default handler handle it!
  return WinDefWindowProc( hwnd, msg, mp1, mp2 );
}

//---------------------------------------------------------------------
// SetupMainWindow
//
// Creates and positions the BattleShips window
//---------------------------------------------------------------------
int SetupMainWindow()
{
  ULONG fcf;

#ifdef DEBUG_BUILD
  printf("SetupMainWindow()\n");
  fflush(stdout);
#endif

  // Register our own window class
  if (!WinRegisterClass(Ab, (PCH) WindowClassName, (PFNWP) WndProc, CS_SIZEREDRAW, 0))
    return FALSE;

  // Create a simple window that will contain only the DIVE output
  fcf =
    FCF_TITLEBAR |
    FCF_SYSMENU |
    FCF_MINBUTTON |
    FCF_MAXBUTTON |
    FCF_SIZEBORDER |
    FCF_TASKLIST;

  hFrame = WinCreateStdWindow( HWND_DESKTOP, 0, &fcf,
			       WindowClassName, AppTitle, 0,
			       0, 0, (HWND FAR *) &hClient );

  if (!hFrame)
    return FALSE;

  WindowIcon=WinLoadPointer(HWND_DESKTOP, DLLHandle, DLG_SETUPWINDOW);
  WinPostMsg(hFrame, WM_SETICON, (MPARAM) WindowIcon, 0);

  // Set its size to the source buffer's size, and position it to the
  // middle of the screen
  WinSetWindowPos( hFrame,
		   HWND_TOP,
		   (WinQuerySysValue( HWND_DESKTOP, SV_CXSCREEN) - srcBufferWidth - WinQuerySysValue( HWND_DESKTOP, SV_CXDLGFRAME)*2)/2,
		   (WinQuerySysValue( HWND_DESKTOP, SV_CYSCREEN) - srcBufferHeight - WinQuerySysValue( HWND_DESKTOP, SV_CYDLGFRAME)*2)/2,
		   srcBufferWidth+WinQuerySysValue( HWND_DESKTOP, SV_CXDLGFRAME)*2,
		   srcBufferHeight + WinQuerySysValue( HWND_DESKTOP, SV_CYTITLEBAR)
		   + WinQuerySysValue( HWND_DESKTOP, SV_CYDLGFRAME)*2
		   + 2,
                   SWP_SIZE | SWP_ACTIVATE | SWP_SHOW | SWP_MOVE );

  return TRUE;
}

//---------------------------------------------------------------------
// DestroyMainWindow
//
// Destroys the BattleShips window
//---------------------------------------------------------------------
void DestroyMainWindow()
{
#ifdef DEBUG_BUILD
  printf("DestroyMainWindow()\n");
  fflush(stdout);
#endif

  WinDestroyWindow(hFrame);
  WinDestroyPointer(WindowIcon);
}

//---------------------------------------------------------------------
// ShutdownDIVE
//
// Stops the sending of Visible Region change Notification messages,
// frees allocated DIVE buffer(s) and closes DIVE.
//---------------------------------------------------------------------
void ShutdownDIVE()
{
#ifdef DEBUG_BUILD
  printf("ShutdownDIVE()\n");
  fflush(stdout);
#endif

  WinSetVisibleRegionNotify( hClient, FALSE ); // Stop WM_VRN* messages

  if (DosRequestMutexSem( hmtxUseDive, SEM_INDEFINITE_WAIT )==NO_ERROR)
  {
    DiveFreeImageBuffer( hDiveInst, DiveImageHandle ); // Free buffer
    DiveClose( hDiveInst );                            // Close DIVE instance
    DosReleaseMutexSem(hmtxUseDive);
  }
}

//---------------------------------------------------------------------
// SetupDIVE
//
// Opens DIVE, allocates DIVE buffer(s) and starts sending VRN messages
//---------------------------------------------------------------------
int SetupDIVE()
{
  void *pFrameBuffer;

#ifdef DEBUG_BUILD
  printf("SetupDIVE()\n");
  fflush(stdout);
#endif

  if (DosRequestMutexSem( hmtxUseDive, SEM_INDEFINITE_WAIT ) == NO_ERROR)
  {
    fVRNDisabled=TRUE;

    // Try to open DIVE
    if (DiveOpen( &hDiveInst, FALSE, &pFrameBuffer )!=NO_ERROR)
    {
#ifdef DEBUG_BUILD
      printf("SetupDIVE(): DiveOpen error!\n");
      fflush(stdout);
#endif

      DosReleaseMutexSem( hmtxUseDive );
      return FALSE; // Error if could not open
    }

    // Let DIVE allocate needed buffer from hardware, if possible
    DiveAllocImageBuffer( hDiveInst,          
			  &DiveImageHandle,
			  srcBufferFOURCC,
			  srcBufferWidth,
			  srcBufferHeight,
			  0,
			  NULL );
#ifdef DEBUG_BUILD
    printf("  DiveAllocImageBuffer of %d x %d!\n", srcBufferWidth, srcBufferHeight);
    fflush(stdout);
#endif

    // Get WM_VRN* messages when visible region changes!
    WinSetVisibleRegionNotify( hClient, TRUE );

    // Initialize blitter (by WndProc)!
    WinPostMsg( hClient, WM_VRNENABLED, 0, 0);

    DosReleaseMutexSem( hmtxUseDive );
  }
#ifdef DEBUG_BUILD
  else
  {
    printf("SetupDIVE(): DosRequestMutexSem error!\n");
    fflush(stdout);
  }
#endif

  return TRUE;
}

//---------------------------------------------------------------------
// ImgPos2TablePos
//
// Given the mouse coordinates and table definition, it calculates the
// position of table, where the mouse is pointing to, or returns 
// (-1;-1) .
// Note: It can be that it returns out of table values, if the mouse is
// not over the table, so it has to be checked if the returned values 
// are between 0 and 10 (or whatever dimensions the table has...)
//---------------------------------------------------------------------
void ImgPos2TablePos(long TableTopLeftX, long TableTopLeftY,    // Top-left corner of the table
		     long TableXOneStepX, long TableXOneStepY,  // One step to the right
                     long TableYOneStepX, long TableYOneStepY,  // One step down
		     long mx, long my,   // Mouse coordinates
		     long *tx, long *ty) // Results (table coordinates)
{
  double A1, B1, A2, B2, Cx, Cy;
  double PCx, PCy, MCx, MCy, Dx, Dy, Rx, Ry;
  int pvx, pvy;

  *ty=*tx=-1; // Set default result to failure!

  // Check if the mouse is at the left side of the down vector!
  // (because it's for sure not over the table then, but to the left of it)
  pvx = mx - TableTopLeftX;
  pvy = my - TableTopLeftY;
  if ((pvx*(-TableYOneStepY) + pvy*TableYOneStepX)>0) return;

  // Check if the mouse is at the left side of the right vector!
  // (because it's for sure not over the table then, but above it)
  if ((pvx*(-TableXOneStepY) + pvy*TableXOneStepX)<0) return;

  // Ok, the mouse is over the table, or to the left or under it.
  // Let's calculate, how far it is from the top-left corner of the
  // given table, in table units!

  Dx=TableYOneStepX; Dy=TableYOneStepY;
  Rx=TableXOneStepX; Ry=TableXOneStepY;

  if ((Dx!=0) && (Rx!=0))
  {
    // y1=A1*x1+B1 (D)
    // y2=A2*x2+B2 (R)
    // Compute the crossing of these to get the C point!

    A1=Dy / Dx;
    B1=(double)TableTopLeftY - A1*TableTopLeftX;
    A2=Ry / Rx;
    B2=(double)my - A2*mx;

    if (A1==A2) return; // Cannot be computed!
    Cx=(B2-B1)/(A1-A2);
    Cy=A1*Cx+B1;
  } else
  if ((Rx==0) && (Dx!=0))
  { // Rx==0!
    // y1=A1*x1+B1 (D)
    // x2=A2*y2-B2 (R)
    // Compute the crossing of these to get the C point!

    if (Ry==0) return; // Cannot be computed!

    A1=Dy / Dx;
    B1=(double)TableTopLeftY - A1*TableTopLeftX;
    A2=Rx / Ry;
    B2=(double)A2*my-mx;

    if (A1*A2==1) return; // Cannot be computed!
    Cy=(B1-A1*A2)/(1-A1*A2);
    Cx=A2*Cy-B2;

  } else
  if ((Rx!=0) && (Dx==0))
  { // Dx==0!
    // x1=A1*y1-B1 (D)
    // y2=A2*x2+B2 (R)
    // Compute the crossing of these to get the C point!

    if (Dy==0) return; // Cannot be computed!

    A1=Dx / Dy;
    B1=A1*(double)TableTopLeftY - TableTopLeftX;
    A2=Ry / Rx;
    B2=(double)my - A2*mx;

    if (A1*A2==1) return; // Cannot be computed!
    Cy=(B2-A2*A1)/(1-A2*A1);
    Cx=A1*Cy-B1;

  } else
  if ((Rx==0) && (Dx==0))
  {
    // x1=A1*y1-B1 (D)
    // x2=A2*y2+B2 (R)
    // Compute the crossing of these to get the C point!
    if ((Dy==0) || (Ry==0)) return; // Cannot be computed!
    A1=Dx / Dy;
    A2=Rx / Ry;
    B1=A1*(double)TableTopLeftY - TableTopLeftX;
    B2=(double)A2*my-mx;

    if (A1==A2) return; // Cannot be computed!
    Cy=(B1-B2)/(A1-A2);
    Cx=A1*Cy-B1;
  }

  PCx=Cx-TableTopLeftX; PCy=Cy-TableTopLeftY;
  MCx=Cx-mx; MCy=Cy-my;

  A1=sqrt(Rx*Rx+Ry*Ry);
  A2=sqrt(Dx*Dx+Dy*Dy);
  if ((A1!=0) && (A2!=0))
  {
    *tx = sqrt(MCx*MCx+MCy*MCy)/A1;
    *ty = sqrt(PCx*PCx+PCy*PCy)/A2;
  }
}

//---------------------------------------------------------------------
// ShipStartingAtPosition
//
// Tells the ID of ship starting at a given table position, or 
// returns -1, if there is no ship there.
//---------------------------------------------------------------------
int ShipStartingAtPosition(int x, int y)
{ 
  int i, j;

  // First check if it fits to the table
  if ((x<0) || (x>9) || (y<0) || (y>9)) return -1;

  for (i=0; i<NUM_OF_SHIPS; i++)
    if ((ShipPosition[i][0]==x) && (ShipPosition[i][1]==y)) return i;
  return -1;
}

//---------------------------------------------------------------------
// CheckForGameOver
//
// Checks if all enemy ships are destroyed, or all of our ships are
// destroyed, and sets gamestate if needed.
//---------------------------------------------------------------------
void CheckForGameOver()
{
  int x,y;
  int AllHits = 0;   // Number of hits needed to sink all ships
  int LocalHits = 0; // Number of hits in LocaTable
  int EnemyHits = 0; // Number of hits in EnemyTable

  // First calculate the total number of hits needed to sink all ships
  for (x=0; x<NUM_OF_SHIPS;x++) AllHits+=Ship_Length[x];

  // Now count the number of hits in tables
  for (y=0; y<10; y++)
    for (x=0; x<10; x++)
    {
      if ((LocalTable[x][y]==TABLEENTRY_HIT) ||
          (LocalTable[x][y]==TABLEENTRY_HIT_AUDIOONLY) ||
	  (LocalTable[x][y]==TABLEENTRY_HIT_AUDIOANDBLOWUP)) LocalHits++;
      if ((EnemyTable[x][y]==TABLEENTRY_HIT) ||
          (EnemyTable[x][y]==TABLEENTRY_HIT_AUDIOONLY) ||
	  (EnemyTable[x][y]==TABLEENTRY_HIT_AUDIOANDBLOWUP)) EnemyHits++;
    }

  // Now check for gameover
  if (EnemyHits == AllHits)
  {
    iGameState = GAMESTATE_END_OF_GAME_WON;
    fFirstTimeRender = TRUE;  // Make sure the new background will be drawn!
    NewMessage_to_MessageBoard(Skin->msgGameWon);
    // Also make audible notification!
    SFXPlayFX(Skin->GameWonSFX);
  }
  if (LocalHits == AllHits)
  {
    iGameState = GAMESTATE_END_OF_GAME_LOST;
    fFirstTimeRender = TRUE;   // Make sure the new background will be drawn!
    NewMessage_to_MessageBoard(Skin->msgGameLost);
    // Also make audible notification!
    SFXPlayFX(Skin->GameLostSFX);
  }
}

//---------------------------------------------------------------------
// ShipAtPosition
//
// It's like ShipStartingAtPosition, but it returns the ship ID of
// the ship that takes place at a given table position, not necessary
// starting there.
//---------------------------------------------------------------------
int ShipAtPosition(int x, int y)
{ 
  int i, j;

  // First check if it fits to the table
  if ((x<0) || (x>9) || (y<0) || (y>9)) return -1;

  for (i=0; i<NUM_OF_SHIPS; i++)
  {
    if ((ShipPosition[i][0]>-1) && (ShipPosition[i][1]>-1))
    { // This ship is on the table, check it's position!
      if ((!ShipDirection[i]) && (ShipPosition[i][0]==x))
      { // it's a DownShip in our column
	for (j=ShipPosition[i][1]; j<ShipPosition[i][1]+Ship_Length[i]; j++)
          if (j==y) return i;
      } else
      if ((ShipDirection[i]) && (ShipPosition[i][1]==y))
      { // it's a RightShip in our row
        for (j=ShipPosition[i][0]; j<ShipPosition[i][0]+Ship_Length[i]; j++)
          if (j==x) return i;
      }
    }
  }
  return -1;
}

//---------------------------------------------------------------------
// AcceptableShipPosition
//
// Returns TRUE, if a given table position is OK to place a given ship
// to, because it will not be placed over any other ships on the table.
// Returns FALSE otherwise.
//---------------------------------------------------------------------
int AcceptableShipPosition(int x, int y, int ShipNum, int ShipDir)
{
  int i, j;

  // First check if it fits to the table
  if ((x<0) || (x>9) || (y<0) || (y>9) || (ShipNum<0) || (ShipNum>=NUM_OF_SHIPS)) return 0;

  // Then if any of the other ships are under us!
  if (!ShipDir)
  {
    // Down-Ship
    if (y>(10-Ship_Length[ShipNum])) return 0;

    for (j=y; j<y+Ship_Length[ShipNum]; j++)
    {
      i=ShipAtPosition(x,j);
      if ((i>-1) && (i!=ShipNum)) return 0;
    }
    return 1; // Ok, it fits!
  }
  else
  {
    // Right-Ship
    if (x>(10-Ship_Length[ShipNum])) return 0;

    for (j=x; j<x+Ship_Length[ShipNum]; j++)
    {
      i=ShipAtPosition(j,y);
      if ((i>-1) && (i!=ShipNum)) return 0;
    }
    return 1; // Ok, it fits!
  }
}

//---------------------------------------------------------------------
// DrawShipAndNeightbours
//
// Draws a ship to a given table. Takes care of visibility, so it first
// calls itself to draws ships which should be covered by this ship,
// only then draws this ship.
//---------------------------------------------------------------------
void DrawShipAndNeightbours(unsigned char *pBuffer, int ShipNum, int *ShipDrawn, int Depth, pTableDef Table)
{
  int i;
  long TableTopLeftCornerX, TableTopLeftCornerY;
  long TableXOneStepX, TableXOneStepY;
  long TableYOneStepX, TableYOneStepY;
  long TableOrientation;
  pSprite Ship;

  if ((Depth>NUM_OF_SHIPS) || (!Table)) return;

  if (Table->TableOrientation==0)
  { // Draw from left to right
    // First draw every other ship that is to the left of our ship, or above!
    for (i=0; i<NUM_OF_SHIPS; i++)
      if ((i!=ShipNum) && (!ShipDrawn[i]) &&
  	  (ShipPosition[i][0]<=ShipPosition[ShipNum][0]) &&
  	  (ShipPosition[i][1]<=ShipPosition[ShipNum][1]))
        DrawShipAndNeightbours(pBuffer, i, ShipDrawn, Depth+1, Table);
  } else
  { // Draw from right to left
    // First draw every other ship that is to the right of our ship, or above!
    for (i=0; i<NUM_OF_SHIPS; i++)
      if ((i!=ShipNum) && (!ShipDrawn[i]) &&
  	  (ShipPosition[i][0]>=ShipPosition[ShipNum][0]) &&
  	  (ShipPosition[i][1]<=ShipPosition[ShipNum][1]))
        DrawShipAndNeightbours(pBuffer, i, ShipDrawn, Depth+1, Table);
  }

  // Now that every ship that is potentionally covered by this ship is
  // drawn, we can draw this ship too!

  // Get a defined ship sprite according to the number of hits it got,
  // and according to its direction!

  if (ShipDirection[ShipNum])
  { // Horizontal ship
    i = ShipNumOfHits[ShipNum];
    while ((!Skin->Ship_Right[ShipNum][i]) && (i>0)) i--;
    Ship = Skin->Ship_Right[ShipNum][i];
  }
  else
  { // Vertical ship
    i = ShipNumOfHits[ShipNum];
    while ((!Skin->Ship_Down[ShipNum][i]) && (i>0)) i--;
    Ship = Skin->Ship_Down[ShipNum][i];
  }

  // Now put this sprite to the screen

  DrawSprite(pBuffer, Skin->BackgroundImage->ulWidth,
             0, 0,
             Skin->BackgroundImage->ulWidth, Skin->BackgroundImage->ulHeight,
	     ShipPosition[ShipNum][0]*Table->TableXOneStepX + ShipPosition[ShipNum][1]*Table->TableYOneStepX +Table->TableTopLeftCornerX,
	     ShipPosition[ShipNum][0]*Table->TableXOneStepY + ShipPosition[ShipNum][1]*Table->TableYOneStepY +Table->TableTopLeftCornerY,
	     Ship, ShipVisibility[ShipNum]);

  // And note that this ship has been drawn
  ShipDrawn[ShipNum]=1;
}

//---------------------------------------------------------------------
// GetButtonSprite
//
// Returns the actual sprite of a button, based on its status.
//---------------------------------------------------------------------
pSprite GetButtonSprite(pButton Button)
{
  if (!Button) return NULL;
  switch (Button->iState)
  {
    case BUTTONSTATE_DISABLED:
      return Button->sprDisabled;
    case BUTTONSTATE_ACTIVE:
      return Button->sprActive;
    case BUTTONSTATE_MOUSEOVER:
      return Button->sprMouseOver;
    case BUTTONSTATE_PRESSED:
      return Button->sprPressed;
    default: break;
  }
  return NULL;
}

//---------------------------------------------------------------------
// DrawButton
//
// Draws and animates a given button
//---------------------------------------------------------------------
void DrawButton(unsigned char *pBuffer, pButton Button)
{
  pSprite Sprite;

  if ((!Button) || (!pBuffer)) return;

  Sprite = GetButtonSprite(Button);
  if (!Sprite) return;

  DrawSprite(pBuffer, Skin->BackgroundImage->ulWidth,
	     0, 0,
	     Skin->BackgroundImage->ulWidth, Skin->BackgroundImage->ulHeight,
	     Button->XPos, Button->YPos,
	     Sprite, 255);
  AnimateSprite(Sprite);
}

//---------------------------------------------------------------------
// MessageBoard_Processing
//
// This function is called at every screen update (~30FPS), and it is
// responsible for the messages on the message board. It checks if there
// are new messages to show, and scrolls them in.
//---------------------------------------------------------------------
void MessageBoard_Processing(unsigned char *pBuffer, pSkin Skin)
{
  int iNumOfLines;
  int iLineLengths[10];
  int iLineWidths[10];
  int iFullHeight, iMidY, iMidX;
  int x,y,i;
  char *msg;
  int iVisibility;

  // Do nothing, if there is no destination buffer, no skin,
  // no message to show, or no font that can be used to show the message.
  if ((!pBuffer) || (!Skin)) return;
  if ((!ActualMessage) || (!Skin->Font1)) return;

  // First check the message to show!
  // Get some statistical data about it, like number of lines,
  // width of lines etc...
  if (VisibleMessage)
  {
    msg=VisibleMessage;
    iNumOfLines=1;
    iLineLengths[0]=0;
    iLineWidths[0]=0;
    while (*msg)
    {
      if (*msg=='\n')
      { // There is an Enter in the string, so it has one more line!
	if ((iLineLengths[iNumOfLines-1]) && (iNumOfLines<9))
	{
	  iNumOfLines++;
	  iLineLengths[iNumOfLines-1]=0;
	  iLineWidths[iNumOfLines-1]=0;
	}
      } else
      { // This character is not an enter, so the current lineis one more charecters long
	iLineLengths[iNumOfLines-1]++;
	if (Skin->Font1->Character[*msg])
	{ // And also count the width in pixels!
	  iLineWidths[iNumOfLines-1]+=(Skin->Font1->Character[*msg]->ulWidth - Skin->Font1->Character[*msg]->ulHotSpotX);
	}
      }
      msg++;
    }
  } else
  { // If there is no message to show, then the number of lines is zero:
    iNumOfLines=0;
  }
  // Calculate the full height of the current message
  iFullHeight = iNumOfLines*Skin->Font1->iHeight;
  // And the middle point of the text area
  iMidY = (Skin->TextAreaY1+Skin->TextAreaY2)/2;
  iMidX = (Skin->TextAreaX1+Skin->TextAreaX2)/2;

  // Move/Scroll the message, if needed!
  if (!MessageBoard_NewMessage)
  {
    // If there is no new message, but this message has not been scrolled in
    // to the middle of the message are, scroll it!
    if (MessagePos<0) MessagePos+=(labs(MessagePos)/10+1);
  } else
  {
    // If there is a new message to show, we should first scroll out the
    // old message, then scroll in the new one.
    // First thing is to check if the old message is scrolled out, and set
    // its new position:
    if (MessagePos<=iFullHeight)
      MessagePos+=(labs((labs(iFullHeight)/10)-(labs(iFullHeight-MessagePos)/10))+1);
    else
    { // The old message has been scrolled out, let the new message come!
      MessageBoard_NewMessage=0;
      VisibleMessage = ActualMessage;
      // Get information about the new message!
      // (The same code as some lines above)
      msg=VisibleMessage;
      iNumOfLines=1;
      iLineLengths[0]=0;
      iLineWidths[0]=0;
      while (*msg)
      {
	if (*msg=='\n')
	{
	  if ((iLineLengths[iNumOfLines-1]) && (iNumOfLines<9))
	  {
            iNumOfLines++;
	    iLineLengths[iNumOfLines-1]=0;
	    iLineWidths[iNumOfLines-1]=0;
	  }
	} else
	{
	  iLineLengths[iNumOfLines-1]++;
	  if (Skin->Font1->Character[*msg])
	  {
	    iLineWidths[iNumOfLines-1]+=(Skin->Font1->Character[*msg]->ulWidth - Skin->Font1->Character[*msg]->ulHotSpotX);
	  }
	}
	msg++;
      }
      iFullHeight = iNumOfLines*Skin->Font1->iHeight;
      iMidY = (Skin->TextAreaY1+Skin->TextAreaY2)/2;
      iMidX = (Skin->TextAreaX1+Skin->TextAreaX2)/2;

      MessagePos = -iMidY; // Start position is out of the text area
    }
  }

  // Now that the message is positioned, let's draw it!

  if (VisibleMessage)
  { // If there is a message to show:
    msg=VisibleMessage;
    y = iMidY-iFullHeight/2 + Skin->Font1->iHeight;
    x = iMidX-iLineWidths[0]/2;
    y+=MessagePos;
    i=1;
    while (*msg)
    {
      if (*msg=='\n')
      { // Next line
	if ((iLineLengths[i-1]) && (i<9))
	{
	  // Step in Y coordinate
	  y+=Skin->Font1->iHeight;
	  // Go to next line of message
	  i++;
	  // And return to the beginning in X coordinate
	  x = iMidX-iLineWidths[i-1]/2;
	}
      } else
      { // Continue with letters in this line
	if (Skin->Font1->Character[*msg])
	{
	  // The transparency of this line is relative to its Y position:
	  iVisibility = 255-(labs(MessagePos)*255/abs(iMidY));
	  if (iVisibility<0) iVisibility=0;
          if (iVisibility>255) iVisibility=255;
          // Draw the letter!
	  DrawSprite(pBuffer, Skin->BackgroundImage->ulWidth,
		     Skin->TextAreaX1, Skin->TextAreaY1,
		     Skin->TextAreaX2, Skin->TextAreaY2,
		     x, y, Skin->Font1->Character[*msg], iVisibility);
	  // Step in X coordinate
	  x+=Skin->Font1->Character[*msg]->ulWidth - Skin->Font1->Character[*msg]->ulHotSpotX;
	}
      }
      msg++;
    }
  }
}

//---------------------------------------------------------------------
// NewMessage_to_MessageBoard
//
// Sets a given string to be displayed on the messageboard.
// The pointer to the string should be valid while it is displayed, so
// don't free it after passing to this function!
//---------------------------------------------------------------------
void NewMessage_to_MessageBoard(char *message)
{
  ActualMessage = message;
  MessageBoard_NewMessage=TRUE;
}

//---------------------------------------------------------------------
// MouseIsOverButton
//
// Returns TRUE if the mouse is over a given button.
//---------------------------------------------------------------------
int MouseIsOverButton(pButton Button)
{
  pSprite Sprite;
  int x1,y1,x2,y2;

  if (!Button) return 0;
  Sprite = GetButtonSprite(Button);

  x1 = (Button->XPos - Sprite->ulHotSpotX);
  y1 = (Button->YPos - Sprite->ulHotSpotY);
  x2 = x1 + Sprite->ulWidth-1;
  y2 = y1 + Sprite->ulHeight-1;

  return ((MouseXPos>=x1) &&
          (MouseYPos>=y1) &&
	  (MouseXPos<=x2) &&
          (MouseYPos<=y2));
}

// Some callbacks, to be called when a given button is pressed:

//---------------------------------------------------------------------
// QuitButtonPressed
//
// Gets called when the Quit button is pressed and released. 
//---------------------------------------------------------------------
void QuitButtonPressed()
{
  WinPostMsg(hFrame, WM_CLOSE, NULL, NULL);
}

//---------------------------------------------------------------------
// VolumeButtonPressed
//
// Gets called when the Volume button is pressed and released.
//---------------------------------------------------------------------
void VolumeButtonPressed()
{
#ifdef DEBUG_BUILD
  printf("VolumeButtonPressed()\n");
  fflush(stdout);
#endif
  bVolumePanelActive = TRUE; // Activate volume panel!
  iVolumePanelState = VOLUMEPANELSTATE_NORMAL;
  Skin->VolumeButton->iState = BUTTONSTATE_MOUSEOVER;
}

//---------------------------------------------------------------------
// SetupThreadFunc
//
// This implements the setup thread.
//---------------------------------------------------------------------
VOID SetupThreadFunc(PVOID Param)
{
  HAB ab = WinInitialize(0);
  HMQ mq = WinCreateMsgQueue(ab, 0);
#ifdef DEBUG_BUILD
  printf("Calling SetupWindowProcessing()...\n");
  fflush(stdout);
#endif

  SetupWindowProcessing(ab, mq, hFrame, DLLHandle, SETUP_CALLED_FROM_GAMEDLL);

#ifdef DEBUG_BUILD
  printf("SetupWindowProcessing() Returned, stopping setup thread!\n");
  fflush(stdout);
#endif

  WinDestroyMsgQueue(mq);
  WinTerminate(ab);
  SetupThread = -1;
  _endthread();
}

//---------------------------------------------------------------------
// SetupButtonPressed
//
// Gets called when the Setup button is pressed and released.
//---------------------------------------------------------------------
void SetupButtonPressed()
{
#ifdef DEBUG_BUILD
  printf("SetupButtonPressed()\n");
  fflush(stdout);
#endif

  if (SetupThread==-1)
    SetupThread = _beginthread( SetupThreadFunc,
				NULL,
				16384,
				NULL);
  Skin->SetupButton->iState = BUTTONSTATE_MOUSEOVER;
}

//---------------------------------------------------------------------
// StartButtonPressed
//
// Gets called when the Start button is pressed and released. 
//---------------------------------------------------------------------
void StartButtonPressed()
{
  // Set the new game state!
  if (iGameState == GAMESTATE_POSITION_SELF)
  {
    iGameState = GAMESTATE_WAIT_FOR_POSITION_ENEMY;
    fFirstTimeRender = TRUE;  // Make sure the new background will be drawn!
    NewMessage_to_MessageBoard(Skin->msgWaitingForEnemy);
  }
  else
  { // If the enemy has already positioned itself, the game can start.
    // Let's see if we can start shooting, or the enemy!
    if (fWeAreTheServer)
    {
      iGameState = GAMESTATE_SELF_ROUND;
      Initialize_ulInitTimeStamp(); // Start counting in-game time!
      fFirstTimeRender = TRUE;      // Make sure the new background will be drawn!
      NewMessage_to_MessageBoard(Skin->msgWePositionedTooSrv);
    } else
    {
      iGameState = GAMESTATE_ENEMY_ROUND;
      Initialize_ulInitTimeStamp(); // Start counting in-game time!
      fFirstTimeRender = TRUE;      // Make sure the new background will be drawn!
      NewMessage_to_MessageBoard(Skin->msgWePositionedTooCli);
    }
  }
  // Also make audible notification!
  if (Skin)
    SFXPlayFX(Skin->GameStartSFX);
  
  // Set the new button state: disabled
  Skin->StartButton->iState = BUTTONSTATE_DISABLED;
  Skin->StartButton->sprDisabled->flActFrame = 0;  // Button animation starting from frame 0
  // Also send a message to the other side that we've positioned the ships (too)
  (*pSendMessage)(Socket, "POSITIONED");
}

//---------------------------------------------------------------------
// UpdateActiveButtonState
//
// Updates the state of a given button, according to the mouse position
// and status of mouse buttons. Calls a given function, if the button
// is pressed and released.
//---------------------------------------------------------------------
void UpdateActiveButtonState(pSkin Skin, pButton Button, void ButtonPressedFn())
{
  if ((!Button) || (!ButtonPressedFn)) return;

  if (MouseIsOverButton(Button))
  { // The mouse is over this button,
    if (Button->iState==BUTTONSTATE_PRESSED)
    { // and the button is already pressed,
      if (MouseLeftPressed)
        // and the mouse button is pressed too:
	Button->iState=BUTTONSTATE_PRESSED;
      else
      {
        // The button was pressed, and the mouse button has been released here.
        // It means that its a button-press event!

        // Make sound
        SFXPlayFX(Skin->PressButtonSFX);
        // Call the function that will do the job
        (*ButtonPressedFn)();
      }
    } else
    { // If the button is not pressed,
      if ((MouseLeftPressed) && (!MouseLeftPressed_Prev))
      { // but just right now pressed the mouse button over this button:
	Button->iState=BUTTONSTATE_PRESSED;
	// Restart animation too!
	Button->sprPressed->flActFrame = 0;
      }
      else
      { // If the button is not pressed, and the mouse button is not pressed too:
	if (Button->iState != BUTTONSTATE_MOUSEOVER)
        { 
          SFXPlayFX(Skin->OverButtonSFX);
	  Button->iState=BUTTONSTATE_MOUSEOVER;
	  // Restart animation too!
	  Button->sprMouseOver->flActFrame = 0;
	}
      }
    }
  }
  else
  { // If the mouse is not over this button:
    if (Button->iState != BUTTONSTATE_ACTIVE)
    { // but it was just before, let's leave it:
      SFXPlayFX(Skin->LeaveButtonSFX);
      Button->iState=BUTTONSTATE_ACTIVE;
      // Restart animation too!
      Button->sprActive->flActFrame = 0;
    }
  }
}

//---------------------------------------------------------------------
// RenderOneFrame_when_PositioningSelf
//
// Renders one frame, when the gamestate is POSITION_SELF
//---------------------------------------------------------------------
void RenderOneFrame_when_PositioningSelf(unsigned char *pBuffer)
{
  long xtemp, ytemp, TableCursorXPos, TableCursorYPos;
  int i, ShipNum;
  int iVisibility;
  int PointedShip;
  pSprite psTemp;

  // First of all, restore the old background by removing the sprites.
  // It's faster than always memcpy'ing the whole background image.
  RemoveDrawnSprites(pBuffer, Skin->BackgroundImage->ulWidth);

  if (fFirstTimeRender) // if it's the first time to build the output image, we should set up the background
			// using memcpy. Otherwise, we should remove the sprites, because that's faster than
			// the memcpy of the whole background...
  {
    if (Skin->BackgroundImage)
      memcpy(pBuffer, Skin->BackgroundImage->pImageData, Skin->BackgroundImage->ulImageSize);
    fFirstTimeRender = FALSE;
  }
  
  // Calculate the position of cursor if over the table:
  TableCursorXPos = TableCursorYPos = -1;
  ImgPos2TablePos(Skin->Table.TableTopLeftCornerX, Skin->Table.TableTopLeftCornerY,
		  Skin->Table.TableXOneStepX, Skin->Table.TableXOneStepY,
		  Skin->Table.TableYOneStepX, Skin->Table.TableYOneStepY,
		  MouseXPos, MouseYPos,
		  &xtemp, &ytemp);
  if ((xtemp>=0) && (ytemp>=0) && (xtemp<=9) && (ytemp<=9))
  {
    TableCursorXPos = xtemp;
    TableCursorYPos = ytemp;
  }

  // First, we check if the user is pointing onto the table, to a ship, and
  // we will remember the ID of that ship! (-1 otherwise)

  if (!bVolumePanelActive)
    PointedShip = ShipAtPosition(TableCursorXPos, TableCursorYPos);
  else
    PointedShip = -1; // Don't let actions here!

  // Now it's time to draw something!

  // First put there the message board text. It will be the bottom-most
  // drawing on the screen.
  MessageBoard_Processing(pBuffer, Skin);

  // Now show the ships on the table:
  // (Should be done in order of visibility!)
  {
    // We a flag for every ship to remember if it is already drawn or not
    int ShipDrawn[NUM_OF_SHIPS];

    // Now for every ship:
    for (i=0; i<NUM_OF_SHIPS; i++)
    {
      // Mark the ships that are on the table, and have to be drawn.
      // Also calculate their transparency.

      // Draw only the ones that are on the table. For the others, set
      // that it has already been drawn, so it won't be drawn here.
      if ((ShipPosition[i][0]==-1) && (ShipPosition[i][1]==-1))
	ShipDrawn[i] = 1;
      else
      { // This should be on the table. Make it fully visible,
	// or half-visible, if its not the pointed one!
	if ((iDragShip>-1) ||
	    ((iDragShip==-1) && (PointedShip>-1) && (i!=PointedShip)))
	{ // if there is a ship being dragged, or
	  //    there is no ship being dragged, and the user 
	  //    is pointing to a ship, which is not this ship,
	  //  then show this ship a bit transparent.
	  if (ShipVisibility[i]>155) ShipVisibility[i]-=20;
	  if (ShipVisibility[i]<155) ShipVisibility[i]=155;
	} else
	{ // Otherwise, show it fully visible!
	  if (ShipVisibility[i]<255) ShipVisibility[i]+=20;
	  if (ShipVisibility[i]>255) ShipVisibility[i]=255;
	}
	ShipDrawn[i] = 0; // And remember, it will have to be drawn!
      }
    }
    // Now draw the ships!
    for (i=0; i<NUM_OF_SHIPS; i++)
      if (!ShipDrawn[i]) DrawShipAndNeightbours(pBuffer, i, ShipDrawn, 0, &(Skin->Table));

  }

  // Now show the ships that are at their start positions! (not positioned yet)
  for (i=0; i<NUM_OF_SHIPS; i++)
  {
    ShipNum=Skin->Ship_StartPosition_DrawOrder[i];
    if (ShipNum!=iDragShip) // If it's not dragged
    {
      if ((ShipPosition[ShipNum][0]==-1) &&
	  (ShipPosition[ShipNum][1]==-1))
      { // This ship is still not on the table, so show it at its base position!
	// If the pointer is at that ship, make it fully visible, otherwise only half-visible!
	ImgPos2TablePos(Skin->Ship_StartPositionX[ShipNum],
			Skin->Ship_StartPositionY[ShipNum],
			Skin->Table.TableXOneStepX, Skin->Table.TableXOneStepY,
			Skin->Table.TableYOneStepX, Skin->Table.TableYOneStepY,
			MouseXPos, MouseYPos,
			&xtemp, &ytemp);

	if (ShipVisibility[ShipNum]<145)
	{
	  ShipVisibility[ShipNum]+=(145-ShipVisibility[ShipNum])/60+5;
	  if (ShipVisibility[ShipNum]>145) ShipVisibility[ShipNum]=145;
	}

	if ((xtemp==0) && (ytemp>=0) &&
	    (ytemp<Ship_Length[ShipNum])) // Take into account the size of the ship
	{ // The mouse cursor is over this ship!
	  if ((!bVolumePanelActive) && (iDragShip==-1) && (MouseRightPressed))
	  { // and the right mouse button is pressed:
	    iDragShip=ShipNum;
	    iDragShipDir = 0; // Drag as a Down-ship
	  }
	  if ((iDragShip==-1) && (ShipVisibility[ShipNum]<255)) ShipVisibility[ShipNum]+=20;
	} else
	{ // The mouse cursor is not over this ship, make it half visible!
	  if (ShipVisibility[ShipNum]>145)
	  {
	    ShipVisibility[ShipNum]-=10;
	    if (ShipVisibility[ShipNum]<145) ShipVisibility[ShipNum]=145;
	  }
	}
	if (ShipVisibility[ShipNum]>255) ShipVisibility[ShipNum]=255;

	DrawSprite(pBuffer, Skin->BackgroundImage->ulWidth,
		   0,0,
		   Skin->BackgroundImage->ulWidth, Skin->BackgroundImage->ulHeight,
		   Skin->Ship_StartPositionX[ShipNum],
		   Skin->Ship_StartPositionY[ShipNum],
		   Skin->Ship_Down[ShipNum][0], ShipVisibility[ShipNum]);
      }
    } 
  }

  // Now show the ship that is being dragged!
  if (iDragShip>-1)
  {
    ShipNum=iDragShip;
    // Make it sure that it will be fully visible
    if (ShipVisibility[ShipNum]<255) ShipVisibility[ShipNum]+=20;
    if (ShipVisibility[ShipNum]>255) ShipVisibility[ShipNum]=255;

    // Get the sprite according to the orientation of this ship
    if (!iDragShipDir)
      psTemp=Skin->Ship_Down[ShipNum][0];
    else
      psTemp=Skin->Ship_Right[ShipNum][0];

    if ((TableCursorXPos==-1) && (TableCursorYPos==-1))
      // If the ship is not over the table, position it relative to the mouse!
      DrawSprite(pBuffer, Skin->BackgroundImage->ulWidth,
		 0, 0,
		 Skin->BackgroundImage->ulWidth, Skin->BackgroundImage->ulHeight,
		 MouseXPos-(Skin->Table.TableXOneStepX+Skin->Table.TableYOneStepX)/2,
		 MouseYPos-(Skin->Table.TableXOneStepY+Skin->Table.TableYOneStepY)/2,
		 psTemp, ShipVisibility[ShipNum]);
    else
    { // If the ship is over the table, use the table as a grid to position it!

      if (AcceptableShipPosition(TableCursorXPos, TableCursorYPos, ShipNum, iDragShipDir))
      { // The position is acceptable, so draw the cursors under the ship:
	if (!iDragShipDir)
	{ // Down-ship
	  for (i=0; i<Ship_Length[iDragShip]; i++)
	  {
	    DrawSprite(pBuffer, Skin->BackgroundImage->ulWidth,
		       0, 0,
		       Skin->BackgroundImage->ulWidth, Skin->BackgroundImage->ulHeight,
		       TableCursorXPos*Skin->Table.TableXOneStepX + (TableCursorYPos+i)*Skin->Table.TableYOneStepX +Skin->Table.TableTopLeftCornerX,
		       TableCursorXPos*Skin->Table.TableXOneStepY + (TableCursorYPos+i)*Skin->Table.TableYOneStepY +Skin->Table.TableTopLeftCornerY,
		       Skin->PositionCursor, 255);
	  }
	} else
	{ // Right-ship
	  for (i=0; i<Ship_Length[iDragShip]; i++)
	  {
	    DrawSprite(pBuffer, Skin->BackgroundImage->ulWidth,
		       0, 0,
		       Skin->BackgroundImage->ulWidth, Skin->BackgroundImage->ulHeight,
		       (TableCursorXPos+i)*Skin->Table.TableXOneStepX + TableCursorYPos*Skin->Table.TableYOneStepX +Skin->Table.TableTopLeftCornerX,
		       (TableCursorXPos+i)*Skin->Table.TableXOneStepY + TableCursorYPos*Skin->Table.TableYOneStepY +Skin->Table.TableTopLeftCornerY,
		       Skin->PositionCursor, 255);
	  }
	}
	// and then draw the ship itself!
	DrawSprite(pBuffer, Skin->BackgroundImage->ulWidth,
		   0, 0,
		   Skin->BackgroundImage->ulWidth, Skin->BackgroundImage->ulHeight,
		   TableCursorXPos*Skin->Table.TableXOneStepX + TableCursorYPos*Skin->Table.TableYOneStepX +Skin->Table.TableTopLeftCornerX,
		   TableCursorXPos*Skin->Table.TableXOneStepY + TableCursorYPos*Skin->Table.TableYOneStepY +Skin->Table.TableTopLeftCornerY,
		   psTemp, ShipVisibility[ShipNum]);
      } else
      { // If the ship is over the table, but doesn't fit to its actutal position
	DrawSprite(pBuffer, Skin->BackgroundImage->ulWidth,
		   0, 0,
		   Skin->BackgroundImage->ulWidth, Skin->BackgroundImage->ulHeight,
		   MouseXPos-(Skin->Table.TableXOneStepX+Skin->Table.TableYOneStepX)/2,
		   MouseYPos-(Skin->Table.TableXOneStepY+Skin->Table.TableYOneStepY)/2,
		   psTemp, ShipVisibility[ShipNum]);
      }
    }
  }

  // Now animate every ship!
  for (i=0; i<NUM_OF_SHIPS; i++)
  {
    int j;
    // Ok, ship is drawn, now animate!
    j = ShipNumOfHits[i];
    while ((!Skin->Ship_Down[i][j]) && (j>0)) j--;
    AnimateSprite(Skin->Ship_Down[i][j]);
    j = ShipNumOfHits[i];
    while ((!Skin->Ship_Down[i][j]) && (j>0)) j--;
    AnimateSprite(Skin->Ship_Right[i][j]);
  }
  // Now animate the cursor
  AnimateSprite(Skin->PositionCursor);
  // Animate the markers
  AnimateSprite(Skin->HitMarker);
  AnimateSprite(Skin->MissMarker);

  if (!bVolumePanelActive)
  {
    // Now set the state of the buttons!
    //  QuitButton:
    if (Skin->QuitButton->iState != BUTTONSTATE_DISABLED)
      UpdateActiveButtonState(Skin, Skin->QuitButton, &QuitButtonPressed);
    //  Setup button:
    if (Skin->SetupButton->iState != BUTTONSTATE_DISABLED)
      UpdateActiveButtonState(Skin, Skin->SetupButton, &SetupButtonPressed);
    //  Volume button:
    if (Skin->VolumeButton->iState != BUTTONSTATE_DISABLED)
      UpdateActiveButtonState(Skin, Skin->VolumeButton, &VolumeButtonPressed);

    //  StartButton:
    // Check if all the ships are on the table or not!
    for (i=0; i<NUM_OF_SHIPS; i++)
      if ((ShipPosition[i][0]==-1) || (ShipPosition[i][1]==-1)) break;
    
    if (i<NUM_OF_SHIPS)
    { // Not all the ships positioned yet, so the start button is disabled
      if (Skin->StartButton->iState !=BUTTONSTATE_DISABLED)
      {
	Skin->StartButton->iState = BUTTONSTATE_DISABLED;
	Skin->StartButton->sprDisabled->flActFrame = 0;
      }
    }
    else
    { // All the ships are positioned, activate start button!
      UpdateActiveButtonState(Skin, Skin->StartButton, &StartButtonPressed);
    }
  }

  // Now put there the buttons!
  DrawButton(pBuffer, Skin->QuitButton);
  DrawButton(pBuffer, Skin->StartButton);
  DrawButton(pBuffer, Skin->SetupButton);
  DrawButton(pBuffer, Skin->VolumeButton);

  if (!bVolumePanelActive)
  {
    // Now do something if the mouse buttons have been pressed:
    if ((MouseLeftPressed_Prev) && (!MouseLeftPressed))
    { // Released the Left mouse button!
      // It means Shi orientation change, if a ship is being dragged
      if (iDragShip>-1) iDragShipDir=1-iDragShipDir;
    }
    
    if ((MouseRightPressed_Prev) && (!MouseRightPressed))
    { // Released the right mouse button!
      // It means dropping a ship, if a ship was being dragged!
      if (iDragShip>-1)
      { // If a ship was dragged:
	if (AcceptableShipPosition(TableCursorXPos, TableCursorYPos, iDragShip, iDragShipDir))
	{
	  ShipPosition[iDragShip][0]=TableCursorXPos;
	  ShipPosition[iDragShip][1]=TableCursorYPos;
	  ShipDirection[iDragShip] = iDragShipDir;
	}
	iDragShip=-1;
      }
    }
    
    if ((!MouseRightPressed_Prev) && (MouseRightPressed) && (PointedShip!=-1))
    { // If right mouse button is just pressed over a ship on the table:
      // Start dragging from the table
      iDragShip=PointedShip;
      iDragShipDir = ShipDirection[PointedShip]; // Drag as it is now
      ShipPosition[PointedShip][0]=-1;
      ShipPosition[PointedShip][1]=-1;
    }
  }
}

//---------------------------------------------------------------------
// DrawRoundTimeScore
//
// Puts status information to screen (Round, Time and Score)
//---------------------------------------------------------------------
void DrawRoundTimeScore(unsigned char *pBuffer, pSkin Skin)
{
  char msgBuf[40];
  int FirstCharHeight;
  int FirstCharHotSpotX;
  pSprite FirstChar;

  if (!Skin) return;
  if (!Skin->Font2) return;

  // Draw Round stuff:
  sprintf(msgBuf, "%d", ulRound);
  FirstCharHeight = 0;
  FirstCharHotSpotX = 0;
  FirstChar = Skin->Font2->Character[msgBuf[0]];
  if (FirstChar)
  {
    FirstCharHeight = FirstChar->ulHotSpotY;
    FirstCharHotSpotX = FirstChar->ulHotSpotX;
  }

  DrawText(pBuffer, Skin->BackgroundImage->ulWidth,
	   Skin->RoundInfoAreaX1, Skin->RoundInfoAreaY1,
	   Skin->RoundInfoAreaX2, Skin->RoundInfoAreaY2,
	   Skin->RoundInfoAreaX1 + FirstCharHotSpotX,
	   Skin->RoundInfoAreaY1 + FirstCharHeight + (Skin->RoundInfoAreaY2 - Skin->RoundInfoAreaY1 - FirstCharHeight)/2,
	   msgBuf, Skin->Font2,
	   255);

  // Draw Score stuff:
  sprintf(msgBuf, "%d", ulScore);
  FirstCharHeight = 0;
  FirstCharHotSpotX = 0;
  FirstChar = Skin->Font2->Character[msgBuf[0]];
  if (FirstChar)
  {
    FirstCharHeight = FirstChar->ulHotSpotY;
    FirstCharHotSpotX = FirstChar->ulHotSpotX;
  }
  DrawText(pBuffer, Skin->BackgroundImage->ulWidth,
	   Skin->ScoreInfoAreaX1, Skin->ScoreInfoAreaY1,
	   Skin->ScoreInfoAreaX2, Skin->ScoreInfoAreaY2,
	   Skin->ScoreInfoAreaX1 + FirstCharHotSpotX,
	   Skin->ScoreInfoAreaY1 + FirstCharHeight + (Skin->ScoreInfoAreaY2 - Skin->ScoreInfoAreaY1 - FirstCharHeight)/2,
	   msgBuf, Skin->Font2,
           255);

  // Draw Time stuff:
  if (((ulTime%1000)<500) ||
      ((iGameState != GAMESTATE_SELF_ROUND) &&
       (iGameState != GAMESTATE_ENEMY_ROUND)))
    sprintf(msgBuf, "%02d:%02d", ulTime/60000, (ulTime%60000)/1000);
  else
    sprintf(msgBuf, "%02d %02d", ulTime/60000, (ulTime%60000)/1000);
  FirstCharHeight = 0;
  FirstCharHotSpotX = 0;
  FirstChar = Skin->Font2->Character[msgBuf[0]];
  if (FirstChar)
  {
    FirstCharHeight = FirstChar->ulHotSpotY;
    FirstCharHotSpotX = FirstChar->ulHotSpotX;
  } else
  {
    printf("No first char in time: '%s'!\n", msgBuf); fflush(stdout);
  }
  DrawText(pBuffer, Skin->BackgroundImage->ulWidth,
	   Skin->TimeInfoAreaX1, Skin->TimeInfoAreaY1,
	   Skin->TimeInfoAreaX2, Skin->TimeInfoAreaY2,
	   Skin->TimeInfoAreaX1 + FirstCharHotSpotX,
	   Skin->TimeInfoAreaY1 + FirstCharHeight + (Skin->TimeInfoAreaY2 - Skin->TimeInfoAreaY1 - FirstCharHeight)/2,
	   msgBuf, Skin->Font2,
           255);
}

//---------------------------------------------------------------------
// Draw_Both_FullTables
//
// Draws local and enemy tables to the screen, both the ships and the
// hit and miss markers.
//---------------------------------------------------------------------
void Draw_Both_FullTables(unsigned char *pBuffer, pSkin Skin)
{
  int x, y, i;
  int ShipDrawn[NUM_OF_SHIPS];
 
  // First draw the LocalTable, so our ships and markers of the shots of enemy

  // First draw the MISS/HIT markers on local table:
  if (Skin->Table1.TableOrientation==0)
  { // Draw from left to right
    for (y=0; y<10; y++)
      for (x=0; x<10; x++)
      {
        if (LocalTable[x][y] == TABLEENTRY_MISS_AUDIOONLY)
        { // It's needed to wait some for the animation.
          if (clock()-LocalTableMiscData[x][y]>=Skin->EnemyMissAnimationWaitTime)
          {
            LocalTable[x][y] = TABLEENTRY_MISS_AUDIOANDBLOWUP;
            LocalTableMiscData[x][y] = 0;
            // *Now* show the message of enemy shot!
            if ((Skin) && (Skin->msgEnemyShot))
              NewMessage_to_MessageBoard(Skin->msgEnemyShot);
          }
        }
	if (LocalTable[x][y] == TABLEENTRY_MISS_AUDIOANDBLOWUP)
	{ // Draw BlowUp animation
	  // Get the BlowUp sprite
	  pSprite BlowUpSprite = Skin->MissBlowUp;
	  if (!BlowUpSprite) LocalTable[x][y]=TABLEENTRY_MISS;
	  else
	  { // If the sprite is defined, set its animation frame number,
	    LocalTableMiscData[x][y]++;
	    BlowUpSprite->flActFrame=BlowUpSprite->flAnimSpeed * LocalTableMiscData[x][y];
	    BlowUpSprite->ulActFrame=BlowUpSprite->flActFrame;
	    if (BlowUpSprite->ulActFrame<BlowUpSprite->ulNumOfFrames)
	    { // and draw this frame, or
	      DrawSprite(pBuffer, Skin->BackgroundImage->ulWidth,
			 0, 0,
			 Skin->BackgroundImage->ulWidth, Skin->BackgroundImage->ulHeight,
			 x*Skin->Table1.TableXOneStepX + y*Skin->Table1.TableYOneStepX +Skin->Table1.TableTopLeftCornerX,
			 x*Skin->Table1.TableXOneStepY + y*Skin->Table1.TableYOneStepY +Skin->Table1.TableTopLeftCornerY,
			 BlowUpSprite, 255);
	    } else
	    { // if all frames drawn, draw miss marker, instead.
	      LocalTableMiscData[x][y]=0;
	      LocalTable[x][y]=TABLEENTRY_MISS;
	    }

	  }
	}
	if (LocalTable[x][y] == TABLEENTRY_MISS)
	{
	  DrawSprite(pBuffer, Skin->BackgroundImage->ulWidth,
		     0, 0,
		     Skin->BackgroundImage->ulWidth, Skin->BackgroundImage->ulHeight,
		     x*Skin->Table1.TableXOneStepX + y*Skin->Table1.TableYOneStepX +Skin->Table1.TableTopLeftCornerX,
		     x*Skin->Table1.TableXOneStepY + y*Skin->Table1.TableYOneStepY +Skin->Table1.TableTopLeftCornerY,
		     Skin->MissMarker, 255);
        }
        if (LocalTable[x][y] == TABLEENTRY_HIT_AUDIOONLY)
        { // It's needed to wait some for the animation.
          if (clock()-LocalTableMiscData[x][y]>=Skin->EnemyHitAnimationWaitTime)
          {
            LocalTable[x][y] = TABLEENTRY_HIT_AUDIOANDBLOWUP;
            LocalTableMiscData[x][y] = 0;
            // *Now* show the message of enemy shot!
            if ((Skin) && (Skin->msgEnemyShot))
              NewMessage_to_MessageBoard(Skin->msgEnemyShot);
          }
        }
	if (LocalTable[x][y] == TABLEENTRY_HIT_AUDIOANDBLOWUP)
        { // The blowup will be drawn by drawing a different ship sprite for this ship:
          ShipNumOfHits[ShipAtPosition(x,y)]++;
          LocalTableMiscData[x][y]=0;
          LocalTable[x][y]=TABLEENTRY_HIT;
	}
	if (LocalTable[x][y] == TABLEENTRY_HIT)
	{
	  DrawSprite(pBuffer, Skin->BackgroundImage->ulWidth,
		     0, 0,
		     Skin->BackgroundImage->ulWidth, Skin->BackgroundImage->ulHeight,
		     x*Skin->Table1.TableXOneStepX + y*Skin->Table1.TableYOneStepX +Skin->Table1.TableTopLeftCornerX,
		     x*Skin->Table1.TableXOneStepY + y*Skin->Table1.TableYOneStepY +Skin->Table1.TableTopLeftCornerY,
		     Skin->HitMarker, 255);
	};
      }
  } else
  { // Draw from right to left
    for (y=0; y<10; y++)
      for (x=9; x>=0; x--)
      {
        if (LocalTable[x][y] == TABLEENTRY_MISS_AUDIOONLY)
        { // It's needed to wait some for the animation.
          if (clock()-LocalTableMiscData[x][y]>=Skin->EnemyMissAnimationWaitTime)
          {
            LocalTable[x][y] = TABLEENTRY_MISS_AUDIOANDBLOWUP;
            LocalTableMiscData[x][y] = 0;
            // *Now* show the message of enemy shot!
            if ((Skin) && (Skin->msgEnemyShot))
              NewMessage_to_MessageBoard(Skin->msgEnemyShot);
          }
        }
	if (LocalTable[x][y] == TABLEENTRY_MISS_AUDIOANDBLOWUP)
	{ // Draw BlowUp animation
	  // Get the BlowUp sprite
	  pSprite BlowUpSprite = Skin->MissBlowUp;
	  if (!BlowUpSprite) LocalTable[x][y]=TABLEENTRY_MISS;
	  else
	  { // If the sprite is defined, set its animation frame number,
	    LocalTableMiscData[x][y]++;
	    BlowUpSprite->flActFrame=BlowUpSprite->flAnimSpeed * LocalTableMiscData[x][y];
	    BlowUpSprite->ulActFrame=BlowUpSprite->flActFrame;
	    if (BlowUpSprite->ulActFrame<BlowUpSprite->ulNumOfFrames)
	    { // and draw this frame, or
	      DrawSprite(pBuffer, Skin->BackgroundImage->ulWidth,
			 0, 0,
			 Skin->BackgroundImage->ulWidth, Skin->BackgroundImage->ulHeight,
			 x*Skin->Table1.TableXOneStepX + y*Skin->Table1.TableYOneStepX +Skin->Table1.TableTopLeftCornerX,
			 x*Skin->Table1.TableXOneStepY + y*Skin->Table1.TableYOneStepY +Skin->Table1.TableTopLeftCornerY,
			 BlowUpSprite, 255);
	    } else
	    { // if all frames drawn, draw miss marker, instead.
	      LocalTableMiscData[x][y]=0;
	      LocalTable[x][y]=TABLEENTRY_MISS;
            }

	  }
	}
	if (LocalTable[x][y] == TABLEENTRY_MISS)
	{
	  DrawSprite(pBuffer, Skin->BackgroundImage->ulWidth,
		     0, 0,
		     Skin->BackgroundImage->ulWidth, Skin->BackgroundImage->ulHeight,
		     x*Skin->Table1.TableXOneStepX + y*Skin->Table1.TableYOneStepX +Skin->Table1.TableTopLeftCornerX,
		     x*Skin->Table1.TableXOneStepY + y*Skin->Table1.TableYOneStepY +Skin->Table1.TableTopLeftCornerY,
		     Skin->MissMarker, 255);
        }

        if (LocalTable[x][y] == TABLEENTRY_HIT_AUDIOONLY)
        { // It's needed to wait some for the animation.
          if (clock()-LocalTableMiscData[x][y]>=Skin->EnemyHitAnimationWaitTime)
          {
            LocalTable[x][y] = TABLEENTRY_HIT;
            LocalTableMiscData[x][y] = 0;
            // *Now* show the message of enemy shot!
            if ((Skin) && (Skin->msgEnemyShot))
              NewMessage_to_MessageBoard(Skin->msgEnemyShot);
          }
        }
	if (LocalTable[x][y] == TABLEENTRY_HIT_AUDIOANDBLOWUP)
        { // The blowup will be drawn by drawing a different ship sprite for this ship:
          ShipNumOfHits[ShipAtPosition(x,y)]++;
          LocalTableMiscData[x][y]=0;
          LocalTable[x][y]=TABLEENTRY_HIT;
	}
	if (LocalTable[x][y] == TABLEENTRY_HIT)
	{
	  DrawSprite(pBuffer, Skin->BackgroundImage->ulWidth,
		     0, 0,
		     Skin->BackgroundImage->ulWidth, Skin->BackgroundImage->ulHeight,
		     x*Skin->Table1.TableXOneStepX + y*Skin->Table1.TableYOneStepX +Skin->Table1.TableTopLeftCornerX,
		     x*Skin->Table1.TableXOneStepY + y*Skin->Table1.TableYOneStepY +Skin->Table1.TableTopLeftCornerY,
		     Skin->HitMarker, 255);
	};
      }
  }

  // Now show the ships on the table:
  // (Should be done in order of visibility!)

  for (i=0; i<NUM_OF_SHIPS; i++)
  {
    // Make all the ships fully visible!
    if (ShipVisibility[i]<255) ShipVisibility[i]+=20;
    if (ShipVisibility[i]>255) ShipVisibility[i]=255;
    if ((ShipPosition[i][0]==-1) && (ShipPosition[i][1]==-1))
      ShipDrawn[i] = 1;
    else
      ShipDrawn[i] = 0;
  }
  // Now draw the ships!
  for (i=0; i<NUM_OF_SHIPS; i++)
    if (!ShipDrawn[i]) DrawShipAndNeightbours(pBuffer, i, ShipDrawn, 0, &(Skin->Table1));

  //
  // Now draw the Enemy Table, so our shots!
  //
 
  if (Skin->Table2.TableOrientation==0)
  { // Draw from left to right
    for (y=0; y<10; y++)
      for (x=0; x<10; x++)
      {
        if (EnemyTable[x][y] == TABLEENTRY_MISS_AUDIOONLY)
        { // It's needed to wait some for the animation.
          if (clock()-EnemyTableMiscData[x][y]>=Skin->WeMissAnimationWaitTime)
          {
            EnemyTable[x][y] = TABLEENTRY_MISS_AUDIOANDBLOWUP;
            EnemyTableMiscData[x][y] = 0;
            // *Now* show the message of enemy shot!
            if ((Skin) && (Skin->msgWeMissed))
              NewMessage_to_MessageBoard(Skin->msgWeMissed);
          }
        }
	if (EnemyTable[x][y] == TABLEENTRY_MISS_AUDIOANDBLOWUP)
	{ // Draw BlowUp animation
	  // Get the BlowUp sprite
	  pSprite BlowUpSprite = Skin->MissBlowUp;
	  if (!BlowUpSprite) EnemyTable[x][y]=TABLEENTRY_MISS;
	  else
	  { // If the sprite is defined, set its animation frame number,
	    EnemyTableMiscData[x][y]++;
	    BlowUpSprite->flActFrame=BlowUpSprite->flAnimSpeed * EnemyTableMiscData[x][y];
	    BlowUpSprite->ulActFrame=BlowUpSprite->flActFrame;
	    if (BlowUpSprite->ulActFrame<BlowUpSprite->ulNumOfFrames)
	    { // and draw this frame, or
	      DrawSprite(pBuffer, Skin->BackgroundImage->ulWidth,
			 0, 0,
			 Skin->BackgroundImage->ulWidth, Skin->BackgroundImage->ulHeight,
			 x*Skin->Table2.TableXOneStepX + y*Skin->Table2.TableYOneStepX +Skin->Table2.TableTopLeftCornerX,
			 x*Skin->Table2.TableXOneStepY + y*Skin->Table2.TableYOneStepY +Skin->Table2.TableTopLeftCornerY,
			 BlowUpSprite, 255);
	    } else
	    { // if all frames drawn, draw miss marker, instead.
	      EnemyTableMiscData[x][y]=0;
	      EnemyTable[x][y]=TABLEENTRY_MISS;
	    }

	  }
	}
	if (EnemyTable[x][y] == TABLEENTRY_MISS)
	{
	  DrawSprite(pBuffer, Skin->BackgroundImage->ulWidth,
		     0, 0,
		     Skin->BackgroundImage->ulWidth, Skin->BackgroundImage->ulHeight,
		     x*Skin->Table2.TableXOneStepX + y*Skin->Table2.TableYOneStepX +Skin->Table2.TableTopLeftCornerX,
		     x*Skin->Table2.TableXOneStepY + y*Skin->Table2.TableYOneStepY +Skin->Table2.TableTopLeftCornerY,
		     Skin->MissMarker, 255);
        }

        if (EnemyTable[x][y] == TABLEENTRY_HIT_AUDIOONLY)
        { // It's needed to wait some for the animation.
          if (clock()-EnemyTableMiscData[x][y]>=Skin->WeHitAnimationWaitTime)
          {
            EnemyTable[x][y] = TABLEENTRY_HIT_AUDIOANDBLOWUP;
            EnemyTableMiscData[x][y] = 0;
            // *Now* show the message of enemy shot!
            if ((Skin) && (Skin->msgWeHit))
              NewMessage_to_MessageBoard(Skin->msgWeHit);
          }
        }
	if (EnemyTable[x][y] == TABLEENTRY_HIT_AUDIOANDBLOWUP)
	{ // Draw BlowUp animation
	  // Get the BlowUp sprite
	  pSprite BlowUpSprite = Skin->HitBlowUp;
	  if (!BlowUpSprite) EnemyTable[x][y]=TABLEENTRY_HIT;
	  else
	  { // If the sprite is defined, set its animation frame number,
	    EnemyTableMiscData[x][y]++;
	    BlowUpSprite->flActFrame=BlowUpSprite->flAnimSpeed * EnemyTableMiscData[x][y];
	    BlowUpSprite->ulActFrame=BlowUpSprite->flActFrame;
	    if (BlowUpSprite->ulActFrame<BlowUpSprite->ulNumOfFrames)
	    { // and draw this frame, or
	      DrawSprite(pBuffer, Skin->BackgroundImage->ulWidth,
			 0, 0,
			 Skin->BackgroundImage->ulWidth, Skin->BackgroundImage->ulHeight,
			 x*Skin->Table2.TableXOneStepX + y*Skin->Table2.TableYOneStepX +Skin->Table2.TableTopLeftCornerX,
			 x*Skin->Table2.TableXOneStepY + y*Skin->Table2.TableYOneStepY +Skin->Table2.TableTopLeftCornerY,
			 BlowUpSprite, 255);
	    } else
	    { // if all frames drawn, draw miss marker, instead.
	      EnemyTableMiscData[x][y]=0;
	      EnemyTable[x][y]=TABLEENTRY_HIT;
	    }

	  }
	}
	if (EnemyTable[x][y] == TABLEENTRY_HIT)
	{
	  DrawSprite(pBuffer, Skin->BackgroundImage->ulWidth,
		     0, 0,
		     Skin->BackgroundImage->ulWidth, Skin->BackgroundImage->ulHeight,
		     x*Skin->Table2.TableXOneStepX + y*Skin->Table2.TableYOneStepX +Skin->Table2.TableTopLeftCornerX,
		     x*Skin->Table2.TableXOneStepY + y*Skin->Table2.TableYOneStepY +Skin->Table2.TableTopLeftCornerY,
		     Skin->HitMarker, 255);
	};
      }
  } else
  { // Draw from right to left
    for (y=0; y<10; y++)
      for (x=9; x>=0; x--)
      {
        if (EnemyTable[x][y] == TABLEENTRY_MISS_AUDIOONLY)
        { // It's needed to wait some for the animation.
          if (clock()-EnemyTableMiscData[x][y]>=Skin->WeMissAnimationWaitTime)
          {
            EnemyTable[x][y] = TABLEENTRY_MISS_AUDIOANDBLOWUP;
            EnemyTableMiscData[x][y] = 0;
            // *Now* show the message of enemy shot!
            if ((Skin) && (Skin->msgWeMissed))
              NewMessage_to_MessageBoard(Skin->msgWeMissed);
          }
        }
	if (EnemyTable[x][y] == TABLEENTRY_MISS_AUDIOANDBLOWUP)
	{ // Draw BlowUp animation
	  // Get the BlowUp sprite
	  pSprite BlowUpSprite = Skin->MissBlowUp;
	  if (!BlowUpSprite) EnemyTable[x][y]=TABLEENTRY_MISS;
	  else
	  { // If the sprite is defined, set its animation frame number,
	    EnemyTableMiscData[x][y]++;
	    BlowUpSprite->flActFrame=BlowUpSprite->flAnimSpeed * EnemyTableMiscData[x][y];
	    BlowUpSprite->ulActFrame=BlowUpSprite->flActFrame;
	    if (BlowUpSprite->ulActFrame<BlowUpSprite->ulNumOfFrames)
	    { // and draw this frame, or
	      DrawSprite(pBuffer, Skin->BackgroundImage->ulWidth,
			 0, 0,
			 Skin->BackgroundImage->ulWidth, Skin->BackgroundImage->ulHeight,
			 x*Skin->Table2.TableXOneStepX + y*Skin->Table2.TableYOneStepX +Skin->Table2.TableTopLeftCornerX,
			 x*Skin->Table2.TableXOneStepY + y*Skin->Table2.TableYOneStepY +Skin->Table2.TableTopLeftCornerY,
			 BlowUpSprite, 255);
	    } else
	    { // if all frames drawn, draw miss marker, instead.
	      EnemyTableMiscData[x][y]=0;
	      EnemyTable[x][y]=TABLEENTRY_MISS;
	    }

	  }
	}
	if (EnemyTable[x][y] == TABLEENTRY_MISS)
	{
	  DrawSprite(pBuffer, Skin->BackgroundImage->ulWidth,
		     0, 0,
		     Skin->BackgroundImage->ulWidth, Skin->BackgroundImage->ulHeight,
		     x*Skin->Table2.TableXOneStepX + y*Skin->Table2.TableYOneStepX +Skin->Table2.TableTopLeftCornerX,
		     x*Skin->Table2.TableXOneStepY + y*Skin->Table2.TableYOneStepY +Skin->Table2.TableTopLeftCornerY,
		     Skin->MissMarker, 255);
        }

	if (EnemyTable[x][y] == TABLEENTRY_HIT_AUDIOONLY)
        { // It's needed to wait some for the animation.
          if (clock()-EnemyTableMiscData[x][y]>=Skin->WeHitAnimationWaitTime)
          {
            EnemyTable[x][y] = TABLEENTRY_HIT_AUDIOANDBLOWUP;
            EnemyTableMiscData[x][y] = 0;
            // *Now* show the message of enemy shot!
            if ((Skin) && (Skin->msgWeHit))
              NewMessage_to_MessageBoard(Skin->msgWeHit);
          }
        }
        if (EnemyTable[x][y] == TABLEENTRY_HIT_AUDIOANDBLOWUP)
	{ // Draw BlowUp animation
	  // Get the BlowUp sprite
	  pSprite BlowUpSprite = Skin->HitBlowUp;
	  if (!BlowUpSprite) EnemyTable[x][y]=TABLEENTRY_HIT;
	  else
	  { // If the sprite is defined, set its animation frame number,
	    EnemyTableMiscData[x][y]++;
	    BlowUpSprite->flActFrame=BlowUpSprite->flAnimSpeed * EnemyTableMiscData[x][y];
	    BlowUpSprite->ulActFrame=BlowUpSprite->flActFrame;
	    if (BlowUpSprite->ulActFrame<BlowUpSprite->ulNumOfFrames)
	    { // and draw this frame, or
	      DrawSprite(pBuffer, Skin->BackgroundImage->ulWidth,
			 0, 0,
			 Skin->BackgroundImage->ulWidth, Skin->BackgroundImage->ulHeight,
			 x*Skin->Table2.TableXOneStepX + y*Skin->Table2.TableYOneStepX +Skin->Table2.TableTopLeftCornerX,
			 x*Skin->Table2.TableXOneStepY + y*Skin->Table2.TableYOneStepY +Skin->Table2.TableTopLeftCornerY,
			 BlowUpSprite, 255);
	    } else
	    { // if all frames drawn, draw miss marker, instead.
	      EnemyTableMiscData[x][y]=0;
	      EnemyTable[x][y]=TABLEENTRY_HIT;
	    }

	  }
	}
	if (EnemyTable[x][y] == TABLEENTRY_HIT)
	{
	  DrawSprite(pBuffer, Skin->BackgroundImage->ulWidth,
		     0, 0,
		     Skin->BackgroundImage->ulWidth, Skin->BackgroundImage->ulHeight,
		     x*Skin->Table2.TableXOneStepX + y*Skin->Table2.TableYOneStepX +Skin->Table2.TableTopLeftCornerX,
		     x*Skin->Table2.TableXOneStepY + y*Skin->Table2.TableYOneStepY +Skin->Table2.TableTopLeftCornerY,
		     Skin->HitMarker, 255);
	};
      }
  }
}

//---------------------------------------------------------------------
// RenderOneFrame_when_WaitForPosEnemy
//
// Renders one frame, when the gamestate is WAITFORPOSENEMY
//---------------------------------------------------------------------
void RenderOneFrame_when_WaitForPosEnemy(unsigned char *pBuffer)
{
  long xtemp, ytemp, TableCursorXPos, TableCursorYPos;
  int i, ShipNum;
  int iVisibility;
  int PointedShip;
  pSprite psTemp;

  // Drawing the background with the same technique used in
  // RenderOneFrame_PositionSelf

  // First of all, put there the background
  // Do it by removing previously drawn sprites:
  RemoveDrawnSprites(pBuffer, Skin->BackgroundImage->ulWidth);
  if (fFirstTimeRender) // if it's the first time to build the output image, we should set up the background
			// using memcpy. Otherwise, we should remove the sprites, because that's faster than
			// the memcpy of the whole background...
  {
    if (Skin->BackgroundImage2)
      memcpy(pBuffer, Skin->BackgroundImage2->pImageData, Skin->BackgroundImage2->ulImageSize);
    fFirstTimeRender = FALSE;
  }

  // Now it's time to draw something!
  // Put some text there!
  MessageBoard_Processing(pBuffer, Skin);

  // Draw the Score/Rounds/Time triumph
  DrawRoundTimeScore(pBuffer, Skin);

  // Draw the special backgrounds on table1, like shot but missed,
  // and shot and hit markers, plus table2 too.
  Draw_Both_FullTables(pBuffer, Skin);

  // Now animate every sprite!
  for (i=0; i<NUM_OF_SHIPS; i++)
  {
    int j;
    // Ok, ship is drawn, now animate!
    j = ShipNumOfHits[i];
    while ((!Skin->Ship_Down[i][j]) && (j>0)) j--;
    AnimateSprite(Skin->Ship_Down[i][j]);
    j = ShipNumOfHits[i];
    while ((!Skin->Ship_Down[i][j]) && (j>0)) j--;
    AnimateSprite(Skin->Ship_Right[i][j]);

  }
  // Now animate the cursor
  AnimateSprite(Skin->PositionCursor);
  // Animate the markers
  AnimateSprite(Skin->HitMarker);
  AnimateSprite(Skin->MissMarker);

  if (!bVolumePanelActive)
  {
    // Now set the state of the buttons!
    //  QuitButton:
    if (Skin->QuitButton->iState != BUTTONSTATE_DISABLED)
      UpdateActiveButtonState(Skin, Skin->QuitButton, &QuitButtonPressed);
    //  Setup button:
    if (Skin->SetupButton->iState != BUTTONSTATE_DISABLED)
      UpdateActiveButtonState(Skin, Skin->SetupButton, &SetupButtonPressed);
    //  Volume button:
    if (Skin->VolumeButton->iState != BUTTONSTATE_DISABLED)
      UpdateActiveButtonState(Skin, Skin->VolumeButton, &VolumeButtonPressed);
  }

  // Now put there the buttons!
  DrawButton(pBuffer, Skin->QuitButton);
  DrawButton(pBuffer, Skin->SetupButton);
  DrawButton(pBuffer, Skin->VolumeButton);
}

//---------------------------------------------------------------------
// WeShootAt
//
// Called when the user shot at a given position. It sends the message
// about shooting to the enemy, and updates round counter.
//---------------------------------------------------------------------
void WeShootAt(int TableXPos, int TableYPos)
{
  char Message[128];

  // First send the message about shooting to the enemy
  sprintf(Message, "SHOOT %d %d", TableXPos, TableYPos);
  (*pSendMessage)(Socket, Message);

  // And change gamestate:
  iGameState = GAMESTATE_ENEMY_ROUND;
  if (!fWeAreTheServer)
    ulRound++; // Next round then!
  fFirstTimeRender = TRUE;  // Make sure the new background will be drawn!
}

//---------------------------------------------------------------------
// EnemyShootsAt
//
// Called when the enemy shot at a given position. It responds to that
// message by sending back hit or miss message, and updates the local
// table to reflect the effect of the shot.
//---------------------------------------------------------------------
void EnemyShootsAt(int TableXPos, int TableYPos)
{
  char Message[128];

  if (ShipAtPosition(TableXPos, TableYPos)>-1)
  { // If we had a ship at that position, it hits!
    sprintf(Message, "HIT %d %d", TableXPos, TableYPos);
    (*pSendMessage)(Socket, Message);

    if (LocalTable[TableXPos][TableYPos] != TABLEENTRY_HIT)
    { // Restart hit animation only if there was no shooting at this position yet!
      LocalTable[TableXPos][TableYPos] = TABLEENTRY_HIT_AUDIOONLY;
      // save actual time, so Audio synchro waiting will know how much to wait!
      LocalTableMiscData[TableXPos][TableYPos] = clock();
    }
    SFXPlayFX(Skin->EnemyHitSFX);
  } else
  { // There is no ship there, hehe, it missed!
    sprintf(Message, "MISS %d %d", TableXPos, TableYPos);
    (*pSendMessage)(Socket, Message);
    LocalTable[TableXPos][TableYPos] = TABLEENTRY_MISS_AUDIOONLY;
    // save actual time, so Audio syncho waiting will know how much to wait!
    LocalTableMiscData[TableXPos][TableYPos] = clock();
    SFXPlayFX(Skin->EnemyMissSFX);
  }
  // The enemy has shot, so it's our turn again!
  iGameState=GAMESTATE_SELF_ROUND;
  if (fWeAreTheServer)
    ulRound++; // Next round then!
  fFirstTimeRender = TRUE;  // Make sure the new background will be drawn!
}

//---------------------------------------------------------------------
// RenderOneFrame_when_Rounds
//
// Renders one frame, when the gamestate is SELFROUND or ENEMYROUND
//---------------------------------------------------------------------
void RenderOneFrame_when_Rounds(unsigned char *pBuffer)
{
  long xtemp, ytemp, TableCursorXPos, TableCursorYPos;
  int i, ShipNum;
  int iVisibility;
  int PointedShip;
  pSprite psTemp;

  // Drawing the background with the same technique used in
  // RenderOneFrame_PositionSelf

  // First of all, put there the background
  // Do it by removing previously drawn sprites:
  RemoveDrawnSprites(pBuffer, Skin->BackgroundImage->ulWidth);
  if (fFirstTimeRender) // if it's the first time to build the output image, we should set up the background
			// using memcpy. Otherwise, we should remove the sprites, because that's faster than
			// the memcpy of the whole background...
  {
    if (Skin->BackgroundImage2)
      memcpy(pBuffer, Skin->BackgroundImage2->pImageData, Skin->BackgroundImage2->ulImageSize);
    fFirstTimeRender = FALSE;
  }

  // Now it's time to draw something!
  // Put some text there!
  MessageBoard_Processing(pBuffer, Skin);

  // Draw the Score/Rounds/Time triumph
  ulTime = ulCurrTimeStamp - ulInitTimeStamp;
  DrawRoundTimeScore(pBuffer, Skin);

  // Calculate the position of cursor if over the table:
  TableCursorXPos = TableCursorYPos = -1;
  ImgPos2TablePos(Skin->Table2.TableTopLeftCornerX, Skin->Table2.TableTopLeftCornerY,
		  Skin->Table2.TableXOneStepX, Skin->Table2.TableXOneStepY,
		  Skin->Table2.TableYOneStepX, Skin->Table2.TableYOneStepY,
		  MouseXPos, MouseYPos,
		  &xtemp, &ytemp);
  if ((xtemp>=0) && (ytemp>=0) && (xtemp<=9) && (ytemp<=9))
  {
    TableCursorXPos = xtemp;
    TableCursorYPos = ytemp;
  }

  // Draw the special backgrounds on table1, like shot but missed,
  // and shot and hit markers! Then the ships there, and then table 2!
  Draw_Both_FullTables(pBuffer, Skin);

  // Now show the cursor, if it's over the enemytable!
  if ((TableCursorXPos!=-1) && (TableCursorYPos!=-1) && (iGameState == GAMESTATE_SELF_ROUND))
  {
    DrawSprite(pBuffer, Skin->BackgroundImage->ulWidth,
	       0, 0,
	       Skin->BackgroundImage->ulWidth, Skin->BackgroundImage->ulHeight,
	       TableCursorXPos*Skin->Table2.TableXOneStepX + TableCursorYPos*Skin->Table2.TableYOneStepX +Skin->Table2.TableTopLeftCornerX,
	       TableCursorXPos*Skin->Table2.TableXOneStepY + TableCursorYPos*Skin->Table2.TableYOneStepY +Skin->Table2.TableTopLeftCornerY,
	       Skin->PositionCursor, 255);

    if ((!bVolumePanelActive) && (!MouseLeftPressed) && (MouseLeftPressed_Prev))
    { // Released the button over this table position! It means shooting!
      if (iGameState == GAMESTATE_SELF_ROUND)
        WeShootAt(TableCursorXPos, TableCursorYPos);
    }
  }

  // Now animate every sprite!
  for (i=0; i<NUM_OF_SHIPS; i++)
  {
    int j;
    // Ok, ship is drawn, now animate!
    j = ShipNumOfHits[i];
    while ((!Skin->Ship_Down[i][j]) && (j>0)) j--;
    AnimateSprite(Skin->Ship_Down[i][j]);
    j = ShipNumOfHits[i];
    while ((!Skin->Ship_Down[i][j]) && (j>0)) j--;
    AnimateSprite(Skin->Ship_Right[i][j]);

  }
  // Now animate the cursor
  AnimateSprite(Skin->PositionCursor);
  // Animate the markers
  AnimateSprite(Skin->HitMarker);
  AnimateSprite(Skin->MissMarker);

  if (!bVolumePanelActive)
  {
    // Now set the state of the buttons!
    //  QuitButton:
    if (Skin->QuitButton->iState != BUTTONSTATE_DISABLED)
      UpdateActiveButtonState(Skin, Skin->QuitButton, &QuitButtonPressed);
    //  Setup button:
    if (Skin->SetupButton->iState != BUTTONSTATE_DISABLED)
      UpdateActiveButtonState(Skin, Skin->SetupButton, &SetupButtonPressed);
    //  Volume button:
    if (Skin->VolumeButton->iState != BUTTONSTATE_DISABLED)
      UpdateActiveButtonState(Skin, Skin->VolumeButton, &VolumeButtonPressed);
  }

  // Now put there the buttons!
  DrawButton(pBuffer, Skin->QuitButton);
  DrawButton(pBuffer, Skin->SetupButton);
  DrawButton(pBuffer, Skin->VolumeButton);

  // Check for game over!
  CheckForGameOver();
}

//---------------------------------------------------------------------
// VolPanelCloseBtnPressed
//
// Callback for volume panel, when the close button is presssed
//---------------------------------------------------------------------
void VolPanelCloseBtnPressed()
{
#ifdef DEBUG_BUILD
  printf("VolPanelCloseBtnPressed()\n");
  fflush(stdout);
#endif
  bVolumePanelActive = FALSE;
}

//---------------------------------------------------------------------
// VolPanelMainArmPressed
//
// Callback for volume panel
//---------------------------------------------------------------------
void VolPanelMainArmPressed()
{
#ifdef DEBUG_BUILD
  printf("VolPanelMainArmPressed()\n");
  fflush(stdout);
#endif
  iVolumePanelState = VOLUMEPANELSTATE_DRAGMAIN;
}

//---------------------------------------------------------------------
// VolPanelMusicArmPressed
//
// Callback for volume panel
//---------------------------------------------------------------------
void VolPanelMusicArmPressed()
{
#ifdef DEBUG_BUILD
  printf("VolPanelMusicArmPressed()\n");
  fflush(stdout);
#endif
  iVolumePanelState = VOLUMEPANELSTATE_DRAGMUSIC;
}

//---------------------------------------------------------------------
// VolPanelSFXArmPressed
//
// Callback for volume panel
//---------------------------------------------------------------------
void VolPanelSFXArmPressed()
{
#ifdef DEBUG_BUILD
  printf("VolPanelSFXArmPressed()\n");
  fflush(stdout);
#endif
  iVolumePanelState = VOLUMEPANELSTATE_DRAGSFX;
}

//---------------------------------------------------------------------
// UpdateActiveArmState
//
// Updates the state of a given button/arm, according to the mouse
// position and status of mouse buttons. Calls a given function, if the
// button/arm is pressed. (does not have to be released too!)
//---------------------------------------------------------------------
void UpdateActiveArmState(pSkin Skin,
			  pButton Button,
			  void ButtonPressedFn())
{
  if ((!Button) || (!ButtonPressedFn)) return;

  if (MouseIsOverButton(Button))
  { // The mouse is over this button,
    if (Button->iState==BUTTONSTATE_PRESSED)
    { // and the button is already pressed,
      if (MouseLeftPressed)
        // and the mouse button is pressed too:
	Button->iState=BUTTONSTATE_PRESSED;
      else
      {
	// The button was pressed, and the mouse button has been released here.
	// We do nothing here, as this is an Slider Arm.
        Button->iState = BUTTONSTATE_MOUSEOVER;
      }
    } else
    { // If the button is not pressed,
      if ((MouseLeftPressed) && (!MouseLeftPressed_Prev))
      { // but just right now pressed the mouse button over this button:
	Button->iState=BUTTONSTATE_PRESSED;
	// Restart animation too!
	Button->sprPressed->flActFrame = 0;
        // Make sound
	SFXPlayFX(Skin->PressButtonSFX);
        // Call the function that will do the job
        (*ButtonPressedFn)();
      }
      else
      { // If the button is not pressed, and the mouse button is not pressed too:
	if (Button->iState != BUTTONSTATE_MOUSEOVER)
        { 
          SFXPlayFX(Skin->OverButtonSFX);
	  Button->iState=BUTTONSTATE_MOUSEOVER;
	  // Restart animation too!
	  Button->sprMouseOver->flActFrame = 0;
	}
      }
    }
  }
  else
  { // If the mouse is not over this button:
    if (Button->iState != BUTTONSTATE_ACTIVE)
    { // but it was just before, let's leave it:
      SFXPlayFX(Skin->LeaveButtonSFX);
      Button->iState=BUTTONSTATE_ACTIVE;
      // Restart animation too!
      Button->sprActive->flActFrame = 0;
    }
  }
}

//---------------------------------------------------------------------
// RenderVolumePanel
//
// Renders the volume setting panel
//---------------------------------------------------------------------
void RenderVolumePanel(unsigned char *pBuffer)
{
  if ((!Skin) || (!pBuffer)) return;

  // Put there and animate the background:
  DrawSprite(pBuffer, Skin->BackgroundImage->ulWidth,
	     0, 0,
	     Skin->BackgroundImage->ulWidth, Skin->BackgroundImage->ulHeight,
             Skin->VolumePanelX, Skin->VolumePanelY,
	     Skin->VolumePanel, 255);
  AnimateSprite(Skin->VolumePanel);

  // Put there and animate the buttons:
  DrawButton(pBuffer, Skin->MainVolumeArm);
  DrawButton(pBuffer, Skin->MusicVolumeArm);
  DrawButton(pBuffer, Skin->SFXVolumeArm);
  DrawButton(pBuffer, Skin->VolumePanelCloseBtn);

  // Now change the state!
  switch (iVolumePanelState)
  {
    case VOLUMEPANELSTATE_NORMAL:

      UpdateActiveArmState(Skin, Skin->MainVolumeArm, VolPanelMainArmPressed);
      UpdateActiveArmState(Skin, Skin->MusicVolumeArm, VolPanelMusicArmPressed);
      UpdateActiveArmState(Skin, Skin->SFXVolumeArm, VolPanelSFXArmPressed);
      UpdateActiveButtonState(Skin, Skin->VolumePanelCloseBtn, VolPanelCloseBtnPressed);
      break;
    case VOLUMEPANELSTATE_DRAGMAIN:
      if (MouseLeftPressed)
      {
	int x = MouseXPos;
	int xstart, xend;
	if ((Skin) && (Skin->VolumePanel))
	{
	  xstart = Skin->VolumePanelX - Skin->VolumePanel->ulHotSpotX +
	    Skin->MainVolumeLineX;
          xend   = Skin->VolumePanelX - Skin->VolumePanel->ulHotSpotX +
	    Skin->MainVolumeLineX + Skin->MainVolumeLineLength;
	  if (x<xstart) x = xstart;
	  if (x>xend) x = xend;
	  SetMainVolume((x - xstart)*SFX_VOLUME__MAX/Skin->MainVolumeLineLength);
#ifdef DEBUG_BUILD
	  printf("New main volume: %d\n",(x - xstart)*SFX_VOLUME__MAX/Skin->MainVolumeLineLength); fflush(stdout);
#endif
	  Skin->MainVolumeArm->XPos = x;
	}
      }
      else
        iVolumePanelState = VOLUMEPANELSTATE_NORMAL;
      break;
    case VOLUMEPANELSTATE_DRAGMUSIC:
      if (MouseLeftPressed)
      {
	int x = MouseXPos;
	int xstart, xend;
	if ((Skin) && (Skin->VolumePanel))
	{
	  xstart = Skin->VolumePanelX - Skin->VolumePanel->ulHotSpotX +
	    Skin->MusicVolumeLineX;
          xend   = Skin->VolumePanelX - Skin->VolumePanel->ulHotSpotX +
	    Skin->MusicVolumeLineX + Skin->MusicVolumeLineLength;
	  if (x<xstart) x = xstart;
	  if (x>xend) x = xend;
	  SetMusicVolume((x - xstart)*SFX_VOLUME__MAX/Skin->MusicVolumeLineLength);
#ifdef DEBUG_BUILD
	  printf("New music volume: %d\n",(x - xstart)*SFX_VOLUME__MAX/Skin->MusicVolumeLineLength); fflush(stdout);
#endif
	  Skin->MusicVolumeArm->XPos = x;
	}
      }
      else
        iVolumePanelState = VOLUMEPANELSTATE_NORMAL;
      break;
    case VOLUMEPANELSTATE_DRAGSFX:
      if (MouseLeftPressed)
      {
	int x = MouseXPos;
	int xstart, xend;
	if ((Skin) && (Skin->VolumePanel))
	{
	  xstart = Skin->VolumePanelX - Skin->VolumePanel->ulHotSpotX +
	    Skin->SFXVolumeLineX;
          xend   = Skin->VolumePanelX - Skin->VolumePanel->ulHotSpotX +
	    Skin->SFXVolumeLineX + Skin->SFXVolumeLineLength;
	  if (x<xstart) x = xstart;
	  if (x>xend) x = xend;
	  SetSFXVolume((x - xstart)*SFX_VOLUME__MAX/Skin->SFXVolumeLineLength);
#ifdef DEBUG_BUILD
	  printf("New SFX volume: %d\n",(x - xstart)*SFX_VOLUME__MAX/Skin->SFXVolumeLineLength); fflush(stdout);
#endif
	  Skin->SFXVolumeArm->XPos = x;
	}
      }
      else
        iVolumePanelState = VOLUMEPANELSTATE_NORMAL;
      break;
    default:
      bVolumePanelActive = FALSE;
      break;
  }
}


//---------------------------------------------------------------------
// ChangeSkin
//
// Called when a game configuration has been changed on the fly.
//---------------------------------------------------------------------
void ChangeSkin()
{
  // The game configuration values has been changed on the fly!
  // It means that the skin has to be reloaded, and the sound
  // has to be reinitialized:
  if (DosRequestMutexSem( hmtxUseDive, SEM_INDEFINITE_WAIT ) == NO_ERROR )
  {
    pConfig pOldConfig=(pConfig) malloc(sizeof(tConfig));
    if (pOldConfig)
    {
  
  #ifdef DEBUG_BUILD
      printf("Changing configuration on the fly!\n");
      fflush(stdout);
  #endif
  
      memcpy(pOldConfig, &CurrentConfig, sizeof(tConfig));
  
      // The new skin could be loaded, so quickly change the old skin to the
      // new one, and reinitialize sound!
      // We need a semaphore to be sure that other threads will not use the
      // current skin and the sound effects, while we are replacing it with the new one!
      if (DosRequestMutexSem( hmtxUseSkin, SEM_INDEFINITE_WAIT ) == NO_ERROR )
      {
	CleanupSkin(&Skin);
	SFXCloseDevice();
  
  #ifdef DEBUG_BUILD
	printf("Setting up new skin...\n");
	fflush(stdout);
  #endif
	memcpy(&CurrentConfig, &NewConfig, sizeof(tConfig));
  
	SFXOpenDevice(CurrentConfig.SFXDeviceID,
		      CurrentConfig.SFXDeviceShared,
		      CurrentConfig.SFXFreq,
		      CurrentConfig.SFXBits,
		      CurrentConfig.SFXChannels,
		      CurrentConfig.SFXNumBufs,
		      CurrentConfig.SFXBufSize);
	// Now load the skin!
	Skin = LoadSkin(CurrentConfig.SkinFilename);
  
	if (!Skin)
	{
	  // Could not load new skin, so go back to the old one!
  #ifdef DEBUG_BUILD
	  printf("Error with new skin, going back to the old one!\n");
	  fflush(stdout);
  #endif
  
	  SFXCloseDevice();
	  memcpy(&CurrentConfig, pOldConfig, sizeof(tConfig));
	  SFXOpenDevice(CurrentConfig.SFXDeviceID,
			CurrentConfig.SFXDeviceShared,
			CurrentConfig.SFXFreq,
			CurrentConfig.SFXBits,
			CurrentConfig.SFXChannels,
			CurrentConfig.SFXNumBufs,
			CurrentConfig.SFXBufSize);
	  // Now load the skin!
	  Skin = LoadSkin(CurrentConfig.SkinFilename);
	  if (!Skin)
	  {
	    // Could not reload the old skin, so close the whole stuff!
	    WinPostMsg(hFrame, WM_QUIT, 0, 0);
	  }
	}
	// Set volumes
	SetMainVolume(CurrentConfig.iMainVolume);
	SetMusicVolume(CurrentConfig.iMusicVolume);
	SetSFXVolume(CurrentConfig.iSFXVolume);
	// Reconfigure DIVE...
  #ifdef DEBUG_BUILD
	printf("Reconfiguring DIVE...\n");
	fflush(stdout);
  #endif
	ShutdownDIVE();
	if (Skin)
	{
	  srcBufferWidth = Skin->BackgroundImage->ulWidth;
	  srcBufferHeight = Skin->BackgroundImage->ulHeight;
	}
	SetupDIVE();
	// Now that the inner screen is bigger/smaller, resize the frame and client to fit size!
	WinSetWindowPos( hFrame,
			 HWND_TOP,
			 0, 0,
			 srcBufferWidth+WinQuerySysValue( HWND_DESKTOP, SV_CXDLGFRAME)*2,
			 srcBufferHeight + WinQuerySysValue( HWND_DESKTOP, SV_CYTITLEBAR)
			 + WinQuerySysValue( HWND_DESKTOP, SV_CYDLGFRAME)*2
			 + 2,
			 SWP_SIZE);
	WinSetWindowPos( hClient,
			 HWND_TOP,
			 0, 0,
			 srcBufferWidth,
			 srcBufferHeight,
			 SWP_SIZE);
	// Make sure the buffers will be fully redrawn, as they are empty now!
	fFirstTimeRender = TRUE;
	if (Skin)
	{
	  MessagePos = Skin->BackgroundImage->ulHeight;
	  ActualMessage = Skin->msgConfigChanged; MessageBoard_NewMessage=TRUE;
	}
	// Start the background music!
	if (Skin)
	  SFXPlayFX(Skin->BackgroundMusic);

	DosReleaseMutexSem( hmtxUseSkin );
      }
      free(pOldConfig);
    }
#ifdef DEBUG_BUILD
    printf("Done skin changing\n");
    fflush(stdout);
#endif
    DosReleaseMutexSem(hmtxUseDive);
  }
#ifdef DEBUG_BUILD
  else
  {
    printf("Could not take hmtxUseDive semaphore!\n");
    fflush(stdout);
  }
#endif

}

//---------------------------------------------------------------------
// RenderOneFrame
//
// Called when a frame should be built. It requests usage for DIVE, 
// grabs the DIVE buffer where the image should be assembled, and
// assembles the new image/frame according to the current gamestate.
// After this, sends the newly rendered image to DIVE to blit it to
// the screen.
//---------------------------------------------------------------------
void RenderOneFrame()
{
  unsigned char *pBuffer;
  ULONG ulScanLineBytes;
  ULONG ulScanLines;

  if ((!DiveImageHandle) |     // Don't blit image, if there is no buffer, or
      (fVRNDisabled)) return;  // the blitting is disabled!

  if (DosRequestMutexSem( hmtxUseDive, SEM_INDEFINITE_WAIT ) == NO_ERROR )
  {
    // Tell DIVE that we want to use its buffer, and ask for a pointer to
    // access it, ask for the number of bytes in a scanline, and the number
    // of scanlines in that buffer!
    DiveBeginImageBufferAccess( hDiveInst, DiveImageHandle, &pBuffer, &ulScanLineBytes, &ulScanLines);

    if (DosRequestMutexSem( hmtxUseSkin, SEM_INDEFINITE_WAIT ) == NO_ERROR )
    {
      // Render the scene, if the skin is real!
      if (Skin)
      {
        // Render the main game stuff
	switch (iGameState)
	{
	  case GAMESTATE_POSITION_SELF:
	  case GAMESTATE_POSITION_SELF_ENEMY_POSITIONED:
	    RenderOneFrame_when_PositioningSelf(pBuffer);
	    break;
	  case GAMESTATE_WAIT_FOR_POSITION_ENEMY:
	    RenderOneFrame_when_WaitForPosEnemy(pBuffer);
	    break;
	  case GAMESTATE_SELF_ROUND:
	  case GAMESTATE_ENEMY_ROUND:
	    RenderOneFrame_when_Rounds(pBuffer);
	    break;
	  case GAMESTATE_END_OF_GAME_WON:
	  case GAMESTATE_END_OF_GAME_LOST:
	    RenderOneFrame_when_WaitForPosEnemy(pBuffer); // The same things to do here...
	    break;
	  default:
	    RenderOneFrame_when_WaitForPosEnemy(pBuffer);
	}

	// Then put there the Volume settings panel, if it's active:
	if (bVolumePanelActive)
          RenderVolumePanel(pBuffer);
      }
  
      // Save this mouse button state as previous for next frame
      MouseRightPressed_Prev = MouseRightPressed;
      MouseLeftPressed_Prev = MouseLeftPressed;

      DosReleaseMutexSem(hmtxUseSkin);
    }

    // Tell DIVE that the access of the image is ended
    DiveEndImageBufferAccess( hDiveInst, DiveImageHandle );

    // and blit it to the screen
    DiveBlitImage( hDiveInst, DiveImageHandle, DIVE_BUFFER_SCREEN );

    // Count the number of blitted images, so we can count FPS!
    ulNumBlitted++;

    // That's all for now, thanks DIVE!
    DosReleaseMutexSem( hmtxUseDive );
  }

  // Ok, now that the image has been drawn, let's check if it's needed
  // to change skin!
  // To change skin/configuration, the ChangeSkin() function is called
  // but *from the main thread*!
  // This must be done, because the SFXOpenDevice() loaded some DLLs
  // (e.g. OGGIDEC.DLL), which must be unloaded from the same thread!
  if (bNewConfigAvailable)
  {
    bNewConfigAvailable = FALSE;
    WinPostMsg(hFrame, WM_CHANGECONFIG, 0, 0);
  }
}

//---------------------------------------------------------------------
// FrameRenderer
//
// This is a thread, which calls RenderOneFrame() when a new frame
// should be displayed. It gets its tick from a FrameRegulation thread,
// which posts an event semaphore at given time intervals.
// When the semaphore is posted, it means that one frame should be 
// displayed, so this thread creates and displays one.
//---------------------------------------------------------------------
VOID FrameRenderer(PVOID Param)
{
  ULONG ulPostCount;

#ifdef DEBUG_BUILD
  printf("Starting FrameRenderer thread\n");
  fflush(stdout);
#endif

  // While the program is running:
  while (!fShuttingDown)
  {
    // Wait for a blit event
    if (DosWaitEventSem( hevBlitNeededEvent, SEM_INDEFINITE_WAIT )==NO_ERROR)
    {
      DosResetEventSem(hevBlitNeededEvent, &ulPostCount);
      // And blit one!
      RenderOneFrame();
    }
    else
    DosSleep(0); // Don't eat all CPU cycles, if there is an error with
                 // DosWaitEventSem, but give up them.
  }

#ifdef DEBUG_BUILD
  printf("Halting FrameRenderer thread\n");
  fflush(stdout);
#endif

  _endthread();
}

//---------------------------------------------------------------------
// SearchForFOURCC
//
// It takes an array of FOURCCs, and checks if a given FOURCC is
// contained or not.
//---------------------------------------------------------------------
FOURCC SearchForFOURCC(FOURCC *where, int space, FOURCC what)
{
  int i;

  // Search in the array!
  for (i=0; i<space; i++)
    if (where[i]==what) break;

  if (i<space)
    return what; // We found it!

  return -1; // Could not find!
}

//---------------------------------------------------------------------
// InspectDIVE
//
// Queries the capabilities of the DIVE subsystem, if it exists, and
// checks if the required BGR3 input format is supported or not.
// Practically, it returns TRUE if DIVE is installed and supports our
// FOURCC, and FALSE otherwise.
//---------------------------------------------------------------------
int InspectDIVE()
{
  PDIVE_CAPS pDiveCaps;
  FOURCC *pfccFormats;
  APIRET rc;
  int i;

#ifdef DEBUG_BUILD
  printf("InspectDIVE()\n");
  fflush(stdout);
#endif

  // Allocate memory for the buffers used
  pDiveCaps = (PDIVE_CAPS) malloc(sizeof(DIVE_CAPS));
  if (!pDiveCaps) return FALSE;
  pfccFormats = (FOURCC *)malloc(sizeof(FOURCC)*100);
  if (!pfccFormats)
  {
    free(pDiveCaps);
    return FALSE;
  }
  // Fill some vital information
  pDiveCaps->ulStructLen = sizeof(DIVE_CAPS);
  pDiveCaps->ulFormatLength = sizeof(FOURCC)*100;
  pDiveCaps->pFormatData = pfccFormats;
  // Query DIVE capabilities
  rc=DiveQueryCaps(pDiveCaps, DIVE_BUFFER_SCREEN);
  if (rc)
  { // No DIVE on this computer???
#ifdef DEBUG_BUILD
    printf("  DiveQueryCaps(), rc=%d\n", rc);
    fflush(stdout);
#endif

    free(pfccFormats);
    free(pDiveCaps);
    return FALSE;
  }

#ifdef DEBUG_BUILD
  {
    char *fcc = (char *) &(pDiveCaps->fccColorEncoding);

    printf(" DIVE Information:\n");
    printf("   Current screen mode: %d x %d, %d BPP, colour encoding %c%c%c%c\n",
	   pDiveCaps->ulHorizontalResolution,
	   pDiveCaps->ulVerticalResolution,
	   pDiveCaps->ulDepth,
	   *fcc,
	   *(fcc+1),
	   *(fcc+2),
	   *(fcc+3));
    printf("   Number of input formats : %d\n",pDiveCaps->ulInputFormats);
    printf("   Number of output formats: %d\n",pDiveCaps->ulOutputFormats);
    printf("   Input formats:  ");

    fcc=(char *)pfccFormats;
    for (i=0; i<pDiveCaps->ulInputFormats; i++)
    {
      printf(" %c%c%c%c",
	     *(fcc),
	     *(fcc+1),
	     *(fcc+2),
	     *(fcc+3));
      fcc+=4;
    }
    printf("\n");
    printf("   Output formats: ");
    for (i=0; i<pDiveCaps->ulOutputFormats; i++)
    {
      printf(" %c%c%c%c",
	     *(fcc),
	     *(fcc+1),
	     *(fcc+2),
	     *(fcc+3));
      fcc+=4;
    }
    printf("\n");
    fflush(stdout);
  }

  printf(" Looking for BGR3 input format...\n"); fflush(stdout);
#endif

  // We will use BGR3 for now. Later it could be implemented to find
  // a usable format that exists, and indicate the selected one in
  // srcBufferFOURCC, so other parts of the code can take care of it.
  srcBufferFOURCC = mmioFOURCC('B','G','R','3');
  // Check if the BGR3 is supported or not!

  for (i=0; i<pDiveCaps->ulInputFormats; i++)
    if (pfccFormats[i]==srcBufferFOURCC) break;

  if (i>=pDiveCaps->ulInputFormats) srcBufferFOURCC = -1; // Indicate that there is no good format found!
  free(pfccFormats);
  free(pDiveCaps)
    ;
#ifdef DEBUG_BUILD
  if (srcBufferFOURCC==-1)
    printf(" Requested input format not supported, cannot continue!\n");
  else
    printf(" Requested input format supported, all right!\n");
  fflush(stdout);
#endif

  return (srcBufferFOURCC!=-1);
}

//                     ------- Skin Cleanup code ------

//---------------------------------------------------------------------
// CleanupButton
//
// Destroys a button object, and frees all memory allocated for it.
//---------------------------------------------------------------------
void CleanupButton(pButton *Button)
{
  // Cleanup all the sprites in the button!
  if ((Button) && (*Button))
  {
    if ((*Button)->sprDisabled)
    {
      CleanupSprite((*Button)->sprDisabled);
      (*Button)->sprDisabled=NULL;
    }
    if ((*Button)->sprActive)
    {
      CleanupSprite((*Button)->sprActive);
      (*Button)->sprActive=NULL;
    }
    if ((*Button)->sprMouseOver)
    {
      CleanupSprite((*Button)->sprMouseOver);
      (*Button)->sprMouseOver=NULL;
    }
    if ((*Button)->sprPressed)
    {
      CleanupSprite((*Button)->sprPressed);
      (*Button)->sprPressed=NULL;
    }
    free(*Button);
    *Button=NULL;
  }
}

//---------------------------------------------------------------------
// CleanupSkin
//
// Destroys a skin object, and frees all memory allocated for it.
//---------------------------------------------------------------------
void CleanupSkin(pSkin *Skin)
{
  int i,j;

#ifdef DEBUG_BUILD
  printf("CleanupSkin()\n");
  if (!Skin) printf("  Skin is NULL!\n");
  fflush(stdout);
#endif

  if (!Skin) return;
  if (!(*Skin)) return;
  // Cleanup images:
  if ((*Skin)->BackgroundImage) CleanupImage((*Skin)->BackgroundImage);
  if ((*Skin)->BackgroundImage2) CleanupImage((*Skin)->BackgroundImage2);
  if ((*Skin)->ImageCache) CleanupImage((*Skin)->ImageCache);
  // Cleanup sprites:
  //  Position Cursor
  if ((*Skin)->PositionCursor) CleanupSprite((*Skin)->PositionCursor);
  //  Markers
  if ((*Skin)->HitMarker) CleanupSprite((*Skin)->HitMarker);
  if ((*Skin)->MissMarker) CleanupSprite((*Skin)->MissMarker);
  //  Blow-up animations
  if ((*Skin)->HitBlowUp) CleanupSprite((*Skin)->HitBlowUp);
  if ((*Skin)->MissBlowUp) CleanupSprite((*Skin)->MissBlowUp);

  //  Ships (Down, Right)
  for (i=0; i<NUM_OF_SHIPS; i++)
    for (j=0; j<5; j++)
    {
  
      if ((*Skin)->Ship_Down[i][j]) CleanupSprite((*Skin)->Ship_Down[i][j]);
      if ((*Skin)->Ship_Right[i][j]) CleanupSprite((*Skin)->Ship_Right[i][j]);
    }
  // Cleanup fonts:
  CleanupFont((*Skin)->Font1); (*Skin)->Font1=NULL;
  CleanupFont((*Skin)->Font2); (*Skin)->Font2=NULL;
  // Cleanup Buttons:
  //  QuitButton:
  CleanupButton(&((*Skin)->QuitButton));
  //  StartButton:
  CleanupButton(&((*Skin)->StartButton));

  // Volume setting panel
  CleanupButton(&((*Skin)->VolumePanelCloseBtn));
  CleanupButton(&((*Skin)->MainVolumeArm));
  CleanupButton(&((*Skin)->MusicVolumeArm));
  CleanupButton(&((*Skin)->SFXVolumeArm));
  if ((*Skin)->VolumePanel) CleanupSprite((*Skin)->VolumePanel);

  // Cleanup messages
  if ((*Skin)->msgWelcome)
    free((*Skin)->msgWelcome);
  if ((*Skin)->msgWaitingForEnemy)
    free((*Skin)->msgWaitingForEnemy);
  if ((*Skin)->msgEnemyDisconnected)
    free((*Skin)->msgEnemyDisconnected);
  if ((*Skin)->msgEnemyPositioned)
    free((*Skin)->msgEnemyPositioned);
  if ((*Skin)->msgEnemyPositionedTooSrv)
    free((*Skin)->msgEnemyPositionedTooSrv);
  if ((*Skin)->msgEnemyPositionedTooCli)
    free((*Skin)->msgEnemyPositionedTooCli);
  if ((*Skin)->msgWePositionedTooSrv)
    free((*Skin)->msgWePositionedTooSrv);
  if ((*Skin)->msgWePositionedTooCli)
    free((*Skin)->msgWePositionedTooCli);
  if ((*Skin)->msgEnemyShot)
    free((*Skin)->msgEnemyShot);
  if ((*Skin)->msgWeHit)
    free((*Skin)->msgWeHit);
  if ((*Skin)->msgWeMissed)
    free((*Skin)->msgWeMissed);
  if ((*Skin)->msgGameWon)
    free((*Skin)->msgGameWon);
  if ((*Skin)->msgGameLost)
    free((*Skin)->msgGameLost);
  if ((*Skin)->msgConfigChanged)
    free((*Skin)->msgConfigChanged);


  // Cleanup sound effects
  if ((*Skin)->BackgroundMusic)
    SFXUnloadFX((*Skin)->BackgroundMusic);
  if ((*Skin)->WeHitSFX)
    SFXUnloadFX((*Skin)->WeHitSFX);
  if ((*Skin)->WeMissSFX)
    SFXUnloadFX((*Skin)->WeMissSFX);
  if ((*Skin)->EnemyHitSFX)
    SFXUnloadFX((*Skin)->EnemyHitSFX);
  if ((*Skin)->EnemyMissSFX)
    SFXUnloadFX((*Skin)->EnemyMissSFX);
  if ((*Skin)->OverButtonSFX)
    SFXUnloadFX((*Skin)->OverButtonSFX);
  if ((*Skin)->LeaveButtonSFX)
    SFXUnloadFX((*Skin)->LeaveButtonSFX);
  if ((*Skin)->PressButtonSFX)
    SFXUnloadFX((*Skin)->PressButtonSFX);
  if ((*Skin)->GameStartSFX)
    SFXUnloadFX((*Skin)->GameStartSFX);
  if ((*Skin)->GameWonSFX)
    SFXUnloadFX((*Skin)->GameWonSFX);
  if ((*Skin)->GameLostSFX)
    SFXUnloadFX((*Skin)->GameLostSFX);

  // That's all!
  free(*Skin); *Skin=NULL;
}

//                         -----  Skin loading  -----

//---------------------------------------------------------------------
// Process_SpriteDefinition_SkinEntry
//
// Loads a sprite based on the spritedef skin entry parameters.
//---------------------------------------------------------------------
void Process_SpriteDefinition_SkinEntry(pSprite *pTheSprite, char *Param, char *path, pImage *pImageCache)
{
  char *pcxfilename;
  char *ptemp1, *ptemp2;
  pImage PCX;


  if ((!Param) || (!path) || (!pTheSprite) || (!pImageCache)) return;

  // First get the filename where to take the sprite from
  pcxfilename = (char *)malloc(strlen(path)+2048);
  if (!pcxfilename) return;
  strcpy(pcxfilename, path);
  ptemp1=pcxfilename+strlen(pcxfilename); // Points to end of pcxfilename (to #0)
  ptemp2=Param;                           // Points to filename in parameter
  // Copy the filename to the end of path
  while ((*ptemp2!='\0') && (*ptemp2!=' ') && (*ptemp2!='\t'))
    *(ptemp1++)=*(ptemp2++);
  *(ptemp1++)='\0';
  // Cool, we have the filename with full skin-path!
  // Try to load the PCX file

  // But don't reload it, if it was the last image to process
  // (so it's in the image cache)
  if ((!(*pImageCache)) ||
      (stricmp((*pImageCache)->sFileName, pcxfilename)!=0))
  { // Not in cache, try to load it!

    if (*pImageCache)
      CleanupImage(*pImageCache);
    *pImageCache = LoadPCX(pcxfilename);
    if (!(*pImageCache))
    {
      free(pcxfilename);
      return;
    }
  }

  free(pcxfilename);
  PCX = *pImageCache; // The image is in the cache!

  // Ok, pcx file in memory!
  // Now get other parameters from Param, then
  // extract the ColorData and ColorMask from PCX image!

  // Find the first real parameter positon in param string:
  ptemp1=Param;
  while ((*ptemp1!='\0') && (*ptemp1!=' ') && (*ptemp1!='\t')) ptemp1++;
  while ((*ptemp1==' ') || (*ptemp1=='\t')) ptemp1++;

  if (*ptemp1!='\0')
  {
    if (*pTheSprite)
      CleanupSprite(*pTheSprite);
    *pTheSprite = LoadSprite(PCX, ptemp1);
  }
}

//---------------------------------------------------------------------
// Process_Position_SkinEntry
//
// Gets an X and Y position from given string.
//---------------------------------------------------------------------
void Process_Position_SkinEntry(long *X, long *Y, char *Params)
{
  sscanf(Params, "%d %d", X, Y);
}

//---------------------------------------------------------------------
// Process_FontDef_SkinEntry
//
// Loads a sprite for a given letter in a given font, using the 
// string (param) describing it.
//---------------------------------------------------------------------
void Process_FontDef_SkinEntry(pFont *pTheFont, char *Param, char *path, pImage *pImageCache)
{
  char *pcxfilename;
  char *ptemp1, *ptemp2;
  pImage PCX;

  if ((!Param) || (!path) || (!pTheFont)) return;

  // Get the PCX filename, which contains the fonts!

  pcxfilename = (char *)malloc(strlen(path)+2048);
  if (!pcxfilename) return;
  strcpy(pcxfilename, path);
  ptemp1=pcxfilename+strlen(pcxfilename); // Points to end of pcxfilename (to #0)
  ptemp2=Param;                           // Points to filename in parameter
  // Copy the filename to the end of path
  while ((*ptemp2!='\0') && (*ptemp2!=' ') && (*ptemp2!='\t'))
    *(ptemp1++)=*(ptemp2++);
  *(ptemp1++)='\0';
  
  // Cool, we have the filename!
  // Try to load the PCX file

  if ((!(*pImageCache)) || (stricmp((*pImageCache)->sFileName, pcxfilename)!=0))
  { // Not in cache, try to load it!
    if (*pImageCache)
      CleanupImage(*pImageCache);
    *pImageCache = LoadPCX(pcxfilename);
    if (!(*pImageCache))
    {
      free(pcxfilename);
      return;
    }
  }
  free(pcxfilename);
  PCX = *pImageCache; // The image is in the cache!

  // Ok, pcx file in memory!
  // Now get other parameters from Param, then
  // extract the ColorData and ColorMask from PCX image!

  // Find the first real parameter positon in param string:
  ptemp1=Param;
  while ((*ptemp1!='\0') && (*ptemp1!=' ') && (*ptemp1!='\t')) ptemp1++;
  while ((*ptemp1==' ') || (*ptemp1=='\t')) ptemp1++;

  // And call in to the Fonts module to load a letter of a font!
  if (*ptemp1!='\0')
    LoadFontDef(pTheFont, PCX, ptemp1);
}

//---------------------------------------------------------------------
// Process_Message_SkinEntry
//
// Loads a text message from the parameter, and stores it.
//---------------------------------------------------------------------
void Process_Message_SkinEntry(char **msg, char *value)
{
  char *temp;

  if ((!value) || (!msg)) return;

  // If there is already a message stored in destination pointer,
  // free it first!
  if (*msg) free(*msg);

  // Now allocate memory for the new string
  (*msg)=(char *)malloc(strlen(value)+1);

  // Copy the message to that place, while replacing "\\n"'s in the
  // string with "\n"'s:
  temp=(*msg);
  while (*value)
  {
    if (*value!='\\')
      (*(temp++)) = (*(value++));
    else
    {
      value++;
      if (*value=='\\') *(temp++)='\\';     // "\\\\" -> "\\"
      else if (*value=='n') *(temp++)='\n'; // "\\n" -> "\n"
      else *(temp++)=*value;                // "\\<else>" -> "\\<else>"
      value++;
    }
  }
  *temp='\0';
}

//---------------------------------------------------------------------
// Process_SFX_SkinEntry
//
// Loads a sound effect using the give parameters.
//---------------------------------------------------------------------
void Process_SFX_SkinEntry(PSFX_FX *SFX, char *path, char *value)
{
  char *p1 = NULL;
  char *p2 = NULL;
  char *p3 = NULL;
  int rc;

  if ((!value) || (!SFX)) return;

  if (*SFX)
  { // Unload SFX, if there was one already loaded here!
    SFXUnloadFX(*SFX);
    *SFX = NULL;
  }

  // First allocate memory for our temporary variables
  p1 = (char *)malloc(strlen(value) + 1);
  p2 = (char *)malloc(strlen(value) + 1);
  p3 = (char *)malloc(strlen(value) + strlen(path) + 1);
  if ((!p1) || (!p2) || (!p3))
  {
    // Not enough memory to get the parameters!
    if (p1) free(p1);
    if (p2) free(p2);
    if (p3) free(p3);
    return;
  }

  // Grab parameters from param. string:
  //   <SFX filename> <Looped|Single> <Disk|Memory>

  rc = sscanf(value, "%s %s %s", p1, p2, p3);
  if (rc == 3)
  { // Ok, we have every parameter!
    if (!stricmp(p3, "MEMORY"))
    { // This file should be loaded as a memory stream
      strcpy(p3, path);   // Make full filename,
      strcat(p3, p1);     // with path!
#ifdef DEBUG_BUILD_SFXLOADMESSAGES
      printf("Loading SFX in MEMORY mode: %s\n", p3); fflush(stdout);
#endif
      *SFX = SFXLoadFX( p3, SFX_LOAD_MODE__MEMORY);
    }
    else
    { // This file shoule be loaded as a normal file stream
      strcpy(p3, path);   // Make full filename,
      strcat(p3, p1);     // with path!
#ifdef DEBUG_BUILD_SFXLOADMESSAGES
      printf("Loading SFX in STREAM mode: %s\n", p3); fflush(stdout);
#endif
      *SFX = SFXLoadFX( p3, SFX_LOAD_MODE__STREAM);
    }
    // If the file is loaded successfully, we should set the
    // play mode (looped or not!)!
    if (*SFX)
    {
      if (!stricmp(p2, "LOOPED"))
        SFXSetFXMode(*SFX, SFX_PLAY_MODE__LOOP);
      else
        SFXSetFXMode(*SFX, SFX_PLAY_MODE__SINGLE);
#ifdef DEBUG_BUILD_SFXLOADMESSAGES
      printf("SFX Loaded successfully, mode = %d, volume = %d\n", SFXGetFXMode(*SFX), SFXGetFXVolume(*SFX)); fflush(stdout);
#endif
    }
  }
  // Cleanup!
  free(p1);
  free(p2);
  free(p3);
}

//---------------------------------------------------------------------
// substr
//
// Extracts a substring from another string.
//---------------------------------------------------------------------
char *substr(char *Dst, char *Src, int Start, int Len)
{
  int i;

  for (i=Start; i<Start+Len; i++) Dst[i-Start]=Src[i];
  Dst[Len]=0;
  return Dst;
}

//---------------------------------------------------------------------
// LoadSkin
//
// Given a skin filename, it opens that file in the skin's base
// directory, processes the skin entries, loads everything from there,
// and returns a pointer to a newly allocated skin description
// structure.
// Returns NULL, if something goes very wrong.
//---------------------------------------------------------------------
pSkin LoadSkin(char *skin_filename)
{
  pSkin result=NULL;
  FILE *skinfile;
  char *key = (char *) malloc(1024);
  char *value = (char *) malloc(4096);
  char *stemp = (char *) malloc(1024);
  int rc, iTemp, iTemp2;
  char *skin_fullpath=(char *)malloc(strlen(Skin_BaseDir)+strlen(skin_filename)+2+1024);
  int LineCount=0;

#ifdef DEBUG_BUILD
  if (skin_filename)
    printf("LoadSkin('%s')\n", skin_filename);
  else
    printf("LoadSkin(NULL)\n");
  fflush(stdout);
#endif

  if ((!skin_fullpath) || (!key) || (!value) || (!stemp))
  {
    free(skin_fullpath);
    free(key);
    free(value);
    free(stemp);
    return result; // Not enough memory!
  }

  // Create full path from skin filename!
  sprintf(skin_fullpath, "%s\\%s.BSS", Skin_BaseDir, skin_filename);

  // Open this text file!
  if ((skinfile=fopen(skin_fullpath, "rt"))!=NULL)
  { // If we could open the file:
    // Let the skin_fullpath point to the direcory where the data files should be
    // for this skin! (so skin_fullpath will not contain the skin filename
    // anymore!)
    skin_fullpath[strlen(Skin_BaseDir)+1+strlen(skin_filename)]='\\';
    skin_fullpath[strlen(Skin_BaseDir)+1+strlen(skin_filename)+1]='\0';

    // Ok, create new skin structure, and fill it!
    result=(pSkin)malloc(sizeof(tSkin));
    if (result)
    { // Memory successfully allocated for new skin, initialize it with zeros
      memset(result, 0, sizeof(tSkin));

      // Now process every lines of the skin definition file, but
      // max. 1000, so if something goes wrong, wo won't end up an infinite
      // loop here.
      while ((!feof(skinfile)) && (LineCount<1000))
      {
	LineCount++;
	// First find the beginning of the key
        fscanf(skinfile, "%*[\n\t= ]");
        // Now get the key 
	rc=fscanf(skinfile, "%[^\n\t= ]",key);
	if ((rc==1) && (key[0]!=';'))
	{ // We have the key! (and not a remark!)
	  // Skip the whitespace after the key, if there are any
	  fscanf(skinfile, "%*[ \t]");
	  // There must be a '=' here
	  rc=fscanf(skinfile, "%[=]", value);
	  if (rc==1)
	  {
	    // Ok, the '=' is there
	    // Skip the whitespace, if there are any
	    fscanf(skinfile, "%*[ \t]");
	    // and read anything until the end of line
	    rc=fscanf(skinfile, "%[^\n]\n", value);
	    if (rc==1)
	    {
/*
#ifdef DEBUG_BUILD
	      printf("%s  =  %s\n", key, value); fflush(stdout);
#endif
*/
              // Cool, there is a key/value pair here, let's process it!
	      if (!stricmp(key, "BACKGROUNDIMAGE"))
	      { // Here is the name of the background image to use
	        // Assemble the full path for this filename in skin_fullpath:
		strcat(skin_fullpath, value);
		// Clean the old background image definition
                if (result->BackgroundImage) CleanupImage(result->BackgroundImage);
                // Load the new image
		result->BackgroundImage=LoadPCX(skin_fullpath);
		// Restore the skin_fullpath variable
                skin_fullpath[strlen(Skin_BaseDir)+1+strlen(skin_filename)]='\\';
		skin_fullpath[strlen(Skin_BaseDir)+1+strlen(skin_filename)+1]='\0';
	      } else
	      if (!stricmp(key, "BACKGROUNDIMAGE2"))
	      { // Here is the name of the background image 2 to use
	        // We do exactly the same as above.
		strcat(skin_fullpath, value);
                if (result->BackgroundImage2) CleanupImage(result->BackgroundImage2);
		result->BackgroundImage2=LoadPCX(skin_fullpath);
                skin_fullpath[strlen(Skin_BaseDir)+1+strlen(skin_filename)]='\\';
		skin_fullpath[strlen(Skin_BaseDir)+1+strlen(skin_filename)+1]='\0';
	      } else
	      if (!stricmp(key, "POSITIONCURSOR"))
              { // Here comes the description of the position cursor!
                Process_SpriteDefinition_SkinEntry(&(result->PositionCursor), value, skin_fullpath, &(result->ImageCache));
	      } else
	      if (!stricmp(key, "HITMARKER"))
	      { // Here comes the description of the hit marker!
                Process_SpriteDefinition_SkinEntry(&(result->HitMarker), value, skin_fullpath, &(result->ImageCache));
	      } else
	      if (!stricmp(key, "MISSMARKER"))
	      { // Here comes the description of the miss marker!
                Process_SpriteDefinition_SkinEntry(&(result->MissMarker), value, skin_fullpath, &(result->ImageCache));
	      } else
              if (!stricmp(key, "HITBLOWUP"))
              { // Here comes the description of the hit animation!
                Process_SpriteDefinition_SkinEntry(&(result->HitBlowUp), value, skin_fullpath, &(result->ImageCache));
              } else
              if (!stricmp(key, "MISSBLOWUP"))
              { // Here comes the description of the miss animation!
                Process_SpriteDefinition_SkinEntry(&(result->MissBlowUp), value, skin_fullpath, &(result->ImageCache));
              } else
	      if (!stricmp(key, "TABLETOPLEFTCORNER"))
	      {
		Process_Position_SkinEntry(&(result->Table.TableTopLeftCornerX),
					   &(result->Table.TableTopLeftCornerY),
					   value);
	      } else
	      if (!stricmp(key, "TABLEXONESTEP"))
	      {
		Process_Position_SkinEntry(&(result->Table.TableXOneStepX),
					   &(result->Table.TableXOneStepY),
					   value);
	      } else
	      if (!stricmp(key, "TABLEYONESTEP"))
	      {
		Process_Position_SkinEntry(&(result->Table.TableYOneStepX),
					   &(result->Table.TableYOneStepY),
					   value);
	      } else
	      if (!stricmp(key, "TABLEORIENTATION"))
	      {
                sscanf(value, "%d", &(result->Table.TableOrientation));
	      } else
	      if (!stricmp(key, "TABLE1TOPLEFTCORNER"))
	      {
		Process_Position_SkinEntry(&(result->Table1.TableTopLeftCornerX),
					   &(result->Table1.TableTopLeftCornerY),
					   value);
	      } else
	      if (!stricmp(key, "TABLE2TOPLEFTCORNER"))
	      {
		Process_Position_SkinEntry(&(result->Table2.TableTopLeftCornerX),
					   &(result->Table2.TableTopLeftCornerY),
					   value);
	      } else
	      if (!stricmp(key, "TABLE1XONESTEP"))
	      {
		Process_Position_SkinEntry(&(result->Table1.TableXOneStepX),
					   &(result->Table1.TableXOneStepY),
					   value);
	      } else
	      if (!stricmp(key, "TABLE1YONESTEP"))
	      {
		Process_Position_SkinEntry(&(result->Table1.TableYOneStepX),
					   &(result->Table1.TableYOneStepY),
					   value);
	      } else
	      if (!stricmp(key, "TABLE1ORIENTATION"))
	      {
                sscanf(value, "%d", &(result->Table1.TableOrientation));
	      } else
	      if (!stricmp(key, "TABLE2XONESTEP"))
	      {
		Process_Position_SkinEntry(&(result->Table2.TableXOneStepX),
					   &(result->Table2.TableXOneStepY),
					   value);
	      } else
	      if (!stricmp(key, "TABLE2YONESTEP"))
	      {
		Process_Position_SkinEntry(&(result->Table2.TableYOneStepX),
					   &(result->Table2.TableYOneStepY),
					   value);
	      } else
	      if (!stricmp(key, "TABLE2ORIENTATION"))
	      {
                sscanf(value, "%d", &(result->Table2.TableOrientation));
	      } else
       	      if ((!stricmp(substr(stemp, key, 0, 4),"SHIP")) &&
		  (!stricmp(substr(stemp, key, 5, 6),"_DOWN_")) &&
                  (!stricmp(substr(stemp, key, 12, 5),"_HITS")))
	      { // SHIP_x_DOWN_y_HITS
		iTemp = atoi(substr(stemp, key, 4, 1))-1;   // x
		iTemp2 = atoi(substr(stemp, key, 11, 1));   // y
		if ((iTemp>=0) && (iTemp<=4) &&
		    (iTemp2>=0) && (iTemp2<=Ship_Length[iTemp]))
  		  Process_SpriteDefinition_SkinEntry(&(result->Ship_Down[iTemp][iTemp2]), value, skin_fullpath, &(result->ImageCache));
	      } else
       	      if ((!stricmp(substr(stemp, key, 0, 4),"SHIP")) &&
		  (!stricmp(substr(stemp, key, 5, 7),"_RIGHT_")) &&
                  (!stricmp(substr(stemp, key, 13, 5),"_HITS")))
	      { // SHIP_x_RIGHT_y_HITS
		iTemp = atoi(substr(stemp, key, 4, 1))-1;   // x
		iTemp2 = atoi(substr(stemp, key, 12, 1));   // y
		if ((iTemp>=0) && (iTemp<=4) &&
		    (iTemp2>=0) && (iTemp2<=Ship_Length[iTemp]))
  		  Process_SpriteDefinition_SkinEntry(&(result->Ship_Right[iTemp][iTemp2]), value, skin_fullpath, &(result->ImageCache));
	      } else
	      if (!stricmp(key, "SHIP1_STARTPOSITION"))
	      {
		Process_Position_SkinEntry(&(result->Ship_StartPositionX[0]),
					   &(result->Ship_StartPositionY[0]),
					   value);
	      } else
	      if (!stricmp(key, "SHIP2_STARTPOSITION"))
	      {
		Process_Position_SkinEntry(&(result->Ship_StartPositionX[1]),
					   &(result->Ship_StartPositionY[1]),
					   value);
	      } else
	      if (!stricmp(key, "SHIP3_STARTPOSITION"))
	      {
		Process_Position_SkinEntry(&(result->Ship_StartPositionX[2]),
					   &(result->Ship_StartPositionY[2]),
					   value);
	      } else
	      if (!stricmp(key, "SHIP4_STARTPOSITION"))
	      {
		Process_Position_SkinEntry(&(result->Ship_StartPositionX[3]),
					   &(result->Ship_StartPositionY[3]),
					   value);
	      } else
	      if (!stricmp(key, "SHIP5_STARTPOSITION"))
	      {
		Process_Position_SkinEntry(&(result->Ship_StartPositionX[4]),
					   &(result->Ship_StartPositionY[4]),
					   value);
	      } else
	      if (!stricmp(key, "SHIP_STARTPOSITION_DRAWORDER"))
	      {
		sscanf(value, "%d %d %d %d %d",
		       &(result->Ship_StartPosition_DrawOrder[0]),
		       &(result->Ship_StartPosition_DrawOrder[1]),
		       &(result->Ship_StartPosition_DrawOrder[2]),
		       &(result->Ship_StartPosition_DrawOrder[3]),
                       &(result->Ship_StartPosition_DrawOrder[4]));
	      } else
	      if (!stricmp(key, "FONT1DEF"))
              {
                Process_FontDef_SkinEntry(&(result->Font1), value, skin_fullpath, &(result->ImageCache));
	      } else
	      if (!stricmp(key, "FONT1HEIGHT"))
              {
		if (sscanf(value, "%d", &iTemp)==1)
		  SetFontHeight(&(result->Font1), iTemp);
	      } else
	      if (!stricmp(key, "FONT2DEF"))
              {
                Process_FontDef_SkinEntry(&(result->Font2), value, skin_fullpath, &(result->ImageCache));
	      } else
	      if (!stricmp(key, "FONT2HEIGHT"))
              {
		if (sscanf(value, "%d", &iTemp)==1)
		  SetFontHeight(&(result->Font2), iTemp);
	      } else
	      if (!stricmp(key, "TEXTAREA"))
	      {
		if (sscanf(value, "%d %d %d %d",
			   &(result->TextAreaX1),
			   &(result->TextAreaY1),
                           &(result->TextAreaX2),
                           &(result->TextAreaY2))!=4)
		{
		  result->TextAreaX1 = 0;
                  result->TextAreaY1 = 0;
		  result->TextAreaX2 = 0;
		  result->TextAreaY2 = 0;
		}
	      } else
	      if (!stricmp(key, "ROUNDINFOAREA"))
	      {
		if (sscanf(value, "%d %d %d %d",
			   &(result->RoundInfoAreaX1),
			   &(result->RoundInfoAreaY1),
                           &(result->RoundInfoAreaX2),
                           &(result->RoundInfoAreaY2))!=4)
		{
		  result->RoundInfoAreaX1 = 0;
                  result->RoundInfoAreaY1 = 0;
		  result->RoundInfoAreaX2 = 0;
		  result->RoundInfoAreaY2 = 0;
		}
	      } else
	      if (!stricmp(key, "SCOREINFOAREA"))
	      {
		if (sscanf(value, "%d %d %d %d",
			   &(result->ScoreInfoAreaX1),
			   &(result->ScoreInfoAreaY1),
                           &(result->ScoreInfoAreaX2),
                           &(result->ScoreInfoAreaY2))!=4)
		{
		  result->ScoreInfoAreaX1 = 0;
                  result->ScoreInfoAreaY1 = 0;
		  result->ScoreInfoAreaX2 = 0;
		  result->ScoreInfoAreaY2 = 0;
		}
	      } else
	      if (!stricmp(key, "TIMEINFOAREA"))
	      {
		if (sscanf(value, "%d %d %d %d",
			   &(result->TimeInfoAreaX1),
			   &(result->TimeInfoAreaY1),
                           &(result->TimeInfoAreaX2),
                           &(result->TimeInfoAreaY2))!=4)
		{
		  result->TimeInfoAreaX1 = 0;
                  result->TimeInfoAreaY1 = 0;
		  result->TimeInfoAreaX2 = 0;
		  result->TimeInfoAreaY2 = 0;
		}
	      } else
	      if (!stricmp(key, "QUITBUTTON_ACTIVE"))
              {
                if (!(result->QuitButton))
                { // Allocate memory for the button, if not allocated yet!
                  result->QuitButton = (pButton) malloc(sizeof(tButton));
		  if (result->QuitButton)
		  {
		    memset(result->QuitButton, 0, sizeof(tButton));
                    result->QuitButton->iState = BUTTONSTATE_ACTIVE;
		  }
                }
                if (result->QuitButton)
                  Process_SpriteDefinition_SkinEntry(&(result->QuitButton->sprActive), value, skin_fullpath, &(result->ImageCache));
	      } else
	      if (!stricmp(key, "QUITBUTTON_MOUSEOVER"))
              {
                if (!(result->QuitButton))
                { // Allocate memory for the button, if not allocated yet!
                  result->QuitButton = (pButton) malloc(sizeof(tButton));
		  if (result->QuitButton)
		  {
		    memset(result->QuitButton, 0, sizeof(tButton));
                    result->QuitButton->iState = BUTTONSTATE_ACTIVE;
		  }
                }
                if (result->QuitButton)
                  Process_SpriteDefinition_SkinEntry(&(result->QuitButton->sprMouseOver), value, skin_fullpath, &(result->ImageCache));
	      } else
	      if (!stricmp(key, "QUITBUTTON_PRESSED"))
              {
                if (!(result->QuitButton))
                { // Allocate memory for the button, if not allocated yet!
                  result->QuitButton = (pButton) malloc(sizeof(tButton));
		  if (result->QuitButton)
		  {
		    memset(result->QuitButton, 0, sizeof(tButton));
                    result->QuitButton->iState = BUTTONSTATE_ACTIVE;
		  }
                }
                if (result->QuitButton)
                  Process_SpriteDefinition_SkinEntry(&(result->QuitButton->sprPressed), value, skin_fullpath, &(result->ImageCache));
	      } else
	      if (!stricmp(key, "QUITBUTTON_POSITION"))
              {
                if (!(result->QuitButton))
                { // Allocate memory for the button, if not allocated yet!
                  result->QuitButton = (pButton) malloc(sizeof(tButton));
		  if (result->QuitButton)
		  {
		    memset(result->QuitButton, 0, sizeof(tButton));
                    result->QuitButton->iState = BUTTONSTATE_ACTIVE;
		  }
                }
                if (result->QuitButton)
                  if (sscanf(value, "%d %d", &(result->QuitButton->XPos), &(result->QuitButton->YPos))!=2)
                  {
                    result->QuitButton->XPos=0;
                    result->QuitButton->YPos=0;
                  }
	      } else
	      if (!stricmp(key, "SETUPBUTTON_ACTIVE"))
              {
                if (!(result->SetupButton))
                { // Allocate memory for the button, if not allocated yet!
                  result->SetupButton = (pButton) malloc(sizeof(tButton));
		  if (result->SetupButton)
		  {
		    memset(result->SetupButton, 0, sizeof(tButton));
                    result->SetupButton->iState = BUTTONSTATE_ACTIVE;
		  }
                }
                if (result->SetupButton)
                  Process_SpriteDefinition_SkinEntry(&(result->SetupButton->sprActive), value, skin_fullpath, &(result->ImageCache));
	      } else
	      if (!stricmp(key, "SETUPBUTTON_MOUSEOVER"))
              {
                if (!(result->SetupButton))
                { // Allocate memory for the button, if not allocated yet!
                  result->SetupButton = (pButton) malloc(sizeof(tButton));
		  if (result->SetupButton)
		  {
		    memset(result->SetupButton, 0, sizeof(tButton));
                    result->SetupButton->iState = BUTTONSTATE_ACTIVE;
		  }
                }
                if (result->SetupButton)
                  Process_SpriteDefinition_SkinEntry(&(result->SetupButton->sprMouseOver), value, skin_fullpath, &(result->ImageCache));
	      } else
	      if (!stricmp(key, "SETUPBUTTON_PRESSED"))
              {
                if (!(result->SetupButton))
                { // Allocate memory for the button, if not allocated yet!
                  result->SetupButton = (pButton) malloc(sizeof(tButton));
		  if (result->SetupButton)
		  {
		    memset(result->SetupButton, 0, sizeof(tButton));
                    result->SetupButton->iState = BUTTONSTATE_ACTIVE;
		  }
                }
                if (result->SetupButton)
                  Process_SpriteDefinition_SkinEntry(&(result->SetupButton->sprPressed), value, skin_fullpath, &(result->ImageCache));
	      } else
	      if (!stricmp(key, "SETUPBUTTON_POSITION"))
              {
                if (!(result->SetupButton))
                { // Allocate memory for the button, if not allocated yet!
                  result->SetupButton = (pButton) malloc(sizeof(tButton));
		  if (result->SetupButton)
		  {
		    memset(result->SetupButton, 0, sizeof(tButton));
                    result->SetupButton->iState = BUTTONSTATE_ACTIVE;
		  }
                }
                if (result->SetupButton)
                  if (sscanf(value, "%d %d", &(result->SetupButton->XPos), &(result->SetupButton->YPos))!=2)
                  {
                    result->SetupButton->XPos=0;
                    result->SetupButton->YPos=0;
                  }
	      } else
	      if (!stricmp(key, "VOLUMEBUTTON_ACTIVE"))
              {
                if (!(result->VolumeButton))
                { // Allocate memory for the button, if not allocated yet!
                  result->VolumeButton = (pButton) malloc(sizeof(tButton));
		  if (result->VolumeButton)
		  {
		    memset(result->VolumeButton, 0, sizeof(tButton));
                    result->VolumeButton->iState = BUTTONSTATE_ACTIVE;
		  }
                }
                if (result->VolumeButton)
                  Process_SpriteDefinition_SkinEntry(&(result->VolumeButton->sprActive), value, skin_fullpath, &(result->ImageCache));
	      } else
	      if (!stricmp(key, "VOLUMEBUTTON_MOUSEOVER"))
              {
                if (!(result->VolumeButton))
                { // Allocate memory for the button, if not allocated yet!
                  result->VolumeButton = (pButton) malloc(sizeof(tButton));
		  if (result->VolumeButton)
		  {
		    memset(result->VolumeButton, 0, sizeof(tButton));
                    result->VolumeButton->iState = BUTTONSTATE_ACTIVE;
		  }
                }
                if (result->VolumeButton)
                  Process_SpriteDefinition_SkinEntry(&(result->VolumeButton->sprMouseOver), value, skin_fullpath, &(result->ImageCache));
	      } else
	      if (!stricmp(key, "VOLUMEBUTTON_PRESSED"))
              {
                if (!(result->VolumeButton))
                { // Allocate memory for the button, if not allocated yet!
                  result->VolumeButton = (pButton) malloc(sizeof(tButton));
		  if (result->VolumeButton)
		  {
		    memset(result->VolumeButton, 0, sizeof(tButton));
                    result->VolumeButton->iState = BUTTONSTATE_ACTIVE;
		  }
                }
                if (result->VolumeButton)
                  Process_SpriteDefinition_SkinEntry(&(result->VolumeButton->sprPressed), value, skin_fullpath, &(result->ImageCache));
	      } else
	      if (!stricmp(key, "VOLUMEBUTTON_POSITION"))
              {
                if (!(result->VolumeButton))
                { // Allocate memory for the button, if not allocated yet!
                  result->VolumeButton = (pButton) malloc(sizeof(tButton));
		  if (result->VolumeButton)
		  {
		    memset(result->VolumeButton, 0, sizeof(tButton));
                    result->VolumeButton->iState = BUTTONSTATE_ACTIVE;
		  }
                }
                if (result->VolumeButton)
                  if (sscanf(value, "%d %d", &(result->VolumeButton->XPos), &(result->VolumeButton->YPos))!=2)
                  {
                    result->VolumeButton->XPos=0;
                    result->VolumeButton->YPos=0;
                  }
	      } else
	      if (!stricmp(key, "STARTBUTTON_DISABLED"))
              {
                if (!(result->StartButton))
                { // Allocate memory for the button, if not allocated yet!
                  result->StartButton = (pButton) malloc(sizeof(tButton));
                  if (result->StartButton)
                    memset(result->StartButton, 0, sizeof(tButton));
                }
                if (result->StartButton)
                  Process_SpriteDefinition_SkinEntry(&(result->StartButton->sprDisabled), value, skin_fullpath, &(result->ImageCache));
	      } else
	      if (!stricmp(key, "STARTBUTTON_ACTIVE"))
              {
                if (!(result->StartButton))
                { // Allocate memory for the button, if not allocated yet!
                  result->StartButton = (pButton) malloc(sizeof(tButton));
                  if (result->StartButton)
                    memset(result->StartButton, 0, sizeof(tButton));
                }
                if (result->StartButton)
                  Process_SpriteDefinition_SkinEntry(&(result->StartButton->sprActive), value, skin_fullpath, &(result->ImageCache));
	      } else
	      if (!stricmp(key, "STARTBUTTON_MOUSEOVER"))
              {
                if (!(result->StartButton))
                { // Allocate memory for the button, if not allocated yet!
                  result->StartButton = (pButton) malloc(sizeof(tButton));
                  if (result->StartButton)
                    memset(result->StartButton, 0, sizeof(tButton));
                }
                if (result->StartButton)
                  Process_SpriteDefinition_SkinEntry(&(result->StartButton->sprMouseOver), value, skin_fullpath, &(result->ImageCache));
	      } else
	      if (!stricmp(key, "STARTBUTTON_PRESSED"))
              {
                if (!(result->StartButton))
                { // Allocate memory for the button, if not allocated yet!
                  result->StartButton = (pButton) malloc(sizeof(tButton));
                  if (result->StartButton)
                    memset(result->StartButton, 0, sizeof(tButton));
                }
                if (result->StartButton)
                  Process_SpriteDefinition_SkinEntry(&(result->StartButton->sprPressed), value, skin_fullpath, &(result->ImageCache));
	      } else
	      if (!stricmp(key, "STARTBUTTON_POSITION"))
              {
                if (!(result->StartButton))
                { // Allocate memory for the button, if not allocated yet!
                  result->StartButton = (pButton) malloc(sizeof(tButton));
                  if (result->StartButton)
                    memset(result->StartButton, 0, sizeof(tButton));
                }
                if (result->StartButton)
                  if (sscanf(value, "%d %d", &(result->StartButton->XPos), &(result->StartButton->YPos))!=2)
                  {
                    result->StartButton->XPos=0;
                    result->StartButton->YPos=0;
		  }
	      } else
	      if (!stricmp(key, "WELCOMEMSG"))
              {
                Process_Message_SkinEntry(&(result->msgWelcome), value);
	      } else
	      if (!stricmp(key, "ENEMYDISCONNECTEDMSG"))
              {
                Process_Message_SkinEntry(&(result->msgEnemyDisconnected), value);
	      } else
	      if (!stricmp(key, "WAITINGFORENEMYMSG"))
              {
                Process_Message_SkinEntry(&(result->msgWaitingForEnemy), value);
	      } else
	      if (!stricmp(key, "ENEMYPOSITIONEDMSG"))
              {
                Process_Message_SkinEntry(&(result->msgEnemyPositioned), value);
	      } else
	      if (!stricmp(key, "ENEMYPOSITIONEDTOOSRVMSG"))
              {
                Process_Message_SkinEntry(&(result->msgEnemyPositionedTooSrv), value);
	      } else
	      if (!stricmp(key, "ENEMYPOSITIONEDTOOCLIMSG"))
              {
                Process_Message_SkinEntry(&(result->msgEnemyPositionedTooCli), value);
	      } else
	      if (!stricmp(key, "WEPOSITIONEDTOOSRVMSG"))
              {
                Process_Message_SkinEntry(&(result->msgWePositionedTooSrv), value);
	      } else
	      if (!stricmp(key, "WEPOSITIONEDTOOCLIMSG"))
              {
                Process_Message_SkinEntry(&(result->msgWePositionedTooCli), value);
	      } else
	      if (!stricmp(key, "ENEMYSHOTMSG"))
              {
                Process_Message_SkinEntry(&(result->msgEnemyShot), value);
	      } else
	      if (!stricmp(key, "WEHITMSG"))
              {
                Process_Message_SkinEntry(&(result->msgWeHit), value);
	      } else
	      if (!stricmp(key, "WEMISSEDMSG"))
              {
                Process_Message_SkinEntry(&(result->msgWeMissed), value);
	      } else
	      if (!stricmp(key, "GAMEWONMSG"))
              {
                Process_Message_SkinEntry(&(result->msgGameWon), value);
	      } else
	      if (!stricmp(key, "GAMELOSTMSG"))
              {
                Process_Message_SkinEntry(&(result->msgGameLost), value);
	      } else
	      if (!stricmp(key, "CONFIGCHANGEDMSG"))
              {
                Process_Message_SkinEntry(&(result->msgConfigChanged), value);
	      } else
	      if (!stricmp(key, "BACKGROUNDMUSIC"))
              {
                Process_SFX_SkinEntry(&(result->BackgroundMusic), skin_fullpath, value);
	      } else
	      if (!stricmp(key, "WEHITSFX"))
              {
                Process_SFX_SkinEntry(&(result->WeHitSFX), skin_fullpath, value);
	      } else
	      if (!stricmp(key, "WEMISSSFX"))
              {
                Process_SFX_SkinEntry(&(result->WeMissSFX), skin_fullpath, value);
	      } else
	      if (!stricmp(key, "ENEMYHITSFX"))
              {
                Process_SFX_SkinEntry(&(result->EnemyHitSFX), skin_fullpath, value);
	      } else
	      if (!stricmp(key, "ENEMYMISSSFX"))
              {
                Process_SFX_SkinEntry(&(result->EnemyMissSFX), skin_fullpath, value);
	      } else
	      if (!stricmp(key, "OVERBUTTONSFX"))
              {
                Process_SFX_SkinEntry(&(result->OverButtonSFX), skin_fullpath, value);
	      } else
	      if (!stricmp(key, "LEAVEBUTTONSFX"))
              {
                Process_SFX_SkinEntry(&(result->LeaveButtonSFX), skin_fullpath, value);
	      } else
	      if (!stricmp(key, "PRESSBUTTONSFX"))
              {
                Process_SFX_SkinEntry(&(result->PressButtonSFX), skin_fullpath, value);
	      } else
	      if (!stricmp(key, "GAMESTARTSFX"))
              {
                Process_SFX_SkinEntry(&(result->GameStartSFX), skin_fullpath, value);
	      } else
	      if (!stricmp(key, "GAMEWONSFX"))
              {
                Process_SFX_SkinEntry(&(result->GameWonSFX), skin_fullpath, value);
	      } else
	      if (!stricmp(key, "GAMELOSTSFX"))
              {
                Process_SFX_SkinEntry(&(result->GameLostSFX), skin_fullpath, value);
              } else
	      if (!stricmp(key, "WEHITANIMATIONWAITTIME"))
              {
		if (sscanf(value, "%d",
			   &(result->WeHitAnimationWaitTime))!=1)
		  result->WeHitAnimationWaitTime = 0;
              } else
	      if (!stricmp(key, "WEMISSANIMATIONWAITTIME"))
              {
		if (sscanf(value, "%d",
			   &(result->WeMissAnimationWaitTime))!=1)
		  result->WeMissAnimationWaitTime = 0;
              } else
	      if (!stricmp(key, "ENEMYHITANIMATIONWAITTIME"))
              {
		if (sscanf(value, "%d",
			   &(result->EnemyHitAnimationWaitTime))!=1)
		  result->EnemyHitAnimationWaitTime = 0;
              } else
	      if (!stricmp(key, "ENEMYMISSANIMATIONWAITTIME"))
              {
		if (sscanf(value, "%d",
			   &(result->EnemyMissAnimationWaitTime))!=1)
		  result->EnemyMissAnimationWaitTime = 0;
	      } else
	      if (!stricmp(key, "VOLUMEPANEL"))
	      { // Here comes the description of the volume setting panel
                Process_SpriteDefinition_SkinEntry(&(result->VolumePanel), value, skin_fullpath, &(result->ImageCache));
	      } else
	      if (!stricmp(key, "MAINVOLUMEARM_ACTIVE"))
              {
                if (!(result->MainVolumeArm))
                { // Allocate memory for the button, if not allocated yet!
                  result->MainVolumeArm = (pButton) malloc(sizeof(tButton));
		  if (result->MainVolumeArm)
		  {
		    memset(result->MainVolumeArm, 0, sizeof(tButton));
                    result->MainVolumeArm->iState = BUTTONSTATE_ACTIVE;
		  }
                }
                if (result->MainVolumeArm)
                  Process_SpriteDefinition_SkinEntry(&(result->MainVolumeArm->sprActive), value, skin_fullpath, &(result->ImageCache));
	      } else
	      if (!stricmp(key, "MAINVOLUMEARM_MOUSEOVER"))
              {
                if (!(result->MainVolumeArm))
                { // Allocate memory for the button, if not allocated yet!
                  result->MainVolumeArm = (pButton) malloc(sizeof(tButton));
		  if (result->MainVolumeArm)
		  {
		    memset(result->MainVolumeArm, 0, sizeof(tButton));
                    result->MainVolumeArm->iState = BUTTONSTATE_ACTIVE;
		  }
                }
                if (result->MainVolumeArm)
                  Process_SpriteDefinition_SkinEntry(&(result->MainVolumeArm->sprMouseOver), value, skin_fullpath, &(result->ImageCache));
	      } else
	      if (!stricmp(key, "MAINVOLUMEARM_PRESSED"))
              {
                if (!(result->MainVolumeArm))
                { // Allocate memory for the button, if not allocated yet!
                  result->MainVolumeArm = (pButton) malloc(sizeof(tButton));
		  if (result->MainVolumeArm)
		  {
		    memset(result->MainVolumeArm, 0, sizeof(tButton));
                    result->MainVolumeArm->iState = BUTTONSTATE_ACTIVE;
		  }
                }
                if (result->MainVolumeArm)
                  Process_SpriteDefinition_SkinEntry(&(result->MainVolumeArm->sprPressed), value, skin_fullpath, &(result->ImageCache));
	      } else
	      if (!stricmp(key, "MUSICVOLUMEARM_ACTIVE"))
              {
                if (!(result->MusicVolumeArm))
                { // Allocate memory for the button, if not allocated yet!
                  result->MusicVolumeArm = (pButton) malloc(sizeof(tButton));
		  if (result->MusicVolumeArm)
		  {
		    memset(result->MusicVolumeArm, 0, sizeof(tButton));
                    result->MusicVolumeArm->iState = BUTTONSTATE_ACTIVE;
		  }
                }
                if (result->MusicVolumeArm)
                  Process_SpriteDefinition_SkinEntry(&(result->MusicVolumeArm->sprActive), value, skin_fullpath, &(result->ImageCache));
	      } else
	      if (!stricmp(key, "MUSICVOLUMEARM_MOUSEOVER"))
              {
                if (!(result->MusicVolumeArm))
                { // Allocate memory for the button, if not allocated yet!
                  result->MusicVolumeArm = (pButton) malloc(sizeof(tButton));
		  if (result->MusicVolumeArm)
		  {
                    memset(result->MusicVolumeArm, 0, sizeof(tButton));
                    result->MusicVolumeArm->iState = BUTTONSTATE_ACTIVE;
		  }
                }
                if (result->MusicVolumeArm)
                  Process_SpriteDefinition_SkinEntry(&(result->MusicVolumeArm->sprMouseOver), value, skin_fullpath, &(result->ImageCache));
	      } else
	      if (!stricmp(key, "MUSICVOLUMEARM_PRESSED"))
              {
                if (!(result->MusicVolumeArm))
                { // Allocate memory for the button, if not allocated yet!
                  result->MusicVolumeArm = (pButton) malloc(sizeof(tButton));
		  if (result->MusicVolumeArm)
		  {
                    memset(result->MusicVolumeArm, 0, sizeof(tButton));
                    result->MusicVolumeArm->iState = BUTTONSTATE_ACTIVE;
		  }
                }
                if (result->MusicVolumeArm)
                  Process_SpriteDefinition_SkinEntry(&(result->MusicVolumeArm->sprPressed), value, skin_fullpath, &(result->ImageCache));
	      } else
	      if (!stricmp(key, "SFXVOLUMEARM_ACTIVE"))
              {
                if (!(result->SFXVolumeArm))
                { // Allocate memory for the button, if not allocated yet!
                  result->SFXVolumeArm = (pButton) malloc(sizeof(tButton));
		  if (result->SFXVolumeArm)
		  {
                    memset(result->SFXVolumeArm, 0, sizeof(tButton));
                    result->SFXVolumeArm->iState = BUTTONSTATE_ACTIVE;
		  }
                }
                if (result->SFXVolumeArm)
                  Process_SpriteDefinition_SkinEntry(&(result->SFXVolumeArm->sprActive), value, skin_fullpath, &(result->ImageCache));
	      } else
	      if (!stricmp(key, "SFXVOLUMEARM_MOUSEOVER"))
              {
                if (!(result->SFXVolumeArm))
                { // Allocate memory for the button, if not allocated yet!
                  result->SFXVolumeArm = (pButton) malloc(sizeof(tButton));
		  if (result->SFXVolumeArm)
		  {
                    memset(result->SFXVolumeArm, 0, sizeof(tButton));
                    result->SFXVolumeArm->iState = BUTTONSTATE_ACTIVE;
		  }
                }
                if (result->SFXVolumeArm)
                  Process_SpriteDefinition_SkinEntry(&(result->SFXVolumeArm->sprMouseOver), value, skin_fullpath, &(result->ImageCache));
	      } else
	      if (!stricmp(key, "SFXVOLUMEARM_PRESSED"))
              {
                if (!(result->SFXVolumeArm))
                { // Allocate memory for the button, if not allocated yet!
                  result->SFXVolumeArm = (pButton) malloc(sizeof(tButton));
		  if (result->SFXVolumeArm)
		  {
                    memset(result->SFXVolumeArm, 0, sizeof(tButton));
                    result->SFXVolumeArm->iState = BUTTONSTATE_ACTIVE;
		  }
                }
                if (result->SFXVolumeArm)
                  Process_SpriteDefinition_SkinEntry(&(result->SFXVolumeArm->sprPressed), value, skin_fullpath, &(result->ImageCache));
	      } else
	      if (!stricmp(key, "VOLUMEPANELCLOSE_ACTIVE"))
              {
                if (!(result->VolumePanelCloseBtn))
                { // Allocate memory for the button, if not allocated yet!
                  result->VolumePanelCloseBtn = (pButton) malloc(sizeof(tButton));
		  if (result->VolumePanelCloseBtn)
		  {
                    memset(result->VolumePanelCloseBtn, 0, sizeof(tButton));
                    result->VolumePanelCloseBtn->iState = BUTTONSTATE_ACTIVE;
		  }
                }
                if (result->VolumePanelCloseBtn)
                  Process_SpriteDefinition_SkinEntry(&(result->VolumePanelCloseBtn->sprActive), value, skin_fullpath, &(result->ImageCache));
	      } else
	      if (!stricmp(key, "VOLUMEPANELCLOSE_MOUSEOVER"))
              {
                if (!(result->VolumePanelCloseBtn))
                { // Allocate memory for the button, if not allocated yet!
                  result->VolumePanelCloseBtn = (pButton) malloc(sizeof(tButton));
		  if (result->VolumePanelCloseBtn)
		  {
                    memset(result->VolumePanelCloseBtn, 0, sizeof(tButton));
                    result->VolumePanelCloseBtn->iState = BUTTONSTATE_ACTIVE;
		  }
                }
                if (result->VolumePanelCloseBtn)
                  Process_SpriteDefinition_SkinEntry(&(result->VolumePanelCloseBtn->sprMouseOver), value, skin_fullpath, &(result->ImageCache));
	      } else
	      if (!stricmp(key, "VOLUMEPANELCLOSE_PRESSED"))
              {
                if (!(result->VolumePanelCloseBtn))
                { // Allocate memory for the button, if not allocated yet!
                  result->VolumePanelCloseBtn = (pButton) malloc(sizeof(tButton));
		  if (result->VolumePanelCloseBtn)
		  {
		    memset(result->VolumePanelCloseBtn, 0, sizeof(tButton));
                    result->VolumePanelCloseBtn->iState = BUTTONSTATE_ACTIVE;
		  }
                }
                if (result->VolumePanelCloseBtn)
                  Process_SpriteDefinition_SkinEntry(&(result->VolumePanelCloseBtn->sprPressed), value, skin_fullpath, &(result->ImageCache));
	      } else
	      if (!stricmp(key, "VOLUMEPANELCLOSE_POSITION"))
	      {
                if (!(result->VolumePanelCloseBtn))
                { // Allocate memory for the button, if not allocated yet!
                  result->VolumePanelCloseBtn = (pButton) malloc(sizeof(tButton));
		  if (result->VolumePanelCloseBtn)
		  {
		    memset(result->VolumePanelCloseBtn, 0, sizeof(tButton));
                    result->VolumePanelCloseBtn->iState = BUTTONSTATE_ACTIVE;
		  }
                }
                if (result->VolumePanelCloseBtn)
                  if (sscanf(value, "%d %d", &(result->VolumePanelCloseBtn->XPos), &(result->VolumePanelCloseBtn->YPos))!=2)
                  {
                    result->VolumePanelCloseBtn->XPos=0;
                    result->VolumePanelCloseBtn->YPos=0;
                  }
	      } else
	      if (!stricmp(key, "MAINVOLUMELINE"))
	      {
		if (sscanf(value, "%d %d %d",
			   &(result->MainVolumeLineX),
			   &(result->MainVolumeLineY),
                           &(result->MainVolumeLineLength))!=3)
		{
		  result->MainVolumeLineX = 0;
                  result->MainVolumeLineY = 0;
		  result->MainVolumeLineLength = 0;
		}
	      } else
	      if (!stricmp(key, "MUSICVOLUMELINE"))
	      {
		if (sscanf(value, "%d %d %d",
			   &(result->MusicVolumeLineX),
			   &(result->MusicVolumeLineY),
                           &(result->MusicVolumeLineLength))!=3)
		{
		  result->MusicVolumeLineX = 0;
                  result->MusicVolumeLineY = 0;
		  result->MusicVolumeLineLength = 0;
		}
	      } else
	      if (!stricmp(key, "SFXVOLUMELINE"))
	      {
		if (sscanf(value, "%d %d %d",
			   &(result->SFXVolumeLineX),
			   &(result->SFXVolumeLineY),
                           &(result->SFXVolumeLineLength))!=3)
		{
		  result->SFXVolumeLineX = 0;
                  result->SFXVolumeLineY = 0;
		  result->SFXVolumeLineLength = 0;
		}
	      } else
	      if (!stricmp(key, "VOLUMEPANELPOS"))
	      {
		if (sscanf(value, "%d %d",
			   &(result->VolumePanelX),
			   &(result->VolumePanelY))!=2)
		{
		  result->VolumePanelX = 0;
                  result->VolumePanelY = 0;
		}
	      }

	    } else fscanf(skinfile, "%*[^\n]\n"); // find EOL
	  } else fscanf(skinfile, "%*[^\n]\n"); // find EOL
	} else fscanf(skinfile, "%*[^\n]\n"); // find EOL
      }
    }
    // Ok, skin file processed.
    fclose(skinfile);
  }

  // Free temporary variables
  free(skin_fullpath);
  free(key);
  free(value);
  free(stemp);

  // Set startup values for volume setting panel:
  if (result)
  {
    if ((result->SFXVolumeArm) && (result->VolumePanel))
    {
      result->SFXVolumeArm->XPos =
	result->VolumePanelX - result->VolumePanel->ulHotSpotX +
	result->SFXVolumeLineX;
      result->SFXVolumeArm->YPos =
	result->VolumePanelY - result->VolumePanel->ulHotSpotY +
	result->SFXVolumeLineY;
    }
    if ((result->MusicVolumeArm) && (result->VolumePanel))
    {
      result->MusicVolumeArm->XPos =
	result->VolumePanelX - result->VolumePanel->ulHotSpotX +
	result->MusicVolumeLineX;
      result->MusicVolumeArm->YPos =
	result->VolumePanelY - result->VolumePanel->ulHotSpotY +
	result->MusicVolumeLineY;
    }
    if ((result->MainVolumeArm) && (result->VolumePanel))
    {
      result->MainVolumeArm->XPos =
	result->VolumePanelX - result->VolumePanel->ulHotSpotX +
	result->MainVolumeLineX;
      result->MainVolumeArm->YPos =
	result->VolumePanelY - result->VolumePanel->ulHotSpotY +
	result->MainVolumeLineY;
    }
    if ((result->VolumePanelCloseBtn) && (result->VolumePanel))
    {
      result->VolumePanelCloseBtn->XPos =
	result->VolumePanelX - result->VolumePanel->ulHotSpotX +
	result->VolumePanelCloseBtn->XPos;
      result->VolumePanelCloseBtn->YPos =
	result->VolumePanelY - result->VolumePanel->ulHotSpotY +
	result->VolumePanelCloseBtn->YPos;
    }
  }

#ifdef DEBUG_BUILD
  printf("Skin loaded\n");
  fflush(stdout);
#endif

  // TODO: more checking (buttons, etc...)
  if (result)
  {
    // Now check if something vital is missing from skin or not!
    if ((!result->BackgroundImage) ||  // Could not load image!
	(!result->PositionCursor))    // Could not load cursor!
    {
  #ifdef DEBUG_BUILD
      printf("Missing BG Image or PositionCursor\n");
      fflush(stdout);
  #endif
      CleanupSkin(&result);
    }
    else
  
    { // Check for presence of ships
      int i;
      for (i=0; i<NUM_OF_SHIPS; i++)
	if((!result->Ship_Down[i]) || // Could not load one ship
	   (!result->Ship_Right[i]))
	{
  #ifdef DEBUG_BUILD
	  printf("Missing ship data\n");
	  fflush(stdout);
  #endif
  
	  CleanupSkin(&result);
	  break;
	}
    }
  
    // Ok, everything checked, and the skin is cleaned and freed if there
    // was something vital missing.
  
    // Clear image cache, no more need for it
    if (result->ImageCache)
    {
      CleanupImage(result->ImageCache);
      result->ImageCache = NULL;
    }
  }
  // Return the result.
  return result;
}

//---------------------------------------------------------------------
// NetworkReceiver
//
// This is the NetworkReceiver thread. It waits for messages coming 
// from the other side through TCP/IP connection, and processes those
// messages.
// Quits, when signaled with the fShuttingDown flag, or when killed.:)
//---------------------------------------------------------------------
VOID NetworkReceiver(PVOID Param)
{
  CHAR Msg[256];           // Grab the message, that arrives, into this
  CHAR cmd[20];            // Extract command from the message into this
  int x,y;

#ifdef DEBUG_BUILD
  printf("Starting NetworkReceiver thread\n");
  fflush(stdout);
#endif

  while (!fShuttingDown) {     // Now process messages!

    // Get a message into the message buffer.
    // Exit the thread, if an error occures!
    memset(Msg, 256, 0);
#ifdef DEBUG_BUILD
    printf("[NR] Waiting for a message...\n"); fflush(stdout);
#endif
    if (!((*pReceiveMessage)( Socket, 255, Msg, -1 ))) goto EndThread;
#ifdef DEBUG_BUILD
    printf("[NR] Something arrived?\n", Msg); fflush(stdout);
#endif
    if (!stricmp(Msg,"")) goto EndThread;
#ifdef DEBUG_BUILD
    printf("[NR] Msg: %s\n", Msg); fflush(stdout);
#endif
    // Ok, we have a message here. Let's see what kind of message,
    // it is!

    // If it's a disconnect message, disconnect!
    if (!stricmp(Msg,"DISCONNECT"))
    {
      if ((iGameState!=GAMESTATE_END_OF_GAME_WON) &&
	  (iGameState!=GAMESTATE_END_OF_GAME_LOST) &&
	  (iGameState!=GAMESTATE_PEER_DISCONNECTED) )
      {
        if (DosRequestMutexSem( hmtxUseSkin, SEM_INDEFINITE_WAIT ) == NO_ERROR )
        {
          if ((Skin) && (Skin->msgEnemyDisconnected))
            NewMessage_to_MessageBoard(Skin->msgEnemyDisconnected);
          DosReleaseMutexSem(hmtxUseSkin);
        }
      }
      iGameState=GAMESTATE_PEER_DISCONNECTED;
      fFirstTimeRender = TRUE;  // Make sure the new background will be drawn!
      goto EndThread;
    } else
    // If it's a "ships positioned" message
    if (!stricmp(Msg,"POSITIONED"))
    {
      if (DosRequestMutexSem( hmtxUseSkin, SEM_INDEFINITE_WAIT ) == NO_ERROR )
      {
	if (iGameState==GAMESTATE_WAIT_FOR_POSITION_ENEMY)
	{ // We are ready, and now the enemy is ready too!
	  // Let the game begin!
          if (fWeAreTheServer)
          {
            if ((Skin) && (Skin->msgEnemyPositionedTooSrv))
              NewMessage_to_MessageBoard(Skin->msgEnemyPositionedTooSrv);
            iGameState=GAMESTATE_SELF_ROUND;
            Initialize_ulInitTimeStamp(); // Start counting in-game time!
            fFirstTimeRender = TRUE;  // Make sure the new background will be drawn!
          } else
          {
            if ((Skin) && (Skin->msgEnemyPositionedTooCli))
              NewMessage_to_MessageBoard(Skin->msgEnemyPositionedTooCli);
            iGameState=GAMESTATE_ENEMY_ROUND;
            Initialize_ulInitTimeStamp(); // Start counting in-game time!
            fFirstTimeRender = TRUE;  // Make sure the new background will be drawn!
          }
          // Also make audible notification!
          if (Skin)
            SFXPlayFX(Skin->GameStartSFX);
        } else
        { // We're not ready yet, but the enemy is!
          if ((Skin) && (Skin->msgEnemyPositioned))
            NewMessage_to_MessageBoard(Skin->msgEnemyPositioned);
          iGameState=GAMESTATE_POSITION_SELF_ENEMY_POSITIONED;
          fFirstTimeRender = TRUE;  // Make sure the new background will be drawn!
        }
        DosReleaseMutexSem( hmtxUseSkin );
      }
    } else
    // If it's a shoot message
    if (!stricmp(substr(cmd, Msg, 0, 6),"SHOOT "))
    {
      if (DosRequestMutexSem( hmtxUseSkin, SEM_INDEFINITE_WAIT ) == NO_ERROR )
      {
        if (sscanf(Msg+6, "%d %d", &x, &y)==2)
        {
          if ((x>=0) && (x<=9) && (y>=0) && (y<=9) &&
              (iGameState==GAMESTATE_ENEMY_ROUND))
            EnemyShootsAt(x, y);
        }
        DosReleaseMutexSem(hmtxUseSkin);
      }
    } else
    // If it's a hit reply
    if (!stricmp(substr(cmd, Msg, 0, 4),"HIT "))
    {
      if (DosRequestMutexSem( hmtxUseSkin, SEM_INDEFINITE_WAIT ) == NO_ERROR )
      {
        if (sscanf(Msg+4, "%d %d", &x, &y)==2)
        {
          if ((x>=0) && (x<=9) && (y>=0) && (y<=9))
          {
            EnemyTableMiscData[x][y] = clock(); // save actual time, so Audio waiting will know how much to wait!
            EnemyTable[x][y] = TABLEENTRY_HIT_AUDIOONLY;
            if (Skin)
              SFXPlayFX(Skin->WeHitSFX);
          }
        }
        DosReleaseMutexSem(hmtxUseSkin);
      }
    } else
    // If it's a miss reply
    if (!stricmp(substr(cmd, Msg, 0, 5),"MISS "))
    {
      if (DosRequestMutexSem( hmtxUseSkin, SEM_INDEFINITE_WAIT ) == NO_ERROR )
      {
        if (sscanf(Msg+5, "%d %d", &x, &y)==2)
        {
          if ((x>=0) && (x<=9) && (y>=0) && (y<=9))
          {
            EnemyTableMiscData[x][y] = clock(); // save actual time, so Audio waiting will know how much to wait!
            EnemyTable[x][y] = TABLEENTRY_MISS_AUDIOONLY;
            if (Skin)
              SFXPlayFX(Skin->WeMissSFX);
          }
        }
        DosReleaseMutexSem(hmtxUseSkin);
      }
    }
    // Otherwise unknown command, do nothing

  } // End of while (fShuttingDown)

 EndThread:
  // Exiting the thread:
#ifdef DEBUG_BUILD
  printf("[NR] EndThread, Msg: %s\n", Msg); fflush(stdout);
#endif
  (*pSendMessage)(Socket, "DISCONNECT");

#ifdef DEBUG_BUILD
  printf("Halting NetworkReceiver thread\n");
  fflush(stdout);
#endif

  if ((iGameState!=GAMESTATE_END_OF_GAME_WON) &&
      (iGameState!=GAMESTATE_END_OF_GAME_LOST) &&
      (iGameState!=GAMESTATE_PEER_DISCONNECTED) )
  { // If it was a sudden disconnection:
    if (DosRequestMutexSem( hmtxUseSkin, SEM_INDEFINITE_WAIT ) == NO_ERROR )
    {
      if ((Skin) && (Skin->msgEnemyDisconnected))
        NewMessage_to_MessageBoard(Skin->msgEnemyDisconnected);
      DosReleaseMutexSem(hmtxUseSkin);
    }
  }
  iGameState=GAMESTATE_PEER_DISCONNECTED;
  fFirstTimeRender = TRUE;  // Make sure the new background will be drawn!
  _endthread();
}

//---------------------------------------------------------------------
// SetMainVolume
//
//---------------------------------------------------------------------
void SetMainVolume(int iVolume)
{
  CurrentConfig.iMainVolume = iVolume;
  SFXSetMainVolume(iVolume);
  if (Skin)
    if (Skin->MainVolumeArm)
    {
      Skin->MainVolumeArm->XPos =
	Skin->VolumePanelX - Skin->VolumePanel->ulHotSpotX +
	Skin->MainVolumeLineX +
        ((Skin->MainVolumeLineLength)*iVolume/SFX_VOLUME__MAX);
    }
}

//---------------------------------------------------------------------
// SetMusicVolume
//
// There is one SFX, that is treated to be special, this is the
// BackgroundMusic. Its volume is always set separated.
//---------------------------------------------------------------------
void SetMusicVolume(int iVolume)
{
  CurrentConfig.iMusicVolume = iVolume;
  if (Skin)
  {
    SFXSetFXVolume(Skin->BackgroundMusic, iVolume);
    if (Skin->MusicVolumeArm)
    {
      Skin->MusicVolumeArm->XPos =
	Skin->VolumePanelX - Skin->VolumePanel->ulHotSpotX +
	Skin->MusicVolumeLineX +
        ((Skin->MusicVolumeLineLength)*iVolume/SFX_VOLUME__MAX);
    }
  }
}

//---------------------------------------------------------------------
// SetSFXVolume
//
// All the SFXes, that are not BackgroundMusic, are having the same
// volume. This is the function that sets the volume for them.
// Remember, that if you add new SFXes to the skin definition,
// they should be also inclueded here!
//---------------------------------------------------------------------
void SetSFXVolume(int iVolume)
{
  CurrentConfig.iSFXVolume = iVolume;
  if (Skin)
  {
    SFXSetFXVolume(Skin->WeHitSFX, iVolume);
    SFXSetFXVolume(Skin->WeMissSFX, iVolume);
    SFXSetFXVolume(Skin->EnemyHitSFX, iVolume);
    SFXSetFXVolume(Skin->EnemyMissSFX, iVolume);
    SFXSetFXVolume(Skin->OverButtonSFX, iVolume);
    SFXSetFXVolume(Skin->LeaveButtonSFX, iVolume);
    SFXSetFXVolume(Skin->PressButtonSFX, iVolume);
    SFXSetFXVolume(Skin->GameStartSFX, iVolume);
    SFXSetFXVolume(Skin->GameWonSFX, iVolume);
    SFXSetFXVolume(Skin->GameLostSFX, iVolume);
    if (Skin->SFXVolumeArm)
    {
      Skin->SFXVolumeArm->XPos =
	Skin->VolumePanelX - Skin->VolumePanel->ulHotSpotX +
	Skin->SFXVolumeLineX +
        ((Skin->SFXVolumeLineLength)*iVolume/SFX_VOLUME__MAX);
    }
  }
}

//---------------------------------------------------------------------
// ErrorWindow
//
// Shows an error dialog window with a given caption and text.
//---------------------------------------------------------------------
void ErrorWindow(char *caption, char *text)
{
  WinMessageBox(HWND_DESKTOP, HWND_DESKTOP, text, caption, 0, MB_OK | MB_ERROR | MB_MOVEABLE);
}

//---------------------------------------------------------------------
// StartGame_General
//
// Event if starting as a server or as a client, this function will do
// the main job (because it's the same for both modes, just some small
// changes needed...)
//---------------------------------------------------------------------
BOOL StartGame_General( ULONG BufSize, pStartClientBuf Buffer )
{
  QMSG Msg;

#ifdef DEBUG_BUILD
  printf("StartGame() [IN]\n");
  fflush(stdout);
#endif

  // Check the parameters!
  if (BufSize!=sizeof(tStartClientBuf))
  {
    ErrorWindow("BattleShips - Internal error",
                "Invalid parameter for StartGame!");
    return FALSE;
  }

  fShuttingDown=FALSE;

  // Get direct pointers to local and remote usernames
  LocalUserName=Buffer->LocalUserName;
  RemoteUserName=Buffer->RemoteUserName;
  // Get TCP/IP socket to communicate with peer
  Socket = Buffer->Socket;
  // Save DLL module handle too:
  DLLHandle = Buffer->hmod;

  // And get some more information from parameters
  Ab = Buffer->ab;  // Anchor block (initialized by NetGameClient.exe)
  Mq = Buffer->mq;  // Message queue (also)

  // Load configuration into NewCOnfig, and set it to be the current one!
  LoadConfig(&CurrentConfig);

  // Initialize main subsystems
  InitializeImages();
  InitializeSprites();
  InitializeFonts();

  // Check the capabilities of the DIVE subsystem
  // Also sets srcBufferFOURCC to the best available fourCC to use
  if (!InspectDIVE())
  {
    UninitializeFonts();
    UninitializeSprites();
    UninitializeImages();
    ErrorWindow("BattleShips - Error",
                "DIVE not installed, or could not find requested video mode!");
    return FALSE;
  }

  // Initialize sound subsystem, according to the config file!
  if (!SFXOpenDevice(CurrentConfig.SFXDeviceID,
                     CurrentConfig.SFXDeviceShared,
                     CurrentConfig.SFXFreq,
                     CurrentConfig.SFXBits,
                     CurrentConfig.SFXChannels,
                     CurrentConfig.SFXNumBufs,
                     CurrentConfig.SFXBufSize))
  {
    ErrorWindow("BattleShips - Warning",
                "Could not initialize sound! Sound will be disabled!");
  }

  // Now load the skin!
  Skin = LoadSkin(CurrentConfig.SkinFilename);

  if (!Skin)
  {
    SFXCloseDevice();
    UninitializeFonts();
    UninitializeSprites();
    UninitializeImages();
    ErrorWindow("BattleShips - Error",
                "Could not load skin!");
    return FALSE;
  }

  // Set volumes
  SetMainVolume(CurrentConfig.iMainVolume);
  SetMusicVolume(CurrentConfig.iMusicVolume);
  SetSFXVolume(CurrentConfig.iSFXVolume);

  // Setup start values
  srcBufferWidth = Skin->BackgroundImage->ulWidth;
  srcBufferHeight = Skin->BackgroundImage->ulHeight;

  // First, create a mutex semaphore for protecting the Dive!
  if (DosCreateMutexSem(NULL, &hmtxUseDive, 0, FALSE)!=NO_ERROR)
  {
    CleanupSkin(&Skin);
    SFXCloseDevice();
    UninitializeFonts();
    UninitializeSprites();
    UninitializeImages();
    ErrorWindow("BattleShips - Error",
                "Could not create mutex semaphore!");
    return FALSE;
  }

  // Then create a mutex semaphore for protecting the Skin!
  if (DosCreateMutexSem(NULL, &hmtxUseSkin, 0, FALSE)!=NO_ERROR)
  {
    DosCloseMutexSem(hmtxUseDive);
    CleanupSkin(&Skin);
    SFXCloseDevice();
    UninitializeFonts();
    UninitializeSprites();
    UninitializeImages();
    ErrorWindow("BattleShips - Error",
                "Could not create mutex semaphore!");
    return FALSE;
  }

  // Then, create an event semaphore for the image blitter!
  if (DosCreateEventSem(NULL, &hevBlitNeededEvent, 0, FALSE)!=NO_ERROR)
  {
    DosCloseMutexSem(hmtxUseDive);
    DosCloseMutexSem(hmtxUseSkin);
    CleanupSkin(&Skin);
    SFXCloseDevice();
    UninitializeFonts();
    UninitializeSprites();
    UninitializeImages();
    ErrorWindow("BattleShips - Error",
                "Could not create event semaphore!");
    return FALSE;
  }

  // Create window
  if (!SetupMainWindow())
  {
    DosCloseEventSem(hevBlitNeededEvent);
    DosCloseMutexSem(hmtxUseDive);
    DosCloseMutexSem(hmtxUseSkin);
    CleanupSkin(&Skin);
    SFXCloseDevice();
    UninitializeFonts();
    UninitializeSprites();
    UninitializeImages();
    ErrorWindow("BattleShips - Error",
                "Could not create main window!");
    return FALSE;
  }

  // Setup DIVE
  if (!SetupDIVE())
  {
    DestroyMainWindow();
    DosCloseMutexSem(hmtxUseDive);
    DosCloseMutexSem(hmtxUseSkin);
    DosCloseEventSem(hevBlitNeededEvent);
    CleanupSkin(&Skin);
    SFXCloseDevice();
    UninitializeFonts();
    UninitializeSprites();
    UninitializeImages();
    ErrorWindow("BattleShips - Error",
                "Could not setup DIVE!");
    return FALSE;
  }
  // Setup some variables:
  iGameState = GAMESTATE_POSITION_SELF;
  fFirstTimeRender = TRUE;  // Make sure the new background will be drawn!
  bVolumePanelActive = FALSE; // No volume panel at startup.

  memset(LocalTable, TABLEENTRY_EMPTY, sizeof(LocalTable));
  memset(EnemyTable, TABLEENTRY_EMPTY, sizeof(EnemyTable));
  memset(LocalTableMiscData, 0, sizeof(LocalTableMiscData));
  memset(EnemyTableMiscData, 0, sizeof(EnemyTableMiscData));

  {
    int i;
    for (i=0; i<NUM_OF_SHIPS; i++)
    {
      ShipPosition[i][0] = ShipPosition[i][1] = -1;
      ShipVisibility[i]=0;
      ShipNumOfHits[i]=0;
    }
  }

  ulTime = ulRound = ulScore = 0;

  MessagePos = Skin->BackgroundImage->ulHeight;
  ActualMessage = Skin->msgWelcome; MessageBoard_NewMessage=TRUE;

  // Now that everything could be done, setup a new thread that will
  // render frames at request

#ifdef DEBUG_BUILD
  printf("beginthread for FrameRendererThread...\n");
  fflush(stdout);
#endif

  FrameRendererThread = _beginthread( FrameRenderer,
				      NULL,
				      65536, // 64K stack!
				      NULL);
  if (FrameRendererThread==-1)
  { // Could not create new thread!
    ShutdownDIVE();
    DestroyMainWindow();
    DosCloseMutexSem(hmtxUseDive);
    DosCloseMutexSem(hmtxUseSkin);
    DosCloseEventSem(hevBlitNeededEvent);
    CleanupSkin(&Skin);
    SFXCloseDevice();
    UninitializeFonts();
    UninitializeSprites();
    UninitializeImages();
    ErrorWindow("BattleShips - Error",
                "Could not create FrameRenderer thread!");
    return FALSE;
  }
  DosSetPriority(PRTYS_THREAD, PRTYC_REGULAR, 1, FrameRendererThread); // Burst the priority of renderer thread
  DosSetPriority(PRTYS_THREAD, PRTYC_REGULAR, 2, 0);                   // Burst the prio. of this (PM message processor) thread

#ifdef DEBUG_BUILD
  printf("beginthread for NetworkReceiverThread...\n");
  fflush(stdout);
#endif
  NetworkReceiverThread = _beginthread( NetworkReceiver,
                                        NULL,
					16384,
					NULL);
  if (NetworkReceiverThread==-1)
  { // Could not create new thread!
    DosKillThread(FrameRendererThread);
    ShutdownDIVE();
    DestroyMainWindow();
    DosCloseMutexSem(hmtxUseDive);
    DosCloseMutexSem(hmtxUseSkin);
    DosCloseEventSem(hevBlitNeededEvent);
    CleanupSkin(&Skin);
    SFXCloseDevice();
    UninitializeFonts();
    UninitializeSprites();
    UninitializeImages();
    ErrorWindow("BattleShips - Error",
                "Could not create NetworkReceiver thread!");
    return FALSE;
  }
  DosSetPriority(PRTYS_THREAD, PRTYC_REGULAR, 3, NetworkReceiverThread); // Set the priority of renderer thread

#ifdef DEBUG_BUILD
  printf("Initializing frame regulation...\n");
  fflush(stdout);
#endif

  // Create a thread that will post the hevBlitNeeded semaphore
  if (!InitializeFrameRegulation(hevBlitNeededEvent))
  {
    DosKillThread(NetworkReceiverThread);
    DosKillThread(FrameRendererThread);
    ShutdownDIVE();
    DestroyMainWindow();
    DosCloseMutexSem(hmtxUseDive);
    DosCloseMutexSem(hmtxUseSkin);
    DosCloseEventSem(hevBlitNeededEvent);
    CleanupSkin(&Skin);
    SFXCloseDevice();
    UninitializeFonts();
    UninitializeSprites();
    UninitializeImages();
    ErrorWindow("BattleShips - Error",
                "Could not initialize frame regulator thread!");
    return FALSE;
  }
  SetupFrameRegulation(IMAGE_REFRESH_FPS); // Set refresh rate to 30 FPS (or whatever the default is)

  // Ok, everything is set up, let's start the background music!
  if (Skin)
    SFXPlayFX(Skin->BackgroundMusic);

#ifdef DEBUG_BUILD
  printf("Entering to message loop\n");
  fflush(stdout);
#endif

  // Ok, now process messages until the window will be closed!
  while (WinGetMsg(Ab, (PQMSG) &Msg, 0, 0, 0))
  {
    if (Msg.msg == WM_CHANGECONFIG)
      ChangeSkin(); // This must be called from this thread, because it unloads a DLL, loaded in this thread!
    else
      WinDispatchMsg(Ab, (PQMSG) &Msg);
  }

#ifdef DEBUG_BUILD
  printf("Leaving message loop\n");
  fflush(stdout);
#endif

  // Stop the background music!
  if (Skin)
    SFXStopFX(Skin->BackgroundMusic);

  // Signal shutdown to frame renderer thread (and all others that take care of fShuttingDown flag)
  fShuttingDown=TRUE;
  DosPostEventSem(hevBlitNeededEvent); // Make sure it will get the signal

  // Stop network receiver thread
  DosKillThread(NetworkReceiverThread);

  // Stop image update
  UninitializeFrameRegulation();

  // Wait for the thread to be dead
  DosWaitThread(&FrameRendererThread, DCWW_WAIT);

  // Cleanup:
  ShutdownDIVE();
  DestroyMainWindow();
  DosCloseMutexSem(hmtxUseDive);
  DosCloseMutexSem(hmtxUseSkin);
  DosCloseEventSem(hevBlitNeededEvent);
  CleanupSkin(&Skin);

  // Cleanup the subsystems:
  // (In reverse order of initialization!)
  SFXCloseDevice();

  UninitializeFonts();
  UninitializeSprites();
  UninitializeImages();

  // Save configuration
  SaveConfig(&CurrentConfig);

#ifdef DEBUG_BUILD
  printf("StartGame() [NORMAL OUT]\n");
  fflush(stdout);
#endif

  // All done, return!
  return TRUE;
}

//---------------------------------------------------------------------
// StartGame_ClientMode
//
// The NetGame Client calls this when the connection has been made, and
// it's time to start the game. This function is called when this side
// acts as the client in this game (we're the invited one).
//---------------------------------------------------------------------
BOOL FAR __syscall __export StartGame_ClientMode( ULONG BufSize, pStartClientBuf Buffer )
{
  fWeAreTheServer=FALSE;
  return StartGame_General(BufSize, (pStartClientBuf) Buffer);
}

//---------------------------------------------------------------------
// StartGame_ServerMode
//
// The NetGame Client calls this when the connection has been made, and
// it's time to start the game. This function is called when this side
// acts as the server in this game (we're the invitor).
//---------------------------------------------------------------------
BOOL FAR __syscall __export StartGame_ServerMode( ULONG BufSize, pStartServerBuf Buffer )
{
  fWeAreTheServer=TRUE;
  return StartGame_General(BufSize, (pStartClientBuf) Buffer);
}

//---------------------------------------------------------------------
// InitializeGameDLL
//
// NetGame Client calls this when the game is about to begin. It pro-
// vides some helper functions that can be used to communicate through
// TCP/IP (but it's not obligatory to use them). All we do here is that
// we save the addresses of these helper functions for later use,
// and check if we will be able to run on this configuration.
//---------------------------------------------------------------------
BOOL FAR __syscall __export InitializeGameDLL( ULONG BufSize, pInitBuf Buffer )
{
#ifdef DEBUG_BUILD
  printf("InitializeGameDLL() [IN]\n");
  fflush(stdout);
#endif

  // Check for valid buffer size
  if (BufSize!=sizeof(tInitBuf))
  {
    ErrorWindow("BattleShips - Internal error",
                "Invalid parameter for InitializeGameDLL!");
    return FALSE;
  }

  // Save helper functions
  pSendMessage = Buffer->SendMessage;
  pReceiveMessage = Buffer->ReceiveMessage;

  // Now check if we can run on this computer

  // (Basically it's a check for DIVE in our case)
  if (!InspectDIVE())
  {
    ErrorWindow("BattleShips - Error",
                "DIVE not installed, or could not find requested video mode!");
    return FALSE;
  }

#ifdef DEBUG_BUILD
  printf("InitializeGameDLL() [NORMAL OUT]\n");
  fflush(stdout);
#endif

  // Return!
  return TRUE;
}

//---------------------------------------------------------------------
// SetupGame
//
// NetGame Client calls this when the user clicks on the Setup button
// for a given GameDLL. This function should process a dialog window,
// and save the new settings for this GameDLL.
//---------------------------------------------------------------------
BOOL FAR __syscall __export SetupGame( ULONG BufSize, pSetupGameBuf Buffer )
{
#ifdef DEBUG_BUILD
  printf("SetupGame() [IN]\n");
  fflush(stdout);
#endif

  // Check for valid buffer size
  if (BufSize!=sizeof(tSetupGameBuf))
  {
#ifdef DEBUG_BUILD
    printf("Wrong parameters! Expected: %d, got: %d\n", sizeof(tSetupGameBuf), BufSize);
    printf("Params: %d %d %d\n",Buffer->ab, Buffer->mq, Buffer->hmod);
    fflush(stdout);
#endif
    ErrorWindow("BattleShips - Internal error",
                "Invalid parameter for SetupGame!");
    return FALSE;
  }

  LoadConfig(&CurrentConfig);

#ifdef DEBUG_BUILD
  printf("Calling SetupWindowProcessing()!\n");
  fflush(stdout);
#endif

  if (!SetupWindowProcessing(Buffer->ab, Buffer->mq, HWND_DESKTOP, Buffer->hmod, SETUP_CALLED_FROM_NETGAMECLIENT))
#ifdef DEBUG_BUILD
  {
    printf(" Error in SetupWindowProcessing()!\n");
    printf("SetupGame() [OUT]\n");
    fflush(stdout);
#endif
    return 0;
#ifdef DEBUG_BUILD
  }
#endif

#ifdef DEBUG_BUILD
  printf("SetupGame() [OUT]\n");
  fflush(stdout);
#endif
  return TRUE;
}

/* END OF BShips.C */
