/*****************************************************************************
 * Module containing setup definitions and setup dialog window code          *
 *                                                                           *
 *****************************************************************************/

#include "Config.h"
#include <math.h>

#include "BSRes.h"      // Resource IDs
#include "SoundFX.h"    // Device query

// Linked list for available skin filenames
typedef struct _tSkinEntry
{
  char FileName[4096];
  ULONG ItemID;
  void *Next;
} tSkinEntry, *pSkinEntry;

HWND hwndSetupNB;   // Window handle of Setup notebook control
HWND hwndSoundNBP;  // Window handle of Sound notebook page
ULONG ulSoundNBPID; // ID of Sound notebook page
HWND hwndSkinNBP;   // Window handle of Skin notebook page
ULONG ulSkinNBPID;  // ID of Skin notebook page
HWND hwndSetupDlg;  // Window handle of main setup dialog window
int iSetupCalledFrom; // Stores where the setup is called from (NetGameClient, or from game...)

tConfig CurrentConfig; // Current configuration
tConfig NewConfig;     // New configuration, to change to!
BOOL bNewConfigAvailable = FALSE; // Flag to indicate when to change config! (usually when Apply button pressed)
char SoundTextUndoBuf[1024];  // Buffer to store old value of the current entry field.
PSFX_FX TestSound = NULL;     // Handle to the test sound being played!
pSkinEntry SkinListHead = NULL;

// ID of the setup timer
#define TMR_SETUPTIMER 0xd00d

void AddNotebookTab( HWND hwndNotebook, HWND hwndNotebookPage, USHORT usType, char *szTabText, char *szStatusText,
		     ULONG *pulPageID, unsigned short *pusTabTextLength, unsigned short *pusTabTextHeight )
{
  ULONG ulPageID;
  HWND hwndPage;
  PDLGTEMPLATE ppb;
  HPS hps;
  FONTMETRICS fmFontMetrics;
  POINTL aptlTabText[TXTBOX_COUNT];

  // Query dimensions of this tab text:
  hps = WinGetPS(hwndNotebook);
  memset(&fmFontMetrics, 0, sizeof(FONTMETRICS));
  if ((GpiQueryFontMetrics(hps, sizeof(FONTMETRICS), &fmFontMetrics)) &&
      (*pusTabTextHeight < fmFontMetrics.lMaxBaselineExt*2))
    *pusTabTextHeight = fmFontMetrics.lMaxBaselineExt*2;

  if ((GpiQueryTextBox(hps, strlen(szTabText), szTabText, TXTBOX_COUNT, aptlTabText)) &&
      (*pusTabTextLength < aptlTabText[TXTBOX_CONCAT].x))
    *pusTabTextLength = aptlTabText[TXTBOX_CONCAT].x;

  // Now add page
  *pulPageID = (ULONG) WinSendMsg( hwndNotebook,
                                 BKM_INSERTPAGE,
                                 0L,
                                 MPFROM2SHORT( (BKA_STATUSTEXTON | BKA_AUTOPAGESIZE | usType ),
					       BKA_LAST ) );
  // And set text for it
  WinSendMsg( hwndNotebook,
              BKM_SETTABTEXT,
              MPFROMLONG( *pulPageID ),
              MPFROMP( szTabText ) );
  WinSendMsg( hwndNotebook,
              BKM_SETSTATUSLINETEXT,
              MPFROMLONG( *pulPageID ),
              MPFROMP( szStatusText ) );
  WinSendMsg(hwndNotebook,
             BKM_SETPAGEWINDOWHWND,
             MPFROMP(*pulPageID),
             MPFROMP(hwndNotebookPage));
}

void AdjustControls(int NBPwidth, int NBPheight)
{
  SWP swpTemp;
  int xsize, ysize;
  RECTL rectl;
  POINTL pointl;
  int buttonheight;
  ULONG CXDLGFRAME = WinQuerySysValue(HWND_DESKTOP, SV_CXDLGFRAME);
  ULONG CYDLGFRAME = WinQuerySysValue(HWND_DESKTOP, SV_CYDLGFRAME);

  // Query button height in pixels
  if (WinQueryWindowPos(WinWindowFromID(hwndSetupDlg, PB_APPLY), &swpTemp))
    buttonheight = swpTemp.cy;
  else
    buttonheight = 0;

  // Set the size and position of Notebook

  // The size is interesting. We should know how much additional space this notebook requires.
  // For this, we calculate how big is a notebook for a known client size (e.g., 0x0)

  rectl.xLeft = 0; rectl.xRight = 0;
  rectl.yBottom = 0; rectl.yTop = 0;

  WinSendMsg(hwndSetupNB, BKM_CALCPAGERECT, MPFROMP(&rectl), MPFROMSHORT(FALSE));
  // Ok, this must be added to client size in order to be able to display the given client dialog:
  NBPwidth += rectl.xRight;
  NBPheight += rectl.yTop;

  // Now check how much shift is inside the window:
  // (client/notebook shift)
  pointl.x = 0;
  pointl.y = 0;
  WinMapWindowPoints(hwndSoundNBP, hwndSetupNB, &pointl, 1);
  // Add this too:
  NBPwidth += pointl.x;
  NBPheight += pointl.y;

  // Notebook position is above the buttons, which is some pixels above the bottom of window:
  swpTemp.x = CXDLGFRAME+2;
  swpTemp.y = CYDLGFRAME + buttonheight + 5;
  // Notebook size is just calculated:
  swpTemp.cx = NBPwidth;
  swpTemp.cy = NBPheight;
  WinSetWindowPos(hwndSetupNB, HWND_TOP,
		  swpTemp.x, swpTemp.y,
		  swpTemp.cx, swpTemp.cy,
		  SWP_MOVE | SWP_SIZE);

  // Ok, the notebook is resized. Now resize the dialog window to be just
  // as big as it's needed for the notebook:
  xsize = swpTemp.cx + CXDLGFRAME*2 + 2*2;
  ysize = swpTemp.cy + CYDLGFRAME*2 + buttonheight + 5 + WinQuerySysValue(HWND_DESKTOP, SV_CYTITLEBAR) + 5;
  WinSetWindowPos(hwndSetupDlg, HWND_TOP,
		  0, 0,
		  xsize,
                  ysize,
		  SWP_SIZE);

  // Set the position of buttons
  //  Apply Button
  WinQueryWindowPos(WinWindowFromID(hwndSetupDlg, PB_APPLY), &swpTemp);
  swpTemp.y = CYDLGFRAME+2;
  swpTemp.x = (xsize/4)-swpTemp.cx/2;
  WinSetWindowPos(WinWindowFromID(hwndSetupDlg, PB_APPLY), HWND_TOP,
		  swpTemp.x, swpTemp.y,
		  swpTemp.cx, swpTemp.cy,
		  SWP_MOVE);
  //  Cancel Button
  WinQueryWindowPos(WinWindowFromID(hwndSetupDlg, PB_CANCEL), &swpTemp);
  swpTemp.y = CYDLGFRAME+2;
  swpTemp.x = (xsize/4)*3-swpTemp.cx/2;
  WinSetWindowPos(WinWindowFromID(hwndSetupDlg, PB_CANCEL), HWND_TOP,
		  swpTemp.x, swpTemp.y,
		  swpTemp.cx, swpTemp.cy,
		  SWP_MOVE);

  // Now disable or enable the Test Sound button, depending on where is the setup initiated from!
  WinEnableControl(hwndSoundNBP, PB_TESTSELECTION, (iSetupCalledFrom == SETUP_CALLED_FROM_NETGAMECLIENT));

  // That's all!
}

MRESULT EXPENTRY fnSoundNotebookPage( HWND hwnd, USHORT msg, MPARAM mp1, MPARAM mp2 )
{
  char sTemp[256];
  int iTemp;
  switch (msg) {
    case WM_COMMAND:  // ------------------- WM_COMMAND messages ---------------------------
      switch SHORT1FROMMP(mp2) {
	case CMDSRC_PUSHBUTTON:           // - - - - - - - Source of WM_COMMAND is a PushButton - - - - - - -
	  switch (SHORT1FROMMP(mp1)) {
	    case PB_TESTSELECTION:                 // Test selection button pressed
	      if (SFXOpenDevice(NewConfig.SFXDeviceID,
				NewConfig.SFXDeviceShared,
				NewConfig.SFXFreq,
				NewConfig.SFXBits,
				NewConfig.SFXChannels,
				NewConfig.SFXNumBufs,
				NewConfig.SFXBufSize))
	      {
		TestSound = SFXLoadFX("BShips.dat\\TEST_SND.OGG", SFX_LOAD_MODE__STREAM);
                // Ok, file loaded, now play it!
                if (TestSound)
                {
                  WinEnableControl(hwndSoundNBP, PB_TESTSELECTION, FALSE);
                  WinEnableControl(hwndSetupDlg, PB_CLOSE, FALSE);

                  SFXSetFXMode(TestSound, SFX_PLAY_MODE__SINGLE);
                  SFXPlayFX(TestSound);
                }
                // If the test sound could not be loaded, it can be that
                // the test file is not there, or it is the nosound device
                // selected. So, do nothing in that case.
	      } else
		WinMessageBox(HWND_DESKTOP, hwndSetupDlg,
			      "Could not open the selected DART device!",
			      "Sound initialization error",
			      0,
			      MB_OK | MB_ERROR | MB_MOVEABLE);

	      return (MRESULT) TRUE;
	    default:
	      break;
	  }
	  break; // End of CMDSRC_PUSHBUTTON processing
	default:
	  break;
      } // end of switch inside WM_COMMAND
      break;

    case WM_CONTROL:  // ------------------- WM_CONTROL messages ---------------------------
      switch SHORT1FROMMP(mp1) {
	case EF_FREQUENCY:           // - - - - - - - Message from Frequency entry field - - - - - - -
	  switch (SHORT2FROMMP(mp1)) {
	    case EN_SETFOCUS:                 // Entered into entry field
	      // Save old text from here:
	      WinQueryDlgItemText(hwnd, SHORT1FROMMP(mp1),
				  1024, SoundTextUndoBuf);
	      break;
	    case EN_KILLFOCUS:                // Leaving entry field
	      {
		// Check the text there, and restore old one, if it's invalid!
		WinQueryDlgItemText(hwnd, SHORT1FROMMP(mp1),
				    256, sTemp);
		iTemp = atoi(sTemp);
		if ((iTemp>=MIN_ACCEPTABLE_FREQ) && (iTemp<=MAX_ACCEPTABLE_FREQ))
		{
		  NewConfig.SFXFreq = iTemp;
		  WinSetDlgItemText(hwnd, SHORT1FROMMP(mp1), (MPARAM) itoa(iTemp, sTemp, 10));
		}
                else
		  WinSetDlgItemText(hwnd, SHORT1FROMMP(mp1), SoundTextUndoBuf);
		// Calculate and show new lag:
                WinSetDlgItemText(hwnd, EF_AUDIOLAG, (MPARAM) itoa(CalculateCurrentLag(&NewConfig), sTemp, 10));

		break;
	      }
	    default:
	      break;
          }
          return (MPARAM) TRUE;
	  break; // End of EF_FREQUENCY processing
	case EF_NUMBUFS:             // - - - - - - - Message from NumBufs entry field - - - - - - -
	  switch (SHORT2FROMMP(mp1)) {
	    case EN_SETFOCUS:                 // Entered into entry field
	      // Save old text from here:
	      WinQueryDlgItemText(hwnd, SHORT1FROMMP(mp1),
				  1024, SoundTextUndoBuf);
	      break;
	    case EN_KILLFOCUS:                // Leaving entry field
	      {
		// Check the text there, and restore old one, if it's invalid!
		WinQueryDlgItemText(hwnd, SHORT1FROMMP(mp1),
				    256, sTemp);
		iTemp = atoi(sTemp);
		if ((iTemp>=MIN_ACCEPTABLE_NUMBUFS) && (iTemp<=MAX_ACCEPTABLE_NUMBUFS))
		{
		  NewConfig.SFXNumBufs = iTemp;
                  WinSetDlgItemText(hwnd, SHORT1FROMMP(mp1), (MPARAM) itoa(iTemp, sTemp, 10));
		}
                else
		  WinSetDlgItemText(hwnd, SHORT1FROMMP(mp1), SoundTextUndoBuf);
		// Calculate and show new lag:
                WinSetDlgItemText(hwnd, EF_AUDIOLAG, (MPARAM) itoa(CalculateCurrentLag(&NewConfig), sTemp, 10));

		break;
	      }
	    default:
	      break;
	  }
	  break; // End of EF_NUMBUFS processing
	case EF_BUFSIZE:             // - - - - - - - Message from BufSize entry field - - - - - - -
	  switch (SHORT2FROMMP(mp1)) {
	    case EN_SETFOCUS:                 // Entered into entry field
	      // Save old text from here:
	      WinQueryDlgItemText(hwnd, SHORT1FROMMP(mp1),
				  1024, SoundTextUndoBuf);
	      break;
	    case EN_KILLFOCUS:                // Leaving entry field
	      {
		// Check the text there, and restore old one, if it's invalid!
		WinQueryDlgItemText(hwnd, SHORT1FROMMP(mp1),
				    256, sTemp);
		iTemp = atoi(sTemp);
		if ((iTemp>=MIN_ACCEPTABLE_BUFSIZE) && (iTemp<=MAX_ACCEPTABLE_BUFSIZE))
		{
		  NewConfig.SFXBufSize = iTemp;
                  WinSetDlgItemText(hwnd, SHORT1FROMMP(mp1), (MPARAM) itoa(iTemp, sTemp, 10));
		}
                else
		  WinSetDlgItemText(hwnd, SHORT1FROMMP(mp1), SoundTextUndoBuf);
		// Calculate and show new lag:
                WinSetDlgItemText(hwnd, EF_AUDIOLAG, (MPARAM) itoa(CalculateCurrentLag(&NewConfig), sTemp, 10));

		break;
	      }
	    default:
	      break;
	  }
          break; // End of EF_NUMBUFS processing
          // -------------------------------- radio buttons --------------------

	case RB_SIXTEENBITS:
	  if (SHORT2FROMMP(mp1) == BN_CLICKED)
	  {
	    NewConfig.SFXBits = 16;
	    // Calculate and show new lag:
	    WinSetDlgItemText(hwnd, EF_AUDIOLAG, (MPARAM) itoa(CalculateCurrentLag(&NewConfig), sTemp, 10));
          }
          break;
	case RB_EIGHTBITS:
          if (SHORT2FROMMP(mp1) == BN_CLICKED)
	  {
	    NewConfig.SFXBits = 8;
	    // Calculate and show new lag:
	    WinSetDlgItemText(hwnd, EF_AUDIOLAG, (MPARAM) itoa(CalculateCurrentLag(&NewConfig), sTemp, 10));
          }
          break;
	case RB_STEREO:
	  if (SHORT2FROMMP(mp1) == BN_CLICKED)
	  {
	    NewConfig.SFXChannels = 2;
	    // Calculate and show new lag:
	    WinSetDlgItemText(hwnd, EF_AUDIOLAG, (MPARAM) itoa(CalculateCurrentLag(&NewConfig), sTemp, 10));
          }
          break;
	case RB_MONO:
          if (SHORT2FROMMP(mp1) == BN_CLICKED)
	  {
	    NewConfig.SFXChannels = 1;
	    // Calculate and show new lag:
	    WinSetDlgItemText(hwnd, EF_AUDIOLAG, (MPARAM) itoa(CalculateCurrentLag(&NewConfig), sTemp, 10));
          }
          break;
          // -------------------------------- Combo boxes --------------------

	case CX_SOUNDDEVICE:
	  if (SHORT2FROMMP(mp1) == LN_SELECT)
	  {
	    NewConfig.SFXDeviceID = (int) (WinSendDlgItemMsg(hwnd, SHORT1FROMMP(mp1),
                                                             LM_QUERYSELECTION, (MPARAM) LIT_CURSOR, 0))-1;
	  }
	  break;
          // -------------------------------- Checkboxes --------------------
	case CB_OPENSHARED:
	  if (SHORT2FROMMP(mp1) == BN_CLICKED)
	  {
	    NewConfig.SFXDeviceShared = WinQueryButtonCheckstate(hwnd, SHORT1FROMMP(mp1));
          }
          break;

        default:
	  break;
      } // end of switch inside WM_CONTROL
      break;
    default:
      break;
  } // end of switch of msg

  // Call the default procedure if it has not been processed yet!
  return WinDefDlgProc(hwnd, msg, mp1, mp2);
}

void SetNewSkinName(ULONG ItemID)
{
  pSkinEntry pTemp = SkinListHead;

  // Find this Skin ID:
  while ((pTemp) && (pTemp->ItemID != ItemID))
    pTemp = pTemp->Next;

  // If found set the filename from there:
  if (pTemp)
    strcpy(NewConfig.SkinFilename, pTemp->FileName);
}

MRESULT EXPENTRY fnSkinNotebookPage( HWND hwnd, USHORT msg, MPARAM mp1, MPARAM mp2 )
{
  char sTemp[256];
  int iTemp;
  switch (msg) {
    case WM_CONTROL:  // ------------------- WM_CONTROL messages ---------------------------
      switch SHORT1FROMMP(mp1) {
          // -------------------------------- Listboxes --------------------

	case LB_SKIN:
	  if (SHORT2FROMMP(mp1) == LN_SELECT)
          {
            // Get
	    ULONG ItemID = (ULONG) WinSendDlgItemMsg(hwnd, SHORT1FROMMP(mp1),
                                                     LM_QUERYSELECTION, (MPARAM) LIT_CURSOR, 0);

            if (ItemID != LIT_NONE)
              SetNewSkinName(ItemID);
	  }
	  break;
        default:
	  break;
      } // end of switch inside WM_CONTROL
      break;
    default:
      break;
  } // end of switch of msg

  // Call the default procedure if it has not been processed yet!
   return WinDefDlgProc(hwnd, msg, mp1, mp2);
}

MRESULT EXPENTRY SetupDlgProc( HWND hwnd, USHORT msg, MPARAM mp1, MPARAM mp2 )
{
  switch (msg) {
    case WM_INITDLG:
      // Create a timer, that will send us WM_TIMER messages in every second:
      WinStartTimer(WinQueryAnchorBlock(hwnd), hwnd, TMR_SETUPTIMER, 500);
      break;
    case WM_CLOSE:
      // Cleanup test sound, if it's being played!
      if (TestSound)
      {
        SFXUnloadFX(TestSound);
        SFXCloseDevice();
        WinEnableControl(hwndSoundNBP, PB_TESTSELECTION, TRUE);
        WinEnableControl(hwndSetupDlg, PB_CANCEL, TRUE);
        TestSound = NULL;
      }
      break;
    case WM_TIMER:
      if (SHORT1FROMMP(mp1)==TMR_SETUPTIMER)
      {
        if (TestSound)
        { // There is a test sound loaded in memory!
          // Check, if it's still being played, and cleanup if it's ended!
          if (SFXGetFXState(TestSound)!=SFX_STATE__PLAYING)
          {
            SFXUnloadFX(TestSound);
            SFXCloseDevice();
            WinEnableControl(hwndSoundNBP, PB_TESTSELECTION, TRUE);
            WinEnableControl(hwndSetupDlg, PB_CANCEL, TRUE);
            TestSound = NULL;
          }
        }
        return (MRESULT) TRUE;
      }
      break;
    case WM_COMMAND:
      switch SHORT1FROMMP(mp2) {
	case CMDSRC_PUSHBUTTON:           // ---- A WM_COMMAND from a pushbutton ------
	  switch (SHORT1FROMMP(mp1)) {
            case PB_APPLY:                               // Apply button pressed
#ifdef SETUP_DEBUG_BUILD
              printf("[SETUP] Apply button pressed!\n");
              fflush(stdout);
#endif
	      SaveConfig(&NewConfig);               // Save configuration
              bNewConfigAvailable = TRUE;
	      return (MRESULT) TRUE;
	    case PB_CANCEL:
	      if (SHORT1FROMMP(mp2)==CMDSRC_PUSHBUTTON)  // Cancel button pressed
	      {
		WinPostMsg(hwnd, WM_CLOSE, 0, 0);
		return (MRESULT) TRUE;
	      }
	      break;
	    default:
	      break;
	  }
	  break; // End os CMDSRC_PUSHBUTTON processing
      } // end of switch inside WM_COMMAND
      break;
    default:
      break;
  } // end of switch of msg

  // Call the default procedure if it has not been processed yet!
  return WinDefDlgProc(hwnd, msg, mp1, mp2);
}

int CalculateCurrentLag(pConfig Config)
{
  int iLag;

  if (!Config) return 0;

  iLag = Config->SFXNumBufs * Config->SFXBufSize;
  if (Config->SFXBits==16) iLag /= 2;
  iLag /= Config->SFXChannels;
  iLag = iLag*1000/Config->SFXFreq;
  return iLag;
}

void PutSkinDescriptionIntoListbox(char *filename)
{
  FILE  *skinfile;
  int rc;
  char textbuf[1024];
  char *key = (char *) malloc(1024);
  char *value = (char *) malloc(4096);
  char *skindesc = (char *) malloc(4096);
  int LineCount = 0;
  ULONG ItemID;
  pSkinEntry SkinEntry;

  if ((!filename) || (!key) || (!value) || (!skindesc))
  {
    free(key);
    free(value);
    free(skindesc);
    return;
  }

  // Create full path filename
  sprintf(textbuf, "BShips.dat\\%s", filename);

  // Create default skin description:
  sprintf(skindesc, "Skin file: %s", filename);

  // Open this text file!
  if ((skinfile=fopen(textbuf, "rt"))!=NULL)
  { // If we could open the file:
    // Load lines from it, and parse for "SkinName = Description" line!

    // Process every lines of the skin definition file, but
    // max. 1000, so if something goes wrong, wo won't end up an infinite
    // loop here.
    while ((!feof(skinfile)) && (LineCount<1000))
    {
      LineCount++;
      // First find the beginning of the key
      fscanf(skinfile, "%*[\n\t= ]");
      // Now get the key 
      rc=fscanf(skinfile, "%[^\n\t= ]",key);
      if ((rc==1) && (key[0]!=';'))
      { // We have the key! (and not a remark!)
        // Skip the whitespace after the key, if there are any
        fscanf(skinfile, "%*[ \t]");
        // There must be a '=' here
        rc=fscanf(skinfile, "%[=]", value);
        if (rc==1)
        {
          // Ok, the '=' is there
          // Skip the whitespace, if there are any
          fscanf(skinfile, "%*[ \t]");
          // and read anything until the end of line
          rc=fscanf(skinfile, "%[^\n]\n", value);
          if (rc==1)
          {
            // Cool, there is a key/value pair here, let's process it!
            if (!stricmp(key, "SKINNAME"))
            { // Here is the name of the skin
              strcpy(skindesc, value);
            }
          } else fscanf(skinfile, "%*[^\n]\n"); // find EOL
        } else fscanf(skinfile, "%*[^\n]\n"); // find EOL
      } else fscanf(skinfile, "%*[^\n]\n"); // find EOL
    }
    // Ok, skin file processed.
    fclose(skinfile);
  }
  free(key);
  free(value);

#ifdef SETUP_DEBUG_BUILD
  printf("  Skin description : %s\n", skindesc);
  fflush(stdout);
#endif

  // Now create a new item in the list of available skins:
  SkinEntry = (pSkinEntry) malloc(sizeof(tSkinEntry));
  if (SkinEntry)
  {
    // Fill values:
    strcpy(SkinEntry->FileName, filename);
    // Make sure the filename will *not* contain the .BSS extension!
    (SkinEntry->FileName)[strlen(SkinEntry->FileName)-4] = '\0';
    // Link it into the linked list:
    SkinEntry->Next = SkinListHead;
    SkinListHead = SkinEntry;
    // Ok, we have a description to insert into this listbox
    ItemID = (ULONG) WinSendDlgItemMsg(hwndSkinNBP, LB_SKIN, LM_INSERTITEM,
                                       (MPARAM) LIT_SORTASCENDING,
                                       (MPARAM) skindesc);
    // Associate the SkinEntry with this listbox item:
    WinSendDlgItemMsg(hwndSkinNBP, LB_SKIN, LM_SETITEMHANDLE,
                      (MPARAM) ItemID, (MPARAM) SkinEntry);
    SkinEntry->ItemID = ItemID;
  }

  free(skindesc);
}

void SetupControls()
{
  PSFX_AVAIL_DEVICES pDevs;
  char Buffer[2048];
  APIRET rc;
  HDIR hdirFindHandle;
  FILEFINDBUF3 FindBuffer;
  ULONG ulResultBufLen;
  ULONG ulFindCount;
  pSkinEntry pTemp;

  // Copy the current configuratio values into NewConfig!
  memcpy(&NewConfig, &CurrentConfig, sizeof(NewConfig));

  // Fill the Sound devices combo box, and set from the configuration:
  WinSendMsg(WinWindowFromID(hwndSoundNBP, CX_SOUNDDEVICE), LM_INSERTITEM,
             (MPARAM) LIT_END, (MPARAM) ("No Sound"));

  pDevs = SFXQueryAvailableDevices();
  if (pDevs)
  {
    int i;
    for (i=0; i<pDevs->iNumOfDevices; i++)
    {
      WinSendMsg(WinWindowFromID(hwndSoundNBP, CX_SOUNDDEVICE), LM_INSERTITEM,
		 (MPARAM) LIT_END, (MPARAM) ((pDevs->DevNames)+(pDevs->lDevNameIndex[i])));
#ifdef SETUP_DEBUG_BUILD
      printf("Dev %d: %s\n", i, (pDevs->DevNames)+(pDevs->lDevNameIndex[i]));
      fflush(stdout);
#endif
    }
    SFXDestroyAvailableDeviceList(pDevs);
    WinSendMsg(WinWindowFromID(hwndSoundNBP, CX_SOUNDDEVICE), LM_SELECTITEM,
		 (MPARAM) (CurrentConfig.SFXDeviceID+1), (MPARAM) (TRUE));
  }
  // Set Shared flag:
  WinSendMsg(WinWindowFromID(hwndSoundNBP, CB_OPENSHARED), BM_SETCHECK,
	     (MPARAM) CurrentConfig.SFXDeviceShared, (MPARAM) (0));
  // Set frequency:
  WinSetDlgItemText(hwndSoundNBP, EF_FREQUENCY, (MPARAM) itoa(CurrentConfig.SFXFreq, Buffer, 10));
  // Set Buffer size:
  WinSetDlgItemText(hwndSoundNBP, EF_BUFSIZE, (MPARAM) itoa(CurrentConfig.SFXBufSize, Buffer, 10));
  // Set number of buffers:
  WinSetDlgItemText(hwndSoundNBP, EF_NUMBUFS, (MPARAM) itoa(CurrentConfig.SFXNumBufs, Buffer, 10));
  // Calculate and set audio lag:
  WinSetDlgItemText(hwndSoundNBP, EF_AUDIOLAG, (MPARAM) itoa(CalculateCurrentLag(&CurrentConfig), Buffer, 10));
  // Number of bits:
  if (CurrentConfig.SFXBits == 16)
    WinSendMsg(WinWindowFromID(hwndSoundNBP, RB_SIXTEENBITS), BM_SETCHECK,
	       (MPARAM) 1, (MPARAM) (0));
  else
    WinSendMsg(WinWindowFromID(hwndSoundNBP, RB_EIGHTBITS), BM_SETCHECK,
	       (MPARAM) 1, (MPARAM) (0));
  // Number of channels:
  if (CurrentConfig.SFXChannels == 2)
    WinSendMsg(WinWindowFromID(hwndSoundNBP, RB_STEREO), BM_SETCHECK,
	       (MPARAM) 1, (MPARAM) (0));
  else
    WinSendMsg(WinWindowFromID(hwndSoundNBP, RB_MONO), BM_SETCHECK,
               (MPARAM) 1, (MPARAM) (0));

  // Now setup the skin selection page:
  hdirFindHandle = HDIR_CREATE;
  ulResultBufLen = sizeof(FindBuffer);
  ulFindCount = 1; // Request 1 result
  rc = DosFindFirst( "BShips.dat\\*.BSS",
                     &hdirFindHandle,
                     FILE_NORMAL,
                     &FindBuffer,
                     ulResultBufLen,
                     &ulFindCount,
                     FIL_STANDARD);
  if (rc==NO_ERROR)
  { // Ok, findbuf filled
    do {
#ifdef SETUP_DEBUG_BUILD
      printf("File: %s\n", FindBuffer.achName);
      fflush(stdout);
#endif
      // Read this file, and put the information about it into the listbox!
      PutSkinDescriptionIntoListbox(FindBuffer.achName);
      ulFindCount = 1; // Request one result
      rc = DosFindNext(hdirFindHandle,
                       &FindBuffer,
                       ulResultBufLen,
                       &ulFindCount);
    } while (rc == NO_ERROR);
    DosFindClose(hdirFindHandle);
  }

  // Ok, now that we've filled the listbox, we should find the
  // current skin in the list, and select that item in the listbox!

  pTemp = SkinListHead;
  while ((pTemp) && (stricmp(pTemp->FileName, CurrentConfig.SkinFilename)))
    pTemp=pTemp->Next;

  if (pTemp)
    WinSendDlgItemMsg(hwndSkinNBP, LB_SKIN, LM_SELECTITEM, (MPARAM) pTemp->ItemID, (MPARAM) TRUE);
}

int SetupWindowProcessing(HAB hAB, HMQ hMQ, HWND hwndParent, HMODULE hDLLHandle, int iCalledFrom)
{
  SWP swpTemp, swpParent;
  unsigned short usTabTextLength=0, usTabTextHeight=0;


#ifdef SETUP_DEBUG_BUILD
  printf("[SETUP] Loading dialog from DLL... (handle = %d)\n", hDLLHandle);
  fflush(stdout);
#endif

  iSetupCalledFrom = iCalledFrom;

  SkinListHead = NULL;

  hwndSetupDlg=WinLoadDlg(HWND_DESKTOP, hwndParent, (PFNWP)SetupDlgProc, hDLLHandle,
			   DLG_SETUPWINDOW, NULL);
  if (!hwndSetupDlg)
  {
#ifdef SETUP_DEBUG_BUILD
    printf("[SETUP] Could not load resource!\n");
    printf("[SETUP]  Error is 0x%x!\n", WinGetLastError(hAB));
    fflush(stdout);
#endif

//    ErrorWindow("BattleShips Warning", "Could not create dialog window for Setup!");
    return 0;
  }

#ifdef SETUP_DEBUG_BUILD
  printf("[SETUP] Dialog loaded, others come...\n");
  fflush(stdout);
#endif

  // Query the handle of the notebook control
  hwndSetupNB = WinWindowFromID(hwndSetupDlg, NBK_SETUPNOTEBOOK);

  // Now load and add the pages, one by one...
  hwndSoundNBP=WinLoadDlg(hwndSetupNB, hwndSetupNB, (PFNWP) fnSoundNotebookPage,
			  hDLLHandle, DLG_SOUNDSETUPDIALOG, NULL);
  hwndSkinNBP=WinLoadDlg(hwndSetupNB, hwndSetupNB, (PFNWP) fnSkinNotebookPage,
			 hDLLHandle, DLG_SKINSELECTIONDIALOG, NULL);
  // Save the size of one notebook page for later use.
  // (after adding, it will be resized, because of the AUDIOSIZE flag of notebook!)
  WinQueryWindowPos(hwndSoundNBP, &swpTemp);

  // Add pages to notebook:
  AddNotebookTab(hwndSetupNB, hwndSoundNBP, BKA_MAJOR, "Sound", "Sound setup", &ulSoundNBPID, &usTabTextLength, &usTabTextHeight);
  AddNotebookTab(hwndSetupNB, hwndSkinNBP, BKA_MAJOR, "Skin", "Skin selection", &ulSkinNBPID, &usTabTextLength, &usTabTextHeight);

  // The followings are discarder on Warp4, but makes things look better on Warp3:

  // Set notebook info field color (default is white... Yipe...)
  WinSendMsg(hwndSetupNB,
             BKM_SETNOTEBOOKCOLORS,
	     MPFROMLONG(SYSCLR_FIELDBACKGROUND), MPFROMSHORT(BKA_BACKGROUNDPAGECOLORINDEX));
  // Adjust tab size:
  WinSendMsg(hwndSetupNB,
             BKM_SETDIMENSIONS,
             MPFROM2SHORT(usTabTextLength + 10, (SHORT)((float)usTabTextHeight * 0.8)),
	     MPFROMSHORT(BKA_MAJORTAB));
  WinSendMsg(hwndSetupNB,
	     BKM_SETDIMENSIONS,
             MPFROM2SHORT(0, 0),
	     MPFROMSHORT(BKA_MINORTAB));


  // Adjust controls and window size
  AdjustControls(swpTemp.cx, swpTemp.cy);

  // Position the window in the middle of its parent!
  WinQueryWindowPos(hwndParent, &swpParent);
  WinQueryWindowPos(hwndSetupDlg, &swpTemp);
  swpTemp.x = (swpParent.cx-swpTemp.cx)/2 + swpParent.x;
  swpTemp.y = (swpParent.cy-swpTemp.cy)/2 + swpParent.y;
  WinSetWindowPos(hwndSetupDlg, HWND_TOP,
                  swpTemp.x, swpTemp.y,
                  0, 0,
                  SWP_MOVE);

  // Now everything is there and positioned, but empty.
  // Let's fill the fields, based on CurrentConfig!
  SetupControls();
  // As we will work in NewConfig, set that up too!
  memcpy(&NewConfig, &CurrentConfig, sizeof(NewConfig));

  // Show window
  WinShowWindow(hwndSetupDlg, TRUE);
  WinSetFocus(HWND_DESKTOP, hwndSetupNB);

  // Process messages
  WinProcessDlg(hwndSetupDlg);

  // Now cleanup
  WinDestroyWindow(hwndSetupDlg);

  while (SkinListHead)
  {
    pSkinEntry pTemp = SkinListHead;
    SkinListHead = SkinListHead->Next;
#ifdef SETUP_DEBUG_BUILD
    printf("Free: %s\n", pTemp->FileName);
    fflush(stdout);
#endif

    free(pTemp);
  }

  return 1;
}

void LoadConfig(pConfig Config) // Loads configuration into NewConfig(!), or loads default one (still into NewConfig).
{
  FILE *fp;
  BOOL bLoaded = FALSE;
  int iTemp;

  if (!Config) return;
  fp = fopen("BShips.cfg", "rt");
  if (fp != NULL)
  {
    iTemp = fscanf(fp, "%d %d %d %d %d %d %d %s\n%d %d %d",
                   &Config->SFXDeviceID,
                   &Config->SFXDeviceShared,
                   &Config->SFXFreq,
                   &Config->SFXBits,
                   &Config->SFXChannels,
                   &Config->SFXNumBufs,
                   &Config->SFXBufSize,
                   &Config->SkinFilename,
                   &Config->iMainVolume,
                   &Config->iMusicVolume,
                   &Config->iSFXVolume);

    if (iTemp==11)
    {
      if (Config->iMainVolume<0) Config->iMainVolume = 0;
      if (Config->iMainVolume>SFX_VOLUME__MAX) Config->iMainVolume = SFX_VOLUME__MAX;
      if (Config->iMusicVolume<0) Config->iMusicVolume = 0;
      if (Config->iMusicVolume>SFX_VOLUME__MAX) Config->iMusicVolume = SFX_VOLUME__MAX;
      if (Config->iSFXVolume<0) Config->iSFXVolume = 0;
      if (Config->iSFXVolume>SFX_VOLUME__MAX) Config->iSFXVolume = SFX_VOLUME__MAX;
      bLoaded = TRUE;
    }

    fclose(fp);
  }

  if (!bLoaded)
  {
    // Set defaults
    Config->SFXDeviceID = 0;
    Config->SFXDeviceShared = 1;
    Config->SFXFreq = 44100;
    Config->SFXBits = 16;
    Config->SFXChannels = 2;
    Config->SFXNumBufs = 2;
    Config->SFXBufSize = 8192;
    Config->iMainVolume = SFX_VOLUME__MAX;
    Config->iMusicVolume = SFX_VOLUME__MAX*2/3;
    Config->iSFXVolume = SFX_VOLUME__MAX;
    strcpy(Config->SkinFilename, "Default");
    // And save this default configuration!
    SaveConfig(Config);
  }
}

void SaveConfig(pConfig Config) // Saves current configuration
{
  FILE *fp;

  if (!Config) return;
  fp = fopen("BShips.cfg", "wt");
  if (fp != NULL)
  {
    fprintf(fp, "%d %d %d %d %d %d %d %s\n%d %d %d",
            Config->SFXDeviceID,
            Config->SFXDeviceShared,
            Config->SFXFreq,
            Config->SFXBits,
            Config->SFXChannels,
            Config->SFXNumBufs,
            Config->SFXBufSize,
            Config->SkinFilename,
            Config->iMainVolume,
            Config->iMusicVolume,
            Config->iSFXVolume);
    fclose(fp);
  }
}

