
/*****************************************************************************
 * OGG decoder module for the SoundFX library                                *
 *                                                                           *
 * Header file                                                               *
 *                                                                           *
 * This library is a wrapper for the integer-only OGG decoder, called TREMOR *
 *****************************************************************************/
/*
              OGG decoder module for SFX engine
 */
#ifndef OGGMODULE_H
#define OGGMODULE_H

// Definitions:

typedef void *OGGHANDLE;

// OGGOpen flags:
#define OPEN_OGG__FILESTREAM   0
#define OPEN_OGG__MEMORYSTREAM 1

#define DECODE_OGG__SIMPLE       0
#define DECODE_OGG__LOOPED       1

// Functions:
// Most functions return 1 for success, 0 for failure!

// OGGUninit
//  Uninitialize OGG decoder module
int OGGUninit();

// OGGInit
//  Initialize OGG decoder module
int OGGInit();

// OGGTest
//  Checks if a file is an OGG file or not.
int OGGTest(char *filename);

// OGGOpen
//  Tries to open an OGG file. Also loads into
//  memory if requested (see flags).
OGGHANDLE OGGOpen(char *filename, int iOpenMode);

// OGGClose
//  Closes an OGG file, or frees memory if it
//  was opened as MEMORY_STREAM...
int OGGClose(OGGHANDLE Handle);

// OGGGetSampleRate
//  Gets the samplerate of an opened OGG stream
int OGGGetSampleRate(OGGHANDLE Handle);

// OGGGetChannels
//  Gets the number of channels in an opened OGG stream
int OGGGetChannels(OGGHANDLE Handle);

// OGGSeek
//  Seeks in file. Actually, it is only used
//  to seek to the beginning of the file!
int OGGSeek(OGGHANDLE Handle, long lPos);

// OGGRead
//  Decodes ulBufSize bytes into pBuffer. Loops the stream
//  if it's said so (iMode), or fills the remaining buffer
//  with zeros.
int OGGDecode(OGGHANDLE Handle, int iMode, unsigned char *pBuffer, unsigned long ulBufSize);

// OGGTellPos
//  Returns the current position of OGG file, in milliseconds
__int64 OGGTellPos(OGGHANDLE Handle);

// OGGTellLength
//  Returns the length of OGG file, in milliseconds
__int64 OGGTellLength(OGGHANDLE Handle);

#endif
