/*****************************************************************************
 * Sound effect library, using DART                                          *
 *                                                                           *
 * This library can be used to load and play sound effects or musics through *
 * the DART interface of OS/2.                                               *
 * Currently only OGG file format is supported.                              *
 *****************************************************************************/
#define INCL_TYPES
#define INCL_WIN
#define INCL_DOS
#define INCL_DOSERRORS
#define INCL_GPI
#include <os2.h>
#define INCL_OS2MM
#include <os2me.h>  // DART stuff
#include <limits.h> // SHRT_MAX/SHRT_MIN, SCHAR_MAX/SCHAR_MIN

// Include Sound FX format decoders:
#include "OGGModule.h"
// #include "MP3Module.h"
// #include "WAVModule.h"
// #include "PCMModule.h"

// Include our header file, and structure definitions!
#include "SoundFX.h"

// Local variables:
MCI_BUFFER_PARMS BufferParms;     // Sound buffer parameters
MCI_MIX_BUFFER *pMixBuffers;      // Sound buffers
MCI_MIXSETUP_PARMS MixSetupParms; // Mixer setup parameters
TID DARTFeederThread;             // Thred ID of DART feeder thread
HEV hevAudioBufferPlayed;         // Event semaphore to indicate that an audio buffer has been played by DART
HMTX hmtxUseSFXList;              // Mutex semaphore to protect the list of SFXes
int fSFXRunning = FALSE;          // Flag to indicate shutdown of SFX engine
PSFX_FX psfxHead;                 // Head of list of loaded SFXes

// Current dart state:
int iSFXDeviceOrdNow;
int iSFXFreqNow;
int iSFXBitsNow;
int iSFXChannelsNow;
int iSFXNumBufsNow;
int iSFXBufSizeNow;

// Current main volume:
int iSFXMainVolume;

// Buffer states:
#define BUFFER_EMPTY       0
#define BUFFER_USED        1

//---------------------------------------------------------------------
// DARTEventFunc
//
// This function is called by DART, when an event occures, like end of 
// playback of a buffer, etc...
//---------------------------------------------------------------------
LONG APIENTRY DARTEventFunc(ULONG ulStatus,
			    PMCI_MIX_BUFFER pBuffer,
			    ULONG ulFlags)
{
  if (ulFlags && MIX_WRITE_COMPLETE)
  { // Playback of buffer completed!

    // Set the buffer to be empty
    (*pBuffer).ulUserParm = BUFFER_EMPTY;
    // And notify DART feeder thread that it will have to work a bit.
    DosPostEventSem(hevAudioBufferPlayed);
  }
  return TRUE;
}

//---------------------------------------------------------------------
// MixOneChannel_* functions
//
// These functions mix one channel into a DART buffer. There are
// different routines when the DART buffer has 16 bits or 8 bits 
// samples, and even more when the channel to be mixes is 16 bits or
// 8 bits.
//---------------------------------------------------------------------
void MixOneChannel_16into16(signed short *pBuffer, PSFX_FX pFX)
{
  // Mix a 16bits audio data into a 16bits audio buffer!
  signed short *pAudioData = (signed short *)(pFX->DecodeBuf);
  int iBufSize = iSFXBufSizeNow/2; // Buffer size in INT's!
  int iFinalVolume = (pFX->iVolume) * iSFXMainVolume;
  int iChnFreq   = pFX->iSampleRate;
  int iChnChannels = pFX->iChannels;
  int iFreqCounter = 0;

  if (iSFXChannelsNow==2)
  { // Output is stereo
    if (iChnChannels == 1)
    { // Convert from MONO to STEREO!
      int i=0;
      int iTemp, iTempCore;
      while (i<iBufSize)
      {
	iTempCore = (((int)(pAudioData[0]) * iFinalVolume) >> (16));
        iTemp = iTempCore + pBuffer[i];
	if (iTemp>SHRT_MAX) iTemp = SHRT_MAX;
        else
        if (iTemp<SHRT_MIN) iTemp = SHRT_MIN;
	pBuffer[i++] = iTemp; // Left side

        iTemp = iTempCore + pBuffer[i];
	if (iTemp>SHRT_MAX) iTemp = SHRT_MAX;
        else
        if (iTemp<SHRT_MIN) iTemp = SHRT_MIN;
	pBuffer[i++] = iTemp; // Right side

	iFreqCounter += iChnFreq; // Step in source buffer
	while (iFreqCounter>=iSFXFreqNow)
	{
	  pAudioData ++;
          iFreqCounter-=iSFXFreqNow;
	}
      }
    } else
    { // Make stereo from a stream of at least 2 channels!
      int i=0;
      int iTemp;
      while (i<iBufSize)
      {
	iTemp = (((int)(pAudioData[0]) * iFinalVolume) >> (16)) + pBuffer[i];
	if (iTemp>SHRT_MAX) iTemp = SHRT_MAX;
        else
        if (iTemp<SHRT_MIN) iTemp = SHRT_MIN;
	pBuffer[i++] = iTemp; // Left side
        iTemp = (((int)(pAudioData[1]) * iFinalVolume) >> (16)) + pBuffer[i];
	if (iTemp>SHRT_MAX) iTemp = SHRT_MAX;
        else
        if (iTemp<SHRT_MIN) iTemp = SHRT_MIN;
	pBuffer[i++] = iTemp; // Right side

	iFreqCounter += iChnFreq; // Step in source buffer
	while (iFreqCounter>=iSFXFreqNow)
	{
	  pAudioData += iChnChannels;
          iFreqCounter-=iSFXFreqNow;
	}
      }
    }
  }
  else
  if (iSFXChannelsNow==1)
  { // Output is mono
    if (iChnChannels>1)
    { // Convert to MONO from a multichannel format!
      int i=0;
      int iTemp;
      while (i<iBufSize)
      {
	iTemp = (((((int)(pAudioData[0]) * iFinalVolume) >> (16)) +
                 (((int)(pAudioData[1]) * iFinalVolume) >> (16))) /2) + pBuffer[i];
	if (iTemp>SHRT_MAX) iTemp = SHRT_MAX;
        else
        if (iTemp<SHRT_MIN) iTemp = SHRT_MIN;
	pBuffer[i++] = iTemp; // Mono output

	iFreqCounter += iChnFreq; // Step in source buffer
	while (iFreqCounter>=iSFXFreqNow)
	{
	  pAudioData += iChnChannels;
          iFreqCounter-=iSFXFreqNow;
	}
      }
    }
    else
    { // Make mono from a mono stream
      int i=0;
      int iTemp;
      while (i<iBufSize)
      {
	iTemp = (((int)(pAudioData[0]) * iFinalVolume) >> (16)) + pBuffer[i];
	if (iTemp>SHRT_MAX) iTemp = SHRT_MAX;
        else
        if (iTemp<SHRT_MIN) iTemp = SHRT_MIN;
	pBuffer[i++] = iTemp; // Mono output

	iFreqCounter += iChnFreq; // Step in source buffer
	while (iFreqCounter>=iSFXFreqNow)
	{
	  pAudioData ++;
          iFreqCounter-=iSFXFreqNow;
	}
      }
    } // else of if (iChnChannels>1)
  } // if (iSFXChannelsNow == 1)
} // void MixOneChannel

void MixOneChannel_8into16(signed short *pBuffer, PSFX_FX pFX)
{
  // Mix an unsigned 8bits audio data into a signed 16bits audio buffer!
  // NOT TESTED YET!
  unsigned char *pAudioData = (unsigned char *)(pFX->DecodeBuf);
  int iBufSize = iSFXBufSizeNow/2; // Buffer size in INT's!
  int iFinalVolume = (pFX->iVolume) * iSFXMainVolume;
  int iChnFreq   = pFX->iSampleRate;
  int iChnChannels = pFX->iChannels;
  int iFreqCounter = 0;

  if (iSFXChannelsNow==2)
  { // Output is stereo
    if (iChnChannels == 1)
    { // Convert from MONO to STEREO!
      int i=0;
      int iTemp, iTempCore;
      while (i<iBufSize)
      {
	iTempCore = (((int)(pAudioData[0]-128) * iFinalVolume) >> (8));
        iTemp = iTempCore + pBuffer[i];
	if (iTemp>SHRT_MAX) iTemp = SHRT_MAX;
        else
        if (iTemp<SHRT_MIN) iTemp = SHRT_MIN;
	pBuffer[i++] = iTemp; // Left side

        iTemp = iTempCore + pBuffer[i];
	if (iTemp>SHRT_MAX) iTemp = SHRT_MAX;
        else
        if (iTemp<SHRT_MIN) iTemp = SHRT_MIN;
	pBuffer[i++] = iTemp; // Right side

	iFreqCounter += iChnFreq; // Step in source buffer
	while (iFreqCounter>=iSFXFreqNow)
	{
	  pAudioData ++;
          iFreqCounter-=iSFXFreqNow;
	}
      }
    } else
    { // Make stereo from a stream of at least 2 channels!
      int i=0;
      int iTemp;
      while (i<iBufSize)
      {
	iTemp = (((int)(pAudioData[0]-128) * iFinalVolume) >> (8)) + pBuffer[i];
	if (iTemp>SHRT_MAX) iTemp = SHRT_MAX;
        else
        if (iTemp<SHRT_MIN) iTemp = SHRT_MIN;
	pBuffer[i++] = iTemp; // Left side
        iTemp = (((int)(pAudioData[1]-128) * iFinalVolume) >> (8)) + pBuffer[i];
	if (iTemp>SHRT_MAX) iTemp = SHRT_MAX;
        else
        if (iTemp<SHRT_MIN) iTemp = SHRT_MIN;
	pBuffer[i++] = iTemp; // Right side

	iFreqCounter += iChnFreq; // Step in source buffer
	while (iFreqCounter>=iSFXFreqNow)
	{
	  pAudioData += iChnChannels;
          iFreqCounter-=iSFXFreqNow;
	}
      }
    }
  }
  else
  if (iSFXChannelsNow==1)
  { // Output is mono
    if (iChnChannels>1)
    { // Convert to MONO from a multichannel format!
      int i=0;
      int iTemp;
      while (i<iBufSize)
      {
	iTemp = (((((int)(pAudioData[0]-128) * iFinalVolume) >> (8)) +
                 (((int)(pAudioData[1]-128) * iFinalVolume) >> (8))) /2) + pBuffer[i];
	if (iTemp>SHRT_MAX) iTemp = SHRT_MAX;
        else
        if (iTemp<SHRT_MIN) iTemp = SHRT_MIN;
	pBuffer[i++] = iTemp; // Mono output

	iFreqCounter += iChnFreq; // Step in source buffer
	while (iFreqCounter>=iSFXFreqNow)
	{
	  pAudioData += iChnChannels;
          iFreqCounter-=iSFXFreqNow;
	}
      }
    }
    else
    { // Make mono from a mono stream
      int i=0;
      int iTemp;
      while (i<iBufSize)
      {
	iTemp = (((int)(pAudioData[0]-128) * iFinalVolume) >> (8)) + pBuffer[i];
	if (iTemp>SHRT_MAX) iTemp = SHRT_MAX;
        else
        if (iTemp<SHRT_MIN) iTemp = SHRT_MIN;
	pBuffer[i++] = iTemp; // Mono output

	iFreqCounter += iChnFreq; // Step in source buffer
	while (iFreqCounter>=iSFXFreqNow)
	{
	  pAudioData ++;
          iFreqCounter-=iSFXFreqNow;
	}
      }
    } // else of if (iChnChannels>1)
  } // if (iSFXChannelsNow == 1)
}

void MixOneChannel_8into8(unsigned char *pBuffer, PSFX_FX pFX)
{
  // Mix unsigned 8bits audio data into an unsigned 8bits audio buffer!
  // NOT TESTED YET!
  unsigned char *pAudioData = (unsigned char *)(pFX->DecodeBuf);
  int iBufSize = iSFXBufSizeNow; // Buffer size in chars!
  int iFinalVolume = (pFX->iVolume) * iSFXMainVolume;
  int iChnFreq   = pFX->iSampleRate;
  int iChnChannels = pFX->iChannels;
  int iFreqCounter = 0;

  if (iSFXChannelsNow==2)
  { // Output is stereo
    if (iChnChannels == 1)
    { // Convert from MONO to STEREO!
      int i=0;
      int iTemp, iTempCore;
      while (i<iBufSize)
      {
	iTempCore = (((int)(pAudioData[0] - 128) * iFinalVolume) >> (16));
        iTemp = iTempCore + pBuffer[i] - 128;
	if (iTemp>SCHAR_MAX) iTemp = SCHAR_MAX;
        else
        if (iTemp<SCHAR_MIN) iTemp = SCHAR_MIN;
	pBuffer[i++] = iTemp + 128; // Left side

        iTemp = iTempCore + pBuffer[i] - 128;
	if (iTemp>SCHAR_MAX) iTemp = SCHAR_MAX;
        else
        if (iTemp<SCHAR_MIN) iTemp = SCHAR_MIN;
	pBuffer[i++] = iTemp + 128; // Right side

	iFreqCounter += iChnFreq; // Step in source buffer
	while (iFreqCounter>=iSFXFreqNow)
	{
	  pAudioData ++;
          iFreqCounter-=iSFXFreqNow;
	}
      }
    } else
    { // Make stereo from a stream of at least 2 channels!
      int i=0;
      int iTemp;
      while (i<iBufSize)
      {
	iTemp = (((int)(pAudioData[0]-128) * iFinalVolume) >> (16)) + pBuffer[i] - 128;
	if (iTemp>SCHAR_MAX) iTemp = SCHAR_MAX;
        else
        if (iTemp<SCHAR_MIN) iTemp = SCHAR_MIN;
	pBuffer[i++] = iTemp+128; // Left side
        iTemp = (((int)(pAudioData[1]-128) * iFinalVolume) >> (16)) + pBuffer[i] - 128;
	if (iTemp>SCHAR_MAX) iTemp = SCHAR_MAX;
        else
        if (iTemp<SCHAR_MIN) iTemp = SCHAR_MIN;
	pBuffer[i++] = iTemp+128; // Right side

	iFreqCounter += iChnFreq; // Step in source buffer
	while (iFreqCounter>=iSFXFreqNow)
	{
	  pAudioData += iChnChannels;
          iFreqCounter-=iSFXFreqNow;
	}
      }
    }
  }
  else
  if (iSFXChannelsNow==1)
  { // Output is mono
    if (iChnChannels>1)
    { // Convert to MONO from a multichannel format!
      int i=0;
      int iTemp;
      while (i<iBufSize)
      {
	iTemp = (((((int)(pAudioData[0]-128) * iFinalVolume) >> (16)) +
                 (((int)(pAudioData[1]-128) * iFinalVolume) >> (16))) /2) + pBuffer[i] - 128;
	if (iTemp>SCHAR_MAX) iTemp = SCHAR_MAX;
        else
        if (iTemp<SCHAR_MIN) iTemp = SCHAR_MIN;
	pBuffer[i++] = iTemp + 128; // Mono output

	iFreqCounter += iChnFreq; // Step in source buffer
	while (iFreqCounter>=iSFXFreqNow)
	{
	  pAudioData += iChnChannels;
          iFreqCounter-=iSFXFreqNow;
	}
      }
    }
    else
    { // Make mono from a mono stream
      int i=0;
      int iTemp;
      while (i<iBufSize)
      {
	iTemp = (((int)(pAudioData[0]-128) * iFinalVolume) >> (16)) + pBuffer[i] - 128;
	if (iTemp>SCHAR_MAX) iTemp = SCHAR_MAX;
        else
        if (iTemp<SCHAR_MIN) iTemp = SCHAR_MIN;
	pBuffer[i++] = iTemp+128; // Mono output

	iFreqCounter += iChnFreq; // Step in source buffer
	while (iFreqCounter>=iSFXFreqNow)
	{
	  pAudioData ++;
          iFreqCounter-=iSFXFreqNow;
	}
      }
    } // else of if (iChnChannels>1)
  } // if (iSFXChannelsNow == 1)
}

void MixOneChannel_16into8(unsigned char *pBuffer, PSFX_FX pFX)
{
  // Mix signed 16bits audio data into an unsigned 8bits audio buffer!
  signed short *pAudioData = (signed short *)(pFX->DecodeBuf);
  int iBufSize = iSFXBufSizeNow; // Buffer size in chars!
  int iFinalVolume = (pFX->iVolume) * iSFXMainVolume;
  int iChnFreq   = pFX->iSampleRate;
  int iChnChannels = pFX->iChannels;
  int iFreqCounter = 0;

  if (iSFXChannelsNow==2)
  { // Output is stereo
    if (iChnChannels == 1)
    { // Convert from MONO to STEREO!
      int i=0;
      int iTemp, iTempCore;
      while (i<iBufSize)
      {
	iTempCore = (((int)(pAudioData[0]) * iFinalVolume) >> (16+8));
        iTemp = iTempCore + pBuffer[i] - 128;
	if (iTemp>SCHAR_MAX) iTemp = SCHAR_MAX;
        else
        if (iTemp<SCHAR_MIN) iTemp = SCHAR_MIN;
	pBuffer[i++] = iTemp + 128; // Left side

        iTemp = iTempCore + pBuffer[i] - 128;
	if (iTemp>SCHAR_MAX) iTemp = SCHAR_MAX;
        else
        if (iTemp<SCHAR_MIN) iTemp = SCHAR_MIN;
	pBuffer[i++] = iTemp + 128; // Right side

	iFreqCounter += iChnFreq; // Step in source buffer
	while (iFreqCounter>=iSFXFreqNow)
	{
	  pAudioData ++;
          iFreqCounter-=iSFXFreqNow;
	}
      }
    } else
    { // Make stereo from a stream of at least 2 channels!
      int i=0;
      int iTemp;
      while (i<iBufSize)
      {
	iTemp = (((int)(pAudioData[0]) * iFinalVolume) >> (16+8)) + pBuffer[i] - 128;
	if (iTemp>SCHAR_MAX) iTemp = SCHAR_MAX;
        else
        if (iTemp<SCHAR_MIN) iTemp = SCHAR_MIN;
	pBuffer[i++] = iTemp+128; // Left side
        iTemp = (((int)(pAudioData[1]) * iFinalVolume) >> (16+8)) + pBuffer[i] - 128;
	if (iTemp>SCHAR_MAX) iTemp = SCHAR_MAX;
        else
        if (iTemp<SCHAR_MIN) iTemp = SCHAR_MIN;
	pBuffer[i++] = iTemp+128; // Right side

	iFreqCounter += iChnFreq; // Step in source buffer
	while (iFreqCounter>=iSFXFreqNow)
	{
	  pAudioData += iChnChannels;
          iFreqCounter-=iSFXFreqNow;
	}
      }
    }
  }
  else
  if (iSFXChannelsNow==1)
  { // Output is mono
    if (iChnChannels>1)
    { // Convert to MONO from a multichannel format!
      int i=0;
      int iTemp;
      while (i<iBufSize)
      {
	iTemp = (((((int)(pAudioData[0]) * iFinalVolume) >> (16+8)) +
                 (((int)(pAudioData[1]) * iFinalVolume) >> (16+8))) /2) + pBuffer[i] - 128;
	if (iTemp>SCHAR_MAX) iTemp = SCHAR_MAX;
        else
        if (iTemp<SCHAR_MIN) iTemp = SCHAR_MIN;
	pBuffer[i++] = iTemp + 128; // Mono output

	iFreqCounter += iChnFreq; // Step in source buffer
	while (iFreqCounter>=iSFXFreqNow)
	{
	  pAudioData += iChnChannels;
          iFreqCounter-=iSFXFreqNow;
	}
      }
    }
    else
    { // Make mono from a mono stream
      int i=0;
      int iTemp;
      while (i<iBufSize)
      {
	iTemp = (((int)(pAudioData[0]) * iFinalVolume) >> (16+8)) + pBuffer[i] - 128;
	if (iTemp>SCHAR_MAX) iTemp = SCHAR_MAX;
        else
        if (iTemp<SCHAR_MIN) iTemp = SCHAR_MIN;
	pBuffer[i++] = iTemp+128; // Mono output

	iFreqCounter += iChnFreq; // Step in source buffer
	while (iFreqCounter>=iSFXFreqNow)
	{
	  pAudioData ++;
          iFreqCounter-=iSFXFreqNow;
	}
      }
    } // else of if (iChnChannels>1)
  } // if (iSFXChannelsNow == 1)
}


//---------------------------------------------------------------------
// MixChannels
//
// Goes through the list of all loaded SFXes, and if their status is
// PLAYING, and its volume is greater than zero, then mixes it into the
// current DART buffer.
//---------------------------------------------------------------------
void MixChannels(unsigned char *pBuffer)
{
  PSFX_FX pFX;
  int rc;

  // Initialize buffer with zeros!
  if (iSFXBitsNow == 16) // signed
    memset(pBuffer, 0, iSFXBufSizeNow);
  else                // unsigned
    memset(pBuffer, 0x80, iSFXBufSizeNow);

  // If the main volume is bigger than zero, mix something,
  // be quiet otherwise!
  if (DosRequestMutexSem(hmtxUseSFXList, 1000)==NO_ERROR)
  {
    pFX = psfxHead;
    while (pFX)
    {
      // Decode one buffer of this FX
      if ((pFX) && (pFX->iState == SFX_STATE__PLAYING))
      {
        switch (pFX->iFileFormat)
        {
          case SFX_FORMAT__OGG:
            if (pFX->iPlayMode == SFX_PLAY_MODE__LOOP)
              rc = OGGDecode((OGGHANDLE) pFX->DataHandle, DECODE_OGG__LOOPED, pFX->DecodeBuf, pFX->DecodeBufSize);
            else
              rc = OGGDecode((OGGHANDLE) pFX->DataHandle, DECODE_OGG__SIMPLE, pFX->DecodeBuf, pFX->DecodeBufSize);
            // Stop this if there was an error!
            if (!rc)
            {
              pFX->iState = SFX_STATE__STOPPED;
              OGGSeek((OGGHANDLE) pFX->DataHandle, 0);
            }
            break;
          default:
            break;
        }
        // Ok, we have a buffer from this, pointers stepped.
        // Now mix it to the main buffer, if the volume is bigger than 0!
        if (pFX->iVolume)
        {
          if (iSFXBitsNow == 16)
          {
            if (pFX->iBits == 16) MixOneChannel_16into16((signed short *)pBuffer, pFX);
            else
            if (pFX->iBits == 8) MixOneChannel_8into16((signed short *)pBuffer, pFX);
          }
          else
          if (iSFXBitsNow == 8)
          {
            if (pFX->iBits == 16) MixOneChannel_16into8((unsigned char *)pBuffer, pFX);
            else
            if (pFX->iBits == 8) MixOneChannel_8into8((unsigned char *)pBuffer, pFX);
          }
        }
      } // if (pFX->iState == PLAYING)

      // Go to next fx
      pFX = pFX->NextSFX;
    }
    DosReleaseMutexSem(hmtxUseSFXList);
  } // if (DosRequestMutexSem())
}

//---------------------------------------------------------------------
// DARTFeederThreadFunc
//
// This is the DART feeder thread, that creates and queues audio data
// into DART.
//---------------------------------------------------------------------
VOID DARTFeederThreadFunc(PVOID Param)
{
  int i;
  ULONG ulPostCount;
#ifdef SFX_DEBUG_BUILD
  printf("Starting DARTFeeder thread\n");
  fflush(stdout);
#endif

  while (fSFXRunning)
  { // Now process messages!
    // Fill unused buffers with audio data, and send the buffers to DART
    for (i=0; i<iSFXNumBufsNow; i++)
    // Check all buffers, and if there is an empty one,
      if (pMixBuffers[i].ulUserParm == BUFFER_EMPTY)
      {
        // mix channels to fill this empty buffer,
	MixChannels(pMixBuffers[i].pBuffer);
	// set that it's filled now,
        pMixBuffers[i].ulUserParm = BUFFER_USED;
        // and send it to DART to be queued
	MixSetupParms.pmixWrite(MixSetupParms.ulMixHandle, &(pMixBuffers[i]), 1);
      }
    // Wait for a buffer to be empty
    if (DosWaitEventSem(hevAudioBufferPlayed, 1000 )==NO_ERROR) // Wait max 1 sec!
      DosResetEventSem(hevAudioBufferPlayed, &ulPostCount);
    else
      DosSleep(0);
  } // End of while (fShuttingDown)

#ifdef DEBUG_BUILD
  printf("Halting DARTFeeder thread\n");
  fflush(stdout);
#endif
  _endthread();
}

//---------------------------------------------------------------------
// SFXCalculateNeededBufSize
//
// It calculates how much audio data is needed for a given frequency,
// bit depth and channel number, so that when it is converted into the
// current DART format, it will be exactly that much data, that is the
// size of the DART buffers.
//---------------------------------------------------------------------
int SFXCalculateNeededBufSize(int iFreq, int iBits, int iChannels)
{
  __int64 result = iSFXBufSizeNow; // Get default DART buffer size

  result = (result * iBits)/iSFXBitsNow;
  result = (result * iFreq)/iSFXFreqNow;
  result = (result * iChannels)/iSFXChannelsNow;

  return (int)result;
}

//---------------------------------------------------------------------
// SFXDestroyAvailableDeviceList
//
// Destroys the list of available DART devices, allocated by the
// SFXQueryAvailableDevices() function.
//---------------------------------------------------------------------
void SFXDestroyAvailableDeviceList(PSFX_AVAIL_DEVICES ad)
{
  if (!ad) return;
  free(ad->lDevNameIndex);
  free(ad->DevNames);
  free(ad);
}

//---------------------------------------------------------------------
// SFXQueryAvailableDevices
//
// Queries the number of available DART devices, and makes a list of
// them. See the header file for the definition of 
// SFX_AVAILABLE_DEVICES structure!
//---------------------------------------------------------------------
PSFX_AVAIL_DEVICES SFXQueryAvailableDevices()
{
  MCI_SYSINFO_PARMS SysInfo;
  MCI_SYSINFO_LOGDEVICE SysInfoLogDevice;
  char Buffer[1024];
  int iNumOfDartDevices = 0;
  int rc;
  PSFX_AVAIL_DEVICES result = NULL;
  long lActStrPos;

  // First query the number of DART devices!

  memset(&SysInfo, 0, sizeof(SysInfo));
  memset(Buffer, 0, sizeof(Buffer));
  SysInfo.usDeviceType = MCI_DEVTYPE_AUDIO_AMPMIX;
  SysInfo.pszReturn = (PSZ) Buffer;
  SysInfo.ulRetSize = 1024;

  rc = mciSendCommand(0,                                // Don't know device ID yet
		      MCI_SYSINFO,
		      MCI_SYSINFO_QUANTITY | MCI_WAIT,
		      (PVOID) &SysInfo,
		      0);                               // No user parameter
  if (rc != MCIERR_SUCCESS)
    return NULL; // Could not query the number of DART devices

  iNumOfDartDevices = atoi(Buffer);
  if (iNumOfDartDevices)
  {
    int i;
    // Ok, there are devices to make a list from, so allocate memory for
    // the result!

    result = (PSFX_AVAIL_DEVICES) malloc(sizeof(SFX_AVAIL_DEVICES));
    if (!result) return NULL; // Could not allocate memory!

    result->iNumOfDevices = 0;
    result->lDevNameIndex = NULL;
    result->DevNames = NULL;
    lActStrPos = 0;

    // Go through all the devices:
    for (i = 0; i<=iNumOfDartDevices; i++)
    {
      // Get the install name of this Dart device
      memset(&SysInfo, 0, sizeof(SysInfo));
      memset(Buffer, 0, sizeof(Buffer));
      SysInfo.usDeviceType = MCI_DEVTYPE_AUDIO_AMPMIX;
      SysInfo.ulNumber = i;
      SysInfo.pszReturn = (PSZ) Buffer;
      SysInfo.ulRetSize = 1024;
      rc = mciSendCommand(0, // Device ID
			  MCI_SYSINFO,
			  MCI_SYSINFO_INSTALLNAME | MCI_WAIT,
			  (PVOID) &SysInfo,
			  0);
      if (rc == MCIERR_SUCCESS)
      {
        // Using the install name, we can query the description too:
	// Get the description for this install name
	memset(&SysInfo, 0, sizeof(SysInfo));
	memset(&SysInfoLogDevice, 0, sizeof(SysInfoLogDevice));
	strcpy(SysInfoLogDevice.szInstallName, Buffer);
  
	SysInfo.ulItem = MCI_SYSINFO_QUERY_DRIVER;
	SysInfo.pSysInfoParm = &SysInfoLogDevice;
	rc = mciSendCommand(0, // Device ID
			    MCI_SYSINFO,
			    MCI_SYSINFO_ITEM | MCI_WAIT,
			    (PVOID) &SysInfo,
			    0);
  
	if (rc == MCIERR_SUCCESS)
	{
	  if (i==0)
	    // Default device
	    sprintf(Buffer, "Default [%s (%s)]", SysInfoLogDevice.szProductInfo, SysInfoLogDevice.szVersionNumber);
	  else
            // Not the default device
	    sprintf(Buffer, "%s (%s)", SysInfoLogDevice.szProductInfo, SysInfoLogDevice.szVersionNumber);

          // Save this information in the result structure!
	  result->iNumOfDevices++;

       	  if (!(result->lDevNameIndex))
	    result->lDevNameIndex = (long *)malloc(result->iNumOfDevices*sizeof(long));
	  else
            result->lDevNameIndex = (long *)realloc(result->lDevNameIndex, result->iNumOfDevices*sizeof(long));

	  result->lDevNameIndex[result->iNumOfDevices-1] = lActStrPos;

          if (!(result->DevNames))
	    result->DevNames = (char *)malloc(strlen(Buffer)+1);
	  else
            result->DevNames = (char *)realloc(result->DevNames, lActStrPos+strlen(Buffer)+1);

	  strcpy(result->DevNames+lActStrPos, Buffer);
          lActStrPos+=strlen(Buffer)+1;
	}
      }
    }
  }
  return result;
}

#ifdef SFX_DEBUG_BUILD
void MciError(ULONG ulError)
{
  char szBuffer[ 2048 ];
  ULONG rc;
  rc = mciGetErrorString(ulError, szBuffer, 2048);
  printf("MCI Error: %s\n",szBuffer);
}
#endif

//---------------------------------------------------------------------
// SFXOpenDevice
//
// Opens the DART device (with the given ordinal, or 0 for the default)
// for playback. Then queries if that device is capable of handling the
// given sound format. If it is, sets up, allocates buffers, and starts
// the mixing thread and the playback.
//---------------------------------------------------------------------
int SFXOpenDevice(int iDeviceOrd, int bOpenShared, int iFreq, int iBits, int iChannels, int iNumBufs, int iBufSize)
{
  MCI_AMP_OPEN_PARMS AmpOpenParms;
  MCI_GENERIC_PARMS GenericParms;
  int iOpenMode;
  int rc;

  // If we're already running, do nothing, but return failure!
  if (fSFXRunning) return 0;

  // Initialize main volume:
  iSFXMainVolume = SFX_VOLUME__MAX;

  // If this is the device -1, it means, it is the NO_SOUND device.
  // In this case, simply return success!
  if (iDeviceOrd == -1)
  {
    // Store the new settings in global variables
    iSFXDeviceOrdNow = iDeviceOrd;
    iSFXFreqNow = iFreq;
    iSFXBitsNow = iBits;
    iSFXChannelsNow = iChannels;
    iSFXNumBufsNow = iNumBufs;
    iSFXBufSizeNow = iBufSize;

    return 1;
  }

  // First thing is to try to open a given DART device!
  memset(&AmpOpenParms, 0, sizeof(MCI_AMP_OPEN_PARMS));
  // pszDeviceType should contain the device type in low word, and device ordinal in high word!
  AmpOpenParms.pszDeviceType = (PSZ) (MCI_DEVTYPE_AUDIO_AMPMIX | (iDeviceOrd << 16));

  iOpenMode = MCI_WAIT | MCI_OPEN_TYPE_ID;
  if (bOpenShared) iOpenMode |= MCI_OPEN_SHAREABLE;
#ifdef SFX_DEBUG_BUILD
  printf("Opening DART, device %d\n", iDeviceOrd);
  fflush(stdout);
#endif
  rc = mciSendCommand( 0, MCI_OPEN,
                       iOpenMode,
		       (PVOID) &AmpOpenParms, 0);
  if (rc==MCIERR_SUCCESS)
  {
#ifdef SFX_DEBUG_BUILD
    printf("DART opened, now setting up  (device ID = %d)!\n", AmpOpenParms.usDeviceID);
    fflush(stdout);
#endif
    // Save the device ID we got from DART!
    // We will use this in the next calls!
    iDeviceOrd = AmpOpenParms.usDeviceID;

    // Now query this device if it supports the given freq/bits/channels!
    memset(&MixSetupParms, 0, sizeof(MCI_MIXSETUP_PARMS));
    MixSetupParms.ulBitsPerSample = iBits;
    MixSetupParms.ulFormatTag = MCI_WAVE_FORMAT_PCM;
    MixSetupParms.ulSamplesPerSec = iFreq;
    MixSetupParms.ulChannels = iChannels;
    MixSetupParms.ulFormatMode = MCI_PLAY;
    MixSetupParms.ulDeviceType = MCI_DEVTYPE_WAVEFORM_AUDIO;
    MixSetupParms.pmixEvent = DARTEventFunc;
    rc = mciSendCommand (iDeviceOrd, MCI_MIXSETUP,
			 MCI_WAIT | MCI_MIXSETUP_QUERYMODE,
			 &MixSetupParms, 0);
    if (rc!=MCIERR_SUCCESS)
    { // The device cannot handle this format!
      // Close DART, and exit with error code!
#ifdef SFX_DEBUG_BUILD
      printf("Cannot handle this format!\n");
      MciError(rc);
      fflush(stdout);
#endif
      mciSendCommand(iDeviceOrd, MCI_CLOSE, MCI_WAIT, &GenericParms, 0);
      return 0;
    }
    // The device can handle this format, so initialize!
    rc = mciSendCommand(iDeviceOrd, MCI_MIXSETUP,
			MCI_WAIT | MCI_MIXSETUP_INIT,
			&MixSetupParms, 0);
    if (rc!=MCIERR_SUCCESS)
    { // The device could not be opened!
      // Close DART, and exit with error code!
#ifdef SFX_DEBUG_BUILD
      printf("Error opening the DART device!\n");
      fflush(stdout);
#endif
      mciSendCommand(iDeviceOrd, MCI_CLOSE, MCI_WAIT, &GenericParms, 0);
      return 0;
    }
    // Ok, the device is initialized.
    // Now we should allocate buffers. For this, we need a place where
    // the buffer descriptors will be:
    pMixBuffers = (MCI_MIX_BUFFER *) malloc(sizeof(MCI_MIX_BUFFER)*iNumBufs);
    if (!pMixBuffers)
    { // Not enough memory!
      // Close DART, and exit with error code!
#ifdef SFX_DEBUG_BUILD
      printf("Error allocating memory for bufferlist!\n");
      fflush(stdout);
#endif
      mciSendCommand(iDeviceOrd, MCI_CLOSE, MCI_WAIT, &GenericParms, 0);
      return 0;
    }
    // Now that we have the place for buffer list, we can ask DART for the
    // buffers!
    BufferParms.ulNumBuffers = iNumBufs;   // Number of buffers
    BufferParms.ulBufferSize = iBufSize;   // each with this size
    BufferParms.pBufList = pMixBuffers;    // getting descriptorts into this list
    // Allocate buffers!
    rc = mciSendCommand(iDeviceOrd, MCI_BUFFER,
			MCI_WAIT | MCI_ALLOCATE_MEMORY,
			&BufferParms, 0);
    if ((rc!=MCIERR_SUCCESS) || (iNumBufs != BufferParms.ulNumBuffers) || (BufferParms.ulBufferSize==0))
    { // Could not allocate memory!
      // Close DART, and exit with error code!
#ifdef SFX_DEBUG_BUILD
      printf("Could not allocate memory for DART buffers!\n");
      fflush(stdout);
#endif
      free(pMixBuffers); pMixBuffers = NULL;
      mciSendCommand(iDeviceOrd, MCI_CLOSE, MCI_WAIT, &GenericParms, 0);
      return 0;
    }
    // Ok, we have all the buffers allocated, let's mark them!
    {
      int i;
      for (i=0; i<iNumBufs; i++)
      {
        // Check if this buffer was really allocated by DART
	if (!(pMixBuffers[i].pBuffer))
	{ // Wrong buffer!
	  // Close DART, and exit with error code!
#ifdef SFX_DEBUG_BUILD
	  printf("Wrong memory allocated for DART buffer %d!\n", i);
          fflush(stdout);
#endif
          mciSendCommand(iDeviceOrd, MCI_BUFFER, MCI_WAIT | MCI_DEALLOCATE_MEMORY, &BufferParms, 0);
          free(pMixBuffers); pMixBuffers = NULL;
          mciSendCommand(iDeviceOrd, MCI_CLOSE, MCI_WAIT, &GenericParms, 0);
          return 0;
	}
	pMixBuffers[i].ulBufferLength = BufferParms.ulBufferSize;
	pMixBuffers[i].ulUserParm = BUFFER_EMPTY; // User parameter: is it filled or not
	pMixBuffers[i].ulFlags = 0; // Some stuff should be flagged here for DART, like end of
	                            // audio data, but as we will continously send
	                            // audio data, there will be no end.:)
	memset(pMixBuffers[i].pBuffer, 0, iBufSize);
      }
    }
    // Ok, buffers ready and marked
#ifdef SFX_DEBUG_BUILD
    printf("SFXOpenDevice done!\n");
    fflush(stdout);
#endif
    // Initialize decoder modules
    if (!OGGInit())
    {
      // Could not initialize OGG module!
      mciSendCommand(iDeviceOrd, MCI_BUFFER, MCI_WAIT | MCI_DEALLOCATE_MEMORY, &BufferParms, 0);
      free(pMixBuffers); pMixBuffers = NULL;
      mciSendCommand(iDeviceOrd, MCI_CLOSE, MCI_WAIT, &GenericParms, 0);
      return 0;
    }
    // Right now, we only have the OGG decoder. If there will be more,
    // they should be initialized here.

    // All decoder modules initialized!
    psfxHead = NULL;
    fSFXRunning = TRUE;
    // Create event semaphore
    if (DosCreateEventSem(NULL, &hevAudioBufferPlayed, 0, FALSE)!=NO_ERROR)
    {
      // Could not create event semaphore!
      OGGUninit();
      mciSendCommand(iDeviceOrd, MCI_BUFFER, MCI_WAIT | MCI_DEALLOCATE_MEMORY, &BufferParms, 0);
      free(pMixBuffers); pMixBuffers = NULL;
      mciSendCommand(iDeviceOrd, MCI_CLOSE, MCI_WAIT, &GenericParms, 0);
      return 0;
    }
    // Create mutex semaphore
    if (DosCreateMutexSem(NULL, &hmtxUseSFXList, 0, FALSE)!=NO_ERROR)
    {
      // Could not create mutex semaphore!
      DosCloseEventSem(hevAudioBufferPlayed);
      OGGUninit();
      mciSendCommand(iDeviceOrd, MCI_BUFFER, MCI_WAIT | MCI_DEALLOCATE_MEMORY, &BufferParms, 0);
      free(pMixBuffers); pMixBuffers = NULL;
      mciSendCommand(iDeviceOrd, MCI_CLOSE, MCI_WAIT, &GenericParms, 0);
      return 0;
    }

    // Store the new settings in global variables
    iSFXDeviceOrdNow = iDeviceOrd;
    iSFXFreqNow = iFreq;
    iSFXBitsNow = iBits;
    iSFXChannelsNow = iChannels;
    iSFXNumBufsNow = iNumBufs;
    iSFXBufSizeNow = iBufSize;
    // Start dartfeeder thread
    DARTFeederThread = _beginthread( DARTFeederThreadFunc,
                                     NULL,
				     16384,
				     NULL);
    if (DARTFeederThread==-1)
    { // Could not create new thread!
#ifdef SFX_DEBUG_BUILD
      printf("Could not create DARTFeeder thread!\n");
      fflush(stdout);
#endif
      DosCloseMutexSem(hmtxUseSFXList);
      DosCloseEventSem(hevAudioBufferPlayed);
      OGGUninit();
      mciSendCommand(iDeviceOrd, MCI_BUFFER, MCI_WAIT | MCI_DEALLOCATE_MEMORY, &BufferParms, 0);
      free(pMixBuffers); pMixBuffers = NULL;
      mciSendCommand(iDeviceOrd, MCI_CLOSE, MCI_WAIT, &GenericParms, 0);
      return 0;
    }
    DosSetPriority(PRTYS_THREAD, PRTYC_FOREGROUNDSERVER, 28, DARTFeederThread); // Burst the priority
  }

  return (rc==MCIERR_SUCCESS);
}

//---------------------------------------------------------------------
// SFXCloseDevice
//
// Signals end to the dart feeder thread, then stops playback, frees
// audio buffers, closes DART, and frees allocated SFXes, if they are
// not freed yet.
//---------------------------------------------------------------------
int SFXCloseDevice()
{
  MCI_GENERIC_PARMS GenericParms;
  int rc;

  if (!fSFXRunning) return 1;

  // Stop DART playback
  rc = mciSendCommand(iSFXDeviceOrdNow, MCI_STOP, MCI_WAIT, &GenericParms, 0);
  if (rc!=MCIERR_SUCCESS)
  {
#ifdef SFX_DEBUG_BUILD
    printf("Could not stop DART playback!\n");
    fflush(stdout);
#endif
  }

  // Stop dartfeeder thread
  fSFXRunning = FALSE;
  // Wait for the thread to be dead
  DosWaitThread(&DARTFeederThread, DCWW_WAIT);

  // Close event semaphore
  DosCloseEventSem(hevAudioBufferPlayed);

  // Deallocate buffers
  rc = mciSendCommand(iSFXDeviceOrdNow, MCI_BUFFER, MCI_WAIT | MCI_DEALLOCATE_MEMORY, &BufferParms, 0);
  if (rc!=MCIERR_SUCCESS)
  {
#ifdef SFX_DEBUG_BUILD
    printf("Could not deallocate memory!\n");
    fflush(stdout);
#endif
  }

  // Free bufferlist
  free(pMixBuffers); pMixBuffers = NULL;

  // Close dart
  rc = mciSendCommand(iSFXDeviceOrdNow, MCI_CLOSE, MCI_WAIT, &GenericParms, 0);
  if (rc!=MCIERR_SUCCESS)
  {
#ifdef SFX_DEBUG_BUILD
    printf("Could not close DART device!\n");
    fflush(stdout);
#endif
  }

  // Unload sound effects
  while (psfxHead) SFXUnloadFX(psfxHead);

  // Now uninitialize decoder modules!
  OGGUninit();

  // Close mutex semaphore
  DosCloseMutexSem(hmtxUseSFXList);
#ifdef SFX_DEBUG_BUILD
  printf("SFXCloseDevice done\n");
  fflush(stdout);
#endif
  return 1;
}

//---------------------------------------------------------------------
// SFXLoadFX
//
// Checks a given file if its format is supported, and loads it. Loads
// the full file into memory, if iLoadMode is __MEMORY, and does not
// load, only on demand, if iLoadMode is __STREAM.
//---------------------------------------------------------------------
PSFX_FX SFXLoadFX(char *filename, int iLoadMode)
{
  PSFX_FX result = (PSFX_FX) malloc(sizeof(SFX_FX));
  int rc;

  if (!fSFXRunning) return NULL;

  if (result && filename)
  {
    // Check the format of this file!
    // Test with every decoder module!
    rc = OGGTest(filename);
    if (rc)
    {
      // Ok, this is an OGG file, let's open it!
      result->iFileFormat = SFX_FORMAT__OGG;     // Store format with SFX
      result->iPlayMode = SFX_PLAY_MODE__SINGLE; // Default mode is not looped
      result->iState = SFX_STATE__STOPPED;       // Default state: stopped
      result->DecodeBuf = NULL;

      // Open in filestream or memorystream mode!
      if (iLoadMode == SFX_LOAD_MODE__STREAM)
	result->DataHandle = (OGGHANDLE) OGGOpen(filename, OPEN_OGG__FILESTREAM);
      else
	result->DataHandle = (OGGHANDLE) OGGOpen(filename, OPEN_OGG__MEMORYSTREAM);
      if (result->DataHandle)
      { // If the stream could be loaded, allocate a buffer, big enough to
	// decode that much data that will fill exactly one DART buffer!
	result->iSampleRate = OGGGetSampleRate(result->DataHandle);
	result->iChannels = OGGGetChannels(result->DataHandle);
	result->iBits = 16; // OGG always decodes 16bits, signed data!

	result->iVolume = SFX_VOLUME__MAX; // Default volume

        result->DecodeBufSize = SFXCalculateNeededBufSize(result->iSampleRate, result->iBits, result->iChannels);
	result->DecodeBuf = (void *) malloc(result->DecodeBufSize);
	if (!result->DecodeBuf)
	{ // Could not allocate memory, cleanup!
	  OGGClose(result->DataHandle);
	  free(result);
          result = NULL;
	}
      }
    } else
    {
#ifdef SFX_DEBUG_BUILD
      printf("Unknown audio file format: %s\n", filename);
      fflush(stdout);
#endif
      free(result); result=NULL;
      return result;
    }

    if (result->DataHandle)
    { // Link it into the list of loaded sounds!
      if (DosRequestMutexSem(hmtxUseSFXList, 1000)==NO_ERROR)
      {
        result->NextSFX = psfxHead;
	result->PrevSFX = NULL;
        if (psfxHead)
   	  psfxHead->PrevSFX = result;
	psfxHead = result;
	DosReleaseMutexSem(hmtxUseSFXList);
      } else
      { // Could not get SFX list mutex!
	// Call its decoder to close the file:
	switch (result->iFileFormat)
	{
	  case SFX_FORMAT__OGG:
	    OGGClose(result->DataHandle);
	    break;
	  default:
#ifdef SFX_DEBUG_BUILD
	    printf("Internal error, wrong fileformat tag!\n");
	    fflush(stdout);
#endif
	    return 0;
	}
	// free its buffers, and return failure (NULL)
        free(result->DecodeBuf); result->DecodeBuf = NULL;
	free(result); result = NULL;
      }
    }
  }

  return result;
}

//---------------------------------------------------------------------
// SFXPauseFX
//
// Pauses the playback of a given SFX. (The part already decoded and 
// stored in the DART buffers will be played, though...)
//---------------------------------------------------------------------
int SFXPauseFX(PSFX_FX pFX)
{
  if (!fSFXRunning) return 1;
  if (!pFX) return 0;
  pFX->iState = SFX_STATE__STOPPED;
  return 1;
}

//---------------------------------------------------------------------
// SFXResumeFX
//
// Resumes the playback of a given SFX.
//---------------------------------------------------------------------
int SFXResumeFX(PSFX_FX pFX)
{
  if (!fSFXRunning) return 1;
  if (!pFX) return 0;
  pFX->iState = SFX_STATE__PLAYING;
  return 1;
}

//---------------------------------------------------------------------
// SFXPauseFX
//
// Pauses the playback of a given SFX, and "rewinds" the stream to its
// start position.
//---------------------------------------------------------------------
int SFXStopFX(PSFX_FX pFX)
{
  if (!fSFXRunning) return 1;

  if (!pFX) return 0;

  if (DosRequestMutexSem(hmtxUseSFXList, 1000)==NO_ERROR)
  {
    pFX->iState = SFX_STATE__STOPPED;
    switch (pFX->iFileFormat)
    {
      case SFX_FORMAT__OGG:
	OGGSeek((OGGHANDLE) pFX->DataHandle, 0);

	break;
      default:
#ifdef SFX_DEBUG_BUILD
	printf("Internal error, wrong fileformat tag!\n");
	fflush(stdout);
#endif
        DosReleaseMutexSem(hmtxUseSFXList);
	return 0;
    }
    DosReleaseMutexSem(hmtxUseSFXList);
  } else return 0;

  return 1;
}

//---------------------------------------------------------------------
// SFXGetFXPos
//
// Returns the current position of a given SFX, in milliseconds.
// Please note, that it returns the position of the decoder, not taking
// care about how much data is in the DART buffers, waiting to be
// played!
//---------------------------------------------------------------------
long long SFXGetFXPos(PSFX_FX pFX)
{
  long long rc;

  if (!pFX) return 0;

  if (DosRequestMutexSem(hmtxUseSFXList, 1000)==NO_ERROR)
  {
    switch (pFX->iFileFormat)
    {
      case SFX_FORMAT__OGG:
        rc = OGGTellPos((OGGHANDLE) pFX->DataHandle);
        break;
      default:
  #ifdef SFX_DEBUG_BUILD
        printf("Internal error, wrong fileformat tag!\n");
        fflush(stdout);
  #endif
        DosReleaseMutexSem(hmtxUseSFXList);
        return 0;
    }
    DosReleaseMutexSem(hmtxUseSFXList);
  } else return 0;
  return rc;
}

//---------------------------------------------------------------------
// SFXGetFXLength
//
// Returns the length of a given SFX, in milliseconds.
//---------------------------------------------------------------------
long long SFXGetFXLength(PSFX_FX pFX)
{
  long long rc;
  if (!fSFXRunning) return 0;
  if (!pFX) return 0;

  if (DosRequestMutexSem(hmtxUseSFXList, 1000)==NO_ERROR)
  {
    switch (pFX->iFileFormat)
    {
      case SFX_FORMAT__OGG:
        rc = OGGTellLength((OGGHANDLE) pFX->DataHandle);
        break;
      default:
#ifdef SFX_DEBUG_BUILD
        printf("Internal error, wrong fileformat tag!\n");
        fflush(stdout);
#endif
        DosReleaseMutexSem(hmtxUseSFXList);
        return 0;
    }
    DosReleaseMutexSem(hmtxUseSFXList);
  } else return 0;
  return rc;
}

//---------------------------------------------------------------------
// SFXUnloadFX
//
// Stops the playback of a given SFX, and frees all the resources
// that are allocated for it.
//---------------------------------------------------------------------
int SFXUnloadFX(PSFX_FX pFX)
{
  PSFX_FX psfxPrev;
  PSFX_FX psfxNext;

  if (!pFX) return 0;

  if (DosRequestMutexSem(hmtxUseSFXList, 1000)==NO_ERROR)
  {
    // Call its decoder to be closed:
    switch (pFX->iFileFormat)
    {
      case SFX_FORMAT__OGG:
	OGGClose((OGGHANDLE) pFX->DataHandle);
	break;
      default:
#ifdef SFX_DEBUG_BUILD
	printf("Internal error, wrong fileformat tag!\n");
	fflush(stdout);
#endif
        DosReleaseMutexSem(hmtxUseSFXList);
	return 0;
    }
    // Now unlink it from the list of FX
    if (pFX->PrevSFX == NULL)
    { // This is the first in the list:
      psfxHead = pFX->NextSFX;
      if (psfxHead)
        psfxHead->PrevSFX = NULL;
    } else
    { // This is a middle SFX, not the first one!
      psfxPrev = (PSFX_FX) pFX->PrevSFX;
      psfxNext = (PSFX_FX) pFX->NextSFX;
      if (psfxPrev)
        psfxPrev->NextSFX = psfxNext;
      if (psfxNext)
        psfxNext->PrevSFX = psfxPrev;
    }
    if (pFX->DecodeBuf)
    {
      free(pFX->DecodeBuf); pFX->DecodeBuf = NULL;
    }
    free(pFX);
    DosReleaseMutexSem(hmtxUseSFXList);
  } else return 0;
  return 1;
}

//---------------------------------------------------------------------
// SFXPlayFX
//
// Starts the playback of a given SFX.
//---------------------------------------------------------------------
int SFXPlayFX(PSFX_FX pFX)
{
  if (!fSFXRunning) return 1;
  if (!pFX) return 0;
  if (DosRequestMutexSem(hmtxUseSFXList, 1000)==NO_ERROR)
  {
    if (pFX->iState == SFX_STATE__PLAYING)
    { // If already playing, rewind it!
      switch (pFX->iFileFormat)
      {
        case SFX_FORMAT__OGG:
          OGGSeek((OGGHANDLE) pFX->DataHandle, 0);
          break;
        default:
    #ifdef SFX_DEBUG_BUILD
          printf("Internal error, wrong fileformat tag!\n");
          fflush(stdout);
    #endif
      }
    } else
      pFX->iState = SFX_STATE__PLAYING;
    DosReleaseMutexSem(hmtxUseSFXList);
  }
  return 1;
}

//---------------------------------------------------------------------
// SFXSetFXVolume
//
// Sets the volume of a given SFX.
//---------------------------------------------------------------------
int SFXSetFXVolume(PSFX_FX pFX, int iVolume)
{
  if (!fSFXRunning) return 0;
  if (!pFX) return 0;

  if (iVolume<0) iVolume=0;
  if (iVolume>SFX_VOLUME__MAX) iVolume = SFX_VOLUME__MAX;
  if (DosRequestMutexSem(hmtxUseSFXList, 1000)==NO_ERROR)
  {
    pFX->iVolume = iVolume;
    DosReleaseMutexSem(hmtxUseSFXList);
  }
  return iVolume;
}

//---------------------------------------------------------------------
// SFXGetFXVolume
//
// Returns the volume of a given SFX.
//---------------------------------------------------------------------
int SFXGetFXVolume(PSFX_FX pFX)
{
  if (!fSFXRunning) return 0;
  if (!pFX) return 0;

  return pFX->iVolume;
}

//---------------------------------------------------------------------
// SFXSetMainVolume
//
// Sets the main volume.
//---------------------------------------------------------------------
int SFXSetMainVolume(int iVolume)
{
  if (iVolume<0) iVolume=0;
  if (iVolume>SFX_VOLUME__MAX) iVolume = SFX_VOLUME__MAX;
  iSFXMainVolume = iVolume;
  return iVolume;
}

//---------------------------------------------------------------------
// SFXGetMainVolume
//
// Returns the main volume.
//---------------------------------------------------------------------
int SFXGetMainVolume()
{
  return iSFXMainVolume;
}

//---------------------------------------------------------------------
// SFXSetFXMode
//
// Sets the playback mode of a SFX to looped or single-shot.
// (See SFX_PLAYMODE__LOOPED and other defines...)
//---------------------------------------------------------------------
int SFXSetFXMode(PSFX_FX pFX, int iMode)
{
  if (!fSFXRunning) return 0;
  if (!pFX) return 0;

  if (DosRequestMutexSem(hmtxUseSFXList, 1000)==NO_ERROR)
  {
    pFX->iPlayMode = iMode;
    DosReleaseMutexSem(hmtxUseSFXList);
  }
  return iMode;
}

//---------------------------------------------------------------------
// SFXGetFXMode
//
// Returns the playback mode of a SFX.
// (See SFX_PLAYMODE__LOOPED and other defines...)
//---------------------------------------------------------------------
int SFXGetFXMode(PSFX_FX pFX)
{
  if (!fSFXRunning) return 0;
  if (!pFX) return 0;

  return pFX->iPlayMode;
}

//---------------------------------------------------------------------
// SFXGetFXState
//
// Returns the current state of a SFX (if it's being played, or if it's
// already stopped. See defines!)
//---------------------------------------------------------------------
int SFXGetFXState(PSFX_FX pFX)
{
  if (!fSFXRunning) return 0;
  if (!pFX) return 0;

  return pFX->iState;
}
