/*****************************************************************************
 * Sound effect library, using DART                                          *
 *                                                                           *
 * Include header file                                                       *
 *                                                                           *
 * This library can be used to load and play sound effects or musics through *
 * the DART interface of OS/2.                                               *
 * Currently only OGG file format is supported.                              *
 *****************************************************************************/
#ifndef SOUNDFX_H
#define SOUNDFX_H

// The SFX Engine currently only supports OGG playback, but
// with unlimited number of channels.
//
#ifdef DEBUG_BUILD
 #define SFX_DEBUG_BUILD
 #include <stdio.h>   // fflush()
#endif

// This is a structure, returned by the
// SFXQueryAvailableDevices() function,
// and destroyed by the
// SFXDestroyAvailableDeviceList() one!

typedef struct _SFX_AVAIL_DEVICES {
  int iNumOfDevices;      // Number of detected AMP devices
  long *lDevNameIndex;    // Indexes to DevNames, pointing to the start position of the null-terminated strings
  char *DevNames;         // Names of devices
} SFX_AVAIL_DEVICES, *PSFX_AVAIL_DEVICES;

#define SFX_VOLUME__MAX          256
#define SFX_VOLUME__MIN          0
#define SFX_VOLUME__MUTE         0

#define SFX_LOAD_MODE__STREAM    0
#define SFX_LOAD_MODE__MEMORY    1

#define SFX_PLAY_MODE__LOOP      0
#define SFX_PLAY_MODE__SINGLE    1

#define SFX_STATE__STOPPED       0
#define SFX_STATE__PLAYING       1

#define SFX_FORMAT__OGG          0
#define SFX_FORMAT__MP3          1
#define SFX_FORMAT__PCM          2
#define SFX_FORMAT__WAV          3

typedef struct _SFX_FX {

  void *DataHandle; // File or Memory-file handle

  int iPlayMode;    // __SINGLE or __LOOP
  int iState;       // __STOPPED or __PLAYING
  int iFileFormat;  // __OGG, __MP3, __PCM ...

  int iVolume;      // 0..256

  void *DecodeBuf;  // Buffer to decode audio data into
  int DecodeBufSize;// Size of that buffer

  int iSampleRate;  // Audio parameters
  int iChannels;
  int iBits;

  void *NextSFX;    // We link the loaded SFX's in a list, so we can keep track
  void *PrevSFX;    // of them anytime, even in the mixer.
} SFX_FX, *PSFX_FX;


//      Functions:

// (For description, check the .c code!)

PSFX_AVAIL_DEVICES SFXQueryAvailableDevices();
void SFXDestroyAvailableDeviceList(PSFX_AVAIL_DEVICES ad);

int SFXOpenDevice(int iDeviceOrd, int bOpenShared, int iFreq, int iBits, int iChannels, int iNumBufs, int iBufSize);
int SFXCloseDevice();

PSFX_FX SFXLoadFX(char *filename, int iLoadMode);

int SFXPlayFX(PSFX_FX pFX);
int SFXPauseFX(PSFX_FX pFX);
int SFXResumeFX(PSFX_FX pFX);
int SFXStopFX(PSFX_FX pFX);

int SFXSetFXVolume(PSFX_FX pFX, int iVolume);
int SFXGetFXVolume(PSFX_FX pFX);

int SFXSetFXMode(PSFX_FX pFX, int iMode);
int SFXGetFXMode(PSFX_FX pFX);

int SFXGetFXState(PSFX_FX pFX);

long long SFXGetFXPos(PSFX_FX pFX);
long long SFXGetFXLength(PSFX_FX pFX);

int SFXUnloadFX(PSFX_FX pFX);

int SFXSetMainVolume(int iVolume);
int SFXGetMainVolume();

#endif
