/*****************************************************************************
 * Sprite support library                                                    *
 *                                                                           *
 * Include header file                                                       *
 *                                                                           *
 * Using this library, one can use "Sprites", transparent and translucent    *
 * small images in his/her program.                                          *
 * The library uses the Images libary to load the images, from where the     *
 * sprite data is extracted.
 *****************************************************************************/
#ifndef SPRITES_H
#define SPRITES_H

#include "Images.h" // This is needed because LoadSprite uses images

#ifdef DEBUG_BUILD
//  #define SPRITES_DEBUG_BUILD
  #include <stdio.h>
#endif

// Type declarations for the Sprites module:
typedef struct _tSprite
{
  // General describer stuff:
  unsigned long ulWidth;        // Sprite width and height
  unsigned long ulHeight;
  unsigned long ulHotSpotX;     // Sprite "HotSpot" coordinate: the coordinate inside the
  unsigned long ulHotSpotY;     // sprite that should be at the requested position when a sprite is to be displayed
  unsigned long ulOneFrameSize; // ulWidth*ulHeight*3;
  unsigned long ulNumOfFrames;  // Number of animation frames of the sprite
  unsigned long ulRestartFrame; // Ordinal (from 0) of frame where to restart animation when all the frames have been displayed
  unsigned long ulSize;         // ulOneFrameSize * ulNumOfFrames;
  float flAnimSpeed;            // Speed of animation (how many animation frames to jump in one display frame)
  unsigned char *pColorData;    // The color data
  unsigned char *pColorMask;    // The transparency data

  // Stuff storing the actual state of the sprite:
  unsigned long ulActFrame;
  float flActFrame;

} tSprite, *pSprite;

// Functions of Sprites module:

// For description, check the .c source code!

void InitializeSprites();
void UninitializeSprites();

pSprite LoadSprite(pImage Image, char *Params);
void CleanupSprite(pSprite Sprite);

void DrawSprite(unsigned char *pDstBuf, unsigned long ulDstBufWidth,
                long ulViewPortTLX, long ulViewPortTLY,
                long ulViewPortBRX, long ulViewPortBRY,
		long ulDstPosX, long ulDstPosY,
		pSprite Sprite,
		unsigned char bTransparency);
void RemoveDrawnSprites(unsigned char *pBuffer, unsigned long ulDstBufWidth);
void AnimateSprite(pSprite Sprite);

#endif
