/*

  This file is an all-in-one header file for OGG Vorbis decoding, using
  the interger-only OGG decoder, called Tremor.
  The purpose of this file is to be able to easily use the OGGIDEC.DLL
  file (output of Tremor on OS/2, using the OpenWatcom compiler).

  This file is derived from the header files found in that library, so it
  has its own licence! Please check http://www.xiph.org/ for more info!
 
  THE OggVorbis 'TREMOR' SOURCE CODE IS (C) COPYRIGHT 1994-2002
  BY THE Xiph.Org FOUNDATION http://www.xiph.org/

*/

#ifndef OGGIDEC_H
#define OGGIDEC_H

#ifdef __cplusplus
extern "C"
{
#endif /* __cplusplus */

/* ------------------------------------------------------- os_types.h --------

 ********************************************************************

 function: #ifdef jail to whip a few platforms into the UNIX ideal.

 ********************************************************************/

/* OS/2 GCC and OS/2 Watcom C*/
typedef short ogg_int16_t;
typedef int ogg_int32_t;
typedef unsigned int ogg_uint32_t;
typedef __int64 ogg_int64_t;

/* ------------------------------------------------------------ ogg.h --------

 ********************************************************************

 function: subsumed libogg includes

 ********************************************************************/

typedef struct {
  long endbyte;
  int  endbit;

  unsigned char *buffer;
  unsigned char *ptr;
  long storage;
} oggpack_buffer;

/* ogg_page is used to encapsulate the data in one Ogg bitstream page *****/

typedef struct {
  unsigned char *header;
  long header_len;
  unsigned char *body;
  long body_len;
} ogg_page;

/* ogg_stream_state contains the current encode/decode state of a logical
   Ogg bitstream **********************************************************/

typedef struct {
  unsigned char   *body_data;    /* bytes from packet bodies */
  long    body_storage;          /* storage elements allocated */
  long    body_fill;             /* elements stored; fill mark */
  long    body_returned;         /* elements of fill returned */


  int     *lacing_vals;      /* The values that will go to the segment table */
  ogg_int64_t *granule_vals; /* granulepos values for headers. Not compact
				this way, but it is simple coupled to the
				lacing fifo */
  long    lacing_storage;
  long    lacing_fill;
  long    lacing_packet;
  long    lacing_returned;

  unsigned char    header[282];      /* working space for header encode */
  int              header_fill;

  int     e_o_s;          /* set when we have buffered the last packet in the
                             logical bitstream */
  int     b_o_s;          /* set after we've written the initial page
                             of a logical bitstream */
  long    serialno;
  long    pageno;
  ogg_int64_t  packetno;      /* sequence number for decode; the framing
                             knows where there's a hole in the data,
                             but we need coupling so that the codec
                             (which is in a seperate abstraction
                             layer) also knows about the gap */
  ogg_int64_t   granulepos;

} ogg_stream_state;

/* ogg_packet is used to encapsulate the data and metadata belonging
   to a single raw Ogg/Vorbis packet *************************************/

typedef struct {
  unsigned char *packet;
  long  bytes;
  long  b_o_s;
  long  e_o_s;

  ogg_int64_t  granulepos;
  
  ogg_int64_t  packetno;     /* sequence number for decode; the framing
				knows where there's a hole in the data,
				but we need coupling so that the codec
				(which is in a seperate abstraction
				layer) also knows about the gap */
} ogg_packet;

typedef struct {
  unsigned char *data;
  int storage;
  int fill;
  int returned;

  int unsynced;
  int headerbytes;
  int bodybytes;
} ogg_sync_state;

/* --------------------------------------------------- ivorbiscodec.h --------

 ********************************************************************

 function: libvorbis codec headers

 ********************************************************************/

typedef struct vorbis_info{
  int version;
  int channels;
  long rate;

  /* The below bitrate declarations are *hints*.
     Combinations of the three values carry the following implications:
     
     all three set to the same value: 
       implies a fixed rate bitstream
     only nominal set: 
       implies a VBR stream that averages the nominal bitrate.  No hard 
       upper/lower limit
     upper and or lower set: 
       implies a VBR bitstream that obeys the bitrate limits. nominal 
       may also be set to give a nominal rate.
     none set:
       the coder does not care to speculate.
  */

  long bitrate_upper;
  long bitrate_nominal;
  long bitrate_lower;
  long bitrate_window;

  void *codec_setup;
} vorbis_info;

/* vorbis_dsp_state buffers the current vorbis audio
   analysis/synthesis state.  The DSP state belongs to a specific
   logical bitstream ****************************************************/
typedef struct vorbis_dsp_state{
  int analysisp;
  vorbis_info *vi;

  ogg_int32_t **pcm;
  ogg_int32_t **pcmret;
  int      pcm_storage;
  int      pcm_current;
  int      pcm_returned;

  int  preextrapolate;
  int  eofflag;

  long lW;
  long W;
  long nW;
  long centerW;

  ogg_int64_t granulepos;
  ogg_int64_t sequence;

  void       *backend_state;
} vorbis_dsp_state;

typedef struct vorbis_block{
  /* necessary stream state for linking to the framing abstraction */
  ogg_int32_t  **pcm;       /* this is a pointer into local storage */ 
  oggpack_buffer opb;
  
  long  lW;
  long  W;
  long  nW;
  int   pcmend;
  int   mode;

  int         eofflag;
  ogg_int64_t granulepos;
  ogg_int64_t sequence;
  vorbis_dsp_state *vd; /* For read-only access of configuration */

  /* local storage to avoid remallocing; it's up to the mapping to
     structure it */
  void               *localstore;
  long                localtop;
  long                localalloc;
  long                totaluse;
  struct alloc_chain *reap;

} vorbis_block;

/* vorbis_block is a single block of data to be processed as part of
the analysis/synthesis stream; it belongs to a specific logical
bitstream, but is independant from other vorbis_blocks belonging to
that logical bitstream. *************************************************/

struct alloc_chain{
  void *ptr;
  struct alloc_chain *next;
};

/* vorbis_info contains all the setup information specific to the
   specific compression/decompression mode in progress (eg,
   psychoacoustic settings, channel setup, options, codebook
   etc). vorbis_info and substructures are in backends.h.
*********************************************************************/

/* the comments are not part of vorbis_info so that vorbis_info can be
   static storage */
typedef struct vorbis_comment{
  /* unlimited user comment fields.  libvorbis writes 'libvorbis'
     whatever vendor is set to in encode */
  char **user_comments;
  int   *comment_lengths;
  int    comments;
  char  *vendor;

} vorbis_comment;

/* Vorbis ERRORS and return codes ***********************************/

#define OV_FALSE      -1  
#define OV_EOF        -2
#define OV_HOLE       -3

#define OV_EREAD      -128
#define OV_EFAULT     -129
#define OV_EIMPL      -130
#define OV_EINVAL     -131
#define OV_ENOTVORBIS -132
#define OV_EBADHEADER -133
#define OV_EVERSION   -134
#define OV_ENOTAUDIO  -135
#define OV_EBADPACKET -136
#define OV_EBADLINK   -137
#define OV_ENOSEEK    -138

/* ---------------------------------------------------- ivorbisfile.h --------

 ********************************************************************

 function: stdio-based convenience library for opening/seeking/decoding

 ********************************************************************/

#define CHUNKSIZE 8192

/* The function prototypes for the callbacks are basically the same as for
 * the stdio functions fread, fseek, fclose, ftell. 
 * The one difference is that the FILE * arguments have been replaced with
 * a void * - this is to be used as a pointer to whatever internal data these
 * functions might need. In the stdio case, it's just a FILE * cast to a void *
 * 
 * If you use other functions, check the docs for these functions and return
 * the right values. For seek_func(), you *MUST* return -1 if the stream is
 * unseekable
 */
typedef struct {
  size_t (*read_func)  (void *ptr, size_t size, size_t nmemb, void *datasource);
  int    (*seek_func)  (void *datasource, ogg_int64_t offset, int whence);
  int    (*close_func) (void *datasource);
  long   (*tell_func)  (void *datasource);
} ov_callbacks;

#define  NOTOPEN   0
#define  PARTOPEN  1
#define  OPENED    2
#define  STREAMSET 3
#define  INITSET   4

typedef struct OggVorbis_File {
  void            *datasource; /* Pointer to a FILE *, etc. */
  int              seekable;
  ogg_int64_t      offset;
  ogg_int64_t      end;
  ogg_sync_state   oy; 

  /* If the FILE handle isn't seekable (eg, a pipe), only the current
     stream appears */
  int              links;
  ogg_int64_t     *offsets;
  ogg_int64_t     *dataoffsets;
  long            *serialnos;
  ogg_int64_t     *pcmlengths;
  vorbis_info     *vi;
  vorbis_comment  *vc;

  /* Decoding working state local storage */
  ogg_int64_t      pcm_offset;
  int              ready_state;
  long             current_serialno;
  int              current_link;

  ogg_int64_t      bittrack;
  ogg_int64_t      samptrack;

  ogg_stream_state os; /* take physical pages, weld into a logical
                          stream of packets */
  vorbis_dsp_state vd; /* central working state for the packet->PCM decoder */
  vorbis_block     vb; /* local working space for packet->PCM decode */

  ov_callbacks callbacks;

} OggVorbis_File;

typedef int (t_ov_clear)(OggVorbis_File *vf);
typedef int (t_ov_open)(FILE *f,OggVorbis_File *vf,char *initial,long ibytes);
typedef int (t_ov_open_callbacks)(void *datasource, OggVorbis_File *vf,	char *initial, long ibytes, ov_callbacks callbacks);

typedef int (t_ov_test)(FILE *f,OggVorbis_File *vf,char *initial,long ibytes);
typedef int (t_ov_test_callbacks)(void *datasource, OggVorbis_File *vf,	char *initial, long ibytes, ov_callbacks callbacks);
typedef int (t_ov_test_open)(OggVorbis_File *vf);

typedef long (t_ov_bitrate)(OggVorbis_File *vf,int i);
typedef long (t_ov_bitrate_instant)(OggVorbis_File *vf);
typedef long (t_ov_streams)(OggVorbis_File *vf);
typedef long (t_ov_seekable)(OggVorbis_File *vf);
typedef long (t_ov_serialnumber)(OggVorbis_File *vf,int i);

typedef ogg_int64_t (t_ov_raw_total)(OggVorbis_File *vf,int i);
typedef ogg_int64_t (t_ov_pcm_total)(OggVorbis_File *vf,int i);
typedef ogg_int64_t (t_ov_time_total)(OggVorbis_File *vf,int i);

typedef int (t_ov_raw_seek)(OggVorbis_File *vf,long pos);
typedef int (t_ov_pcm_seek)(OggVorbis_File *vf,ogg_int64_t pos);
typedef int (t_ov_pcm_seek_page)(OggVorbis_File *vf,ogg_int64_t pos);
typedef int (t_ov_time_seek)(OggVorbis_File *vf,ogg_int64_t pos);
typedef int (t_ov_time_seek_page)(OggVorbis_File *vf,ogg_int64_t pos);

typedef ogg_int64_t (t_ov_raw_tell)(OggVorbis_File *vf);
typedef ogg_int64_t (t_ov_pcm_tell)(OggVorbis_File *vf);
typedef ogg_int64_t (t_ov_time_tell)(OggVorbis_File *vf);

typedef vorbis_info *(t_ov_info)(OggVorbis_File *vf,int link);
typedef vorbis_comment *(t_ov_comment)(OggVorbis_File *vf,int link);

typedef long (t_ov_read)(OggVorbis_File *vf,char *buffer,int length, int *bitstream);

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif
