#define INCL_TYPES
#define INCL_WIN
#define INCL_DOSPROCESS
#include <os2.h>
// Include window resource ID's
#include "Chat.h"
// Include GameDLL definitions
#include "GameDLLdef.h"

//---- Game Information ----

tGameInfoBuf GameInfo = { 1, 0, 0,                   // Game version: v1.0.0
                          1, 0,                      // NetComm version: v1.0
                          "Chat",                    // Game name
                          "Simple two-people chat",  // Game description
                          GT_DUEL                    // Game type
                        };

//---- Helper functions from NetGame Client ----

fnSendMessage    *pSendMessage;
fnReceiveMessage *pReceiveMessage;

//---- Variables -----------

ULONG Socket;
CHAR *RemoteUserName;
CHAR *LocalUserName;
BOOL ConnectionAlive=FALSE;
BOOL bEntryFieldActive=FALSE;   // True, if the Chat-entry field is active
HSWITCH hSwitchEntry;           // Handle to Window List entry

//---- Code ----

//---------------------------------------------------------------------
// LibMain
//
// This gets called at DLL initialization and termination.
// We have nothing to do here.
//---------------------------------------------------------------------
unsigned _System LibMain( unsigned hmod, unsigned termination )
{
  if (termination) {
    // Cleanup!
  } else {
    // Startup!
  }
  return 1;
}

//---------------------------------------------------------------------
// GetGameInfo
//
// NetGame Client calls this to get some information about this module.
// All we have to do here is to provide a GameInfoBuf, which describes
// this module. (See the declaration above!)
//---------------------------------------------------------------------
BOOL FAR __syscall __export GetGameInfo( ULONG BufSize, pGameInfoBuf Buffer )
{
  // Check for valid buffer size
  if (BufSize!=sizeof(tGameInfoBuf)) return FALSE;

  memcpy(Buffer, &GameInfo, sizeof(GameInfo));
  return TRUE;
}

//---------------------------------------------------------------------
// InitializeGameDLL
//
// NetGame Client calls this when the game is about to begin. It pro-
// vides some helper functions that can be used to communicate through
// TCP/IP (but it's not obligatory to use them). All we do here is that
// we save the addresses of these helper functions for later use.
//---------------------------------------------------------------------
BOOL FAR __syscall __export InitializeGameDLL( ULONG BufSize, pInitBuf Buffer )
{
  // Check for valid buffer size
  if (BufSize!=sizeof(tInitBuf)) return FALSE;

  // Save helper functions
  pSendMessage = Buffer->SendMessage;
  pReceiveMessage = Buffer->ReceiveMessage;

  // Return!
  return TRUE;
}

// ---- Now some local functions that are used by this module: ----

//---------------------------------------------------------------------
// SendChatText
//
// Sends a chat message to the other side.
//---------------------------------------------------------------------
BOOL SendChatText(int Socket, char *Msg)
{
  char *FullMsg=(char *)malloc(strlen(Msg)+9);
  BOOL rc;

  // Create the data to send: "CHATMSG <message>"
  strcpy( FullMsg, "CHATMSG ");

  strcat( FullMsg, Msg);

  // Now send this string to the other Chat DLL!
  rc=(*pSendMessage)(Socket, FullMsg);

  // Cleanup, and return!
  free(FullMsg);
  return rc;
}

//---------------------------------------------------------------------
// WriteChatText
//
// Appends a line at the end of the MLE, describing the user who wrote
// the message, and the message itself.
//---------------------------------------------------------------------
void WriteChatText(HWND DlgWnd, char *UserName, char *Message)
{
  IPT iptOffset;
  char *FullMessage = (char *)malloc(strlen(Message)+strlen(UserName)+8);

  // Create the line that will be appended to the end of the MLE!
  // It will be something like: "<UserName> Message"
  strcpy(FullMessage,"<");
  strcat(FullMessage,UserName);
  strcat(FullMessage,"> ");
  strcat(FullMessage,Message);
  strcat(FullMessage,"\n");

  // Set Import/Export buffer
  WinSendDlgItemMsg( DlgWnd, MLE_CHATFIELD, MLM_SETIMPORTEXPORT,
                         MPFROMP(FullMessage),
                     MPFROMSHORT(strlen(FullMessage)));

  // Get a pointer to the end of the contents of MLE
  iptOffset = 
    (IPT)WinSendDlgItemMsg( DlgWnd, MLE_CHATFIELD, MLM_QUERYTEXTLENGTH, 
                            (MPARAM)0, (MPARAM)0);

  // Insert the text there
  WinSendDlgItemMsg( DlgWnd, MLE_CHATFIELD, MLM_IMPORT, MPFROMP(&iptOffset),
                     MPFROMSHORT(strlen(FullMessage)));

  // Query the new end-of-MLE position
  iptOffset = 
    (IPT)WinSendDlgItemMsg( DlgWnd, MLE_CHATFIELD, MLM_QUERYTEXTLENGTH, 
                            (MPARAM)0, (MPARAM)0);

  // And clear the text-selection, if there is any
  WinSendDlgItemMsg( DlgWnd, MLE_CHATFIELD, MLM_SETSEL, (MPARAM)(iptOffset),
                     (MPARAM)(iptOffset));

  // Cleanup!
  free(FullMessage);
}

//---------------------------------------------------------------------
// ResizeAndRepositionControls
//
// Gets called at dialog init time, and every time when the size of the
// window changes.
//---------------------------------------------------------------------
void ResizeAndRepositionControls(HWND hwnd)
{
  SWP Swp1, Swp2;
  int CXDlgFrame = WinQuerySysValue( HWND_DESKTOP,  SV_CXDLGFRAME);
  int CYDlgFrame = WinQuerySysValue( HWND_DESKTOP,  SV_CYDLGFRAME);
  
  // Disable window update while repositioning controls!
  WinEnableWindowUpdate(hwnd, FALSE);
  
  // First position the close button to the bottom-left corner!
  WinQueryWindowPos( hwnd, &Swp1);
  WinQueryWindowPos( WinWindowFromID( hwnd, PB_CLOSE), &Swp2);
  WinSetWindowPos( WinWindowFromID( hwnd, PB_CLOSE), HWND_TOP,
		   Swp1.cx - CXDlgFrame - Swp2.cx - 8, CYDlgFrame + 5,
		   0, 0,
		   SWP_MOVE);
  
  // Put the Entry field next to the button, full length long!
  WinQueryWindowPos( WinWindowFromID( hwnd, PB_CLOSE), &Swp2);
  WinSetWindowPos( WinWindowFromID( hwnd, EF_ENTRY), HWND_TOP,
		   CXDlgFrame + 8, CYDlgFrame + 5 + (Swp2.cy - 18) / 2,
		   Swp2.x - 10 - (CXDlgFrame+8) , 18,
		   SWP_MOVE | SWP_SIZE);
  
  // Move the Information Groupbox to its new positon!
  WinQueryWindowPos( WinWindowFromID( hwnd, GB_INFORMATION), &Swp2);
  WinSetWindowPos( WinWindowFromID( hwnd, GB_INFORMATION), HWND_TOP,
		   CXDlgFrame + 8, Swp1.cy - CYDlgFrame - WinQuerySysValue(HWND_DESKTOP, SV_CYTITLEBAR) - Swp2.cy - 8,
		   Swp1.cx - 2*(CXDlgFrame+8), Swp2.cy,
		   SWP_MOVE | SWP_SIZE);
  
  // Now the Ping button
  WinQueryWindowPos( WinWindowFromID( hwnd, PB_PING), &Swp1);
  WinQueryWindowPos( WinWindowFromID( hwnd, GB_INFORMATION), &Swp2);
  WinSetWindowPos( WinWindowFromID( hwnd, PB_PING), HWND_TOP,
		   Swp2.x+Swp2.cx - Swp1.cx - 8, Swp2.y + (Swp2.cy-Swp1.cy-2*CYDlgFrame)/2,
		   0, 0,
		   SWP_MOVE);
  // The 'Chatting with' text:
  WinQueryWindowPos( WinWindowFromID( hwnd, ST_CHATTINGWITH), &Swp1);
  WinSetWindowPos( WinWindowFromID( hwnd, ST_CHATTINGWITH), HWND_TOP,
		   Swp2.x+8, Swp2.y + (Swp2.cy-Swp1.cy-2*CYDlgFrame)/2,
		   0, 0,
		   SWP_MOVE);
  // Peer name entry field:
  WinQueryWindowPos( WinWindowFromID( hwnd, PB_PING), &Swp1);
  WinQueryWindowPos( WinWindowFromID( hwnd, ST_CHATTINGWITH), &Swp2);
  WinSetWindowPos( WinWindowFromID( hwnd, EF_PEERNAME), HWND_TOP,
		   Swp2.x + Swp2.cx + 8, Swp1.y + CYDlgFrame,
		   Swp1.x -8 - (Swp2.x + Swp2.cx + 8), Swp1.cy - 2*CYDlgFrame,
		   SWP_MOVE | SWP_SIZE);
  
  // Move/Size the ChatField MLE
  WinQueryWindowPos( WinWindowFromID( hwnd, PB_CLOSE), &Swp1);
  WinQueryWindowPos( WinWindowFromID( hwnd, GB_INFORMATION), &Swp2);
  WinSetWindowPos( WinWindowFromID( hwnd, MLE_CHATFIELD), HWND_TOP,
		   CXDlgFrame + 8, Swp1.y + Swp1.cy + 10,
		   Swp2.cx, Swp2.y - Swp1.y - Swp1.cy -20,
		   SWP_MOVE | SWP_SIZE);
  
  // Now we can enable window update.
  WinEnableWindowUpdate(hwnd, TRUE);
}

//---------------------------------------------------------------------
// MyDlgProc
//
// The window procedure that processes messages for the Chat window.
//---------------------------------------------------------------------
MRESULT EXPENTRY MyDlgProc( HWND hwnd, USHORT msg, MPARAM mp1, MPARAM mp2 )
{
  CHAR Buffer[256]; // Work buffer

  // Ok, process some interesting messages!

  switch (msg)
  {
    case WM_INITDLG:
      // Dialog init time: Set initial position and size of controls!
      ResizeAndRepositionControls(hwnd);
      break;
    case WM_CHAR:
      // Keypress:
      if ((bEntryFieldActive) &&               // If the user is at the entry field,
	  !(SHORT1FROMMP(mp1)&KC_KEYUP) &&     // and a key is pressed,
	  (SHORT1FROMMP(mp1)&KC_VIRTUALKEY) && // which is an Enter key:
	  ((SHORT2FROMMP(mp2)==VK_ENTER) || (SHORT2FROMMP(mp2)==VK_NEWLINE)))
      {

	// Get the user-entered text into out local buffer
	memset(Buffer, 256, 0);
	WinQueryDlgItemText(hwnd, EF_ENTRY, 256, Buffer);

	// Send it, if the connection is alive! Write it into the MLE too, if
	// the sending was successful!
	if ((ConnectionAlive) && (SendChatText(Socket, Buffer)))
	  WriteChatText(hwnd, LocalUserName, Buffer);

	// Clear the entry field, and return
	WinSetDlgItemText(hwnd, EF_ENTRY, "");
	return (MRESULT)TRUE;
      }
      break;

    case WM_COMMAND:
      // Some command, needs more investigation:
      switch SHORT1FROMMP(mp1) {
	case PB_PING: // It's a message from the Ping button

	  if (SHORT1FROMMP(mp2)==CMDSRC_PUSHBUTTON)
	  { // If it's that the button has been pressed:
            char StrValue[40];

            // Send a "CHATPING <nowtime>" to the other side!
            memset(Buffer, 256, 0);
            strcpy(Buffer,"CHATPING ");
            itoa(clock(), StrValue, 10);
	    strcat(Buffer, StrValue);

	    if ((ConnectionAlive) && ((*pSendMessage)(Socket, Buffer)))
	      WriteChatText(hwnd, "*", "Pinging...");
            return (MRESULT)TRUE;
          }
	  break;

	case PB_CLOSE: // It's a message from the Close button

          if (SHORT1FROMMP(mp2)==CMDSRC_PUSHBUTTON) 
	  { // If it's that the button has been pressed:
            // Post a WM_CLOSE message to this chat window, so exit!
            WinPostMsg(hwnd, WM_CLOSE, 0, 0);
            return (MRESULT)TRUE;
          }
	  break;

	default: ; // Never forget about the defaults!
      } 
      break;

    case WM_CONTROL:
      // Some kind of Control message from somebody:

      switch SHORT1FROMMP(mp1) {
	case EF_ENTRY:
	  // Ok, it's a control message from the Entry field.
	  // Save it into a variable if it's the active control
	  // right now, or not!
          if (SHORT2FROMMP(mp1)==EN_KILLFOCUS) bEntryFieldActive=FALSE;
          if (SHORT2FROMMP(mp1)==EN_SETFOCUS) bEntryFieldActive=TRUE;
	  break;

        default: ; // Never forger about the defaults!
      }
      break;

    case WM_QUERYTRACKINFO:
      // We can limit the window size by catching this message,
      // and modifying the result of WinDefDlgProc!
      {
	PTRACKINFO ti = (PTRACKINFO) mp2;

	WinDefDlgProc(hwnd, msg, mp1, mp2);
	ti->ptlMinTrackSize.x = 300;
	ti->ptlMinTrackSize.y = 250;
        return (MRESULT) TRUE;
      }

    case WM_FORMATFRAME:
      // The size of the window has been changed, we have to
      // reposition/resize controls inside!
      ResizeAndRepositionControls(hwnd);
      // And now let the default procedure do the real job!
    default: ; // Never forger about the defaults!
  }

  // Handle the messages that were not handled
  return WinDefDlgProc(hwnd, msg, mp1, mp2);
}

// ---- tThreadParams is a structure, that will be passed to the
//      thread, that listens for messages arriving from the other
//      side ----
typedef struct _tThreadParams {
  ULONG Socket; // Socket to listen for messages in
  HWND  DlgWnd; // Window handle
} tThreadParams;
typedef tThreadParams *pThreadParams;

//---------------------------------------------------------------------
// substr
//
// Extracts a substring from another string.
//---------------------------------------------------------------------
char *substr(char *Dst, char *Src, int Start, int Len)
{
  int i;

  for (i=Start; i<Len; i++) Dst[i-Start]=Src[i];
  Dst[Len]=0;
  return Dst;
}

//---------------------------------------------------------------------
// MsgReceiver
//
// This is a thread-function, that listens for messages coming from
// the other client, and processes them.
//---------------------------------------------------------------------
VOID MsgReceiver(PVOID Param)
{
  pThreadParams P = Param; // Access the thread parameters throught this
  CHAR Msg[256];           // Grab the message, that arrives, into this
  HAB AB;                  // Anchor block for this thread
  HMQ MQ;                  // Message queue for this thread
  IPT iptOffset;           // Insert position of MLE
  CHAR cmd[20];            // Extract command from the message into this
  int i;                   // General variable

  AB=WinInitialize(0);         // Prepare thread for
  MQ=WinCreateMsgQueue(AB, 0); // using the Presentation Manager

  ConnectionAlive=TRUE;        // Indicate that the connection is alive

  while (1) {                  // Now process messages!

    // Get a message into the message buffer.
    // Exit the thread, if an error occures!
    memset(Msg, 256, 0);
    if (!((*pReceiveMessage)( P->Socket, 255, Msg, -1 ))) goto EndThread;
    if (!strcmp(Msg,"")) goto EndThread;

    // Ok, we have a message here. Let's see if it's a chat message,
    // a ping, a reply to the ping, or a disconnect message!

    // If it's a chat message, write it into the MLE!
    if (!strcmp(substr(cmd, Msg, 0, 8),"CHATMSG "))
      WriteChatText(P->DlgWnd, RemoteUserName, Msg+8);
    else
      // If it's a disconnect message, end the thread!
      if (!strcmp(substr(cmd, Msg, 0, 10),"DISCONNECT")) goto EndThread;
      else
	// If it's a PING message, send back the same thing, but
        // with PONG inside! (replace 'I' with 'O')
	if (!strcmp(substr(cmd, Msg, 0, 9),"CHATPING "))
	{
	  WriteChatText(P->DlgWnd, "*", "Replying to Ping request.");
	  Msg[5]='O';
	  (*pSendMessage)(Socket, Msg);
	}
	else
	  // If it's a PONG message, calculate the elapsed time,
          // and inform the user
	  if (!strcmp(substr(cmd, Msg, 0, 9),"CHATPONG "))
	  {
	    char StrValue[40];
	    i=atoi(Msg+9);
	    strcpy(Msg,"Ping reply from peer: ");
	    itoa(clock() - i, StrValue, 10);
	    strcat(Msg, StrValue);
	    strcat(Msg," ms");
	    WriteChatText(P->DlgWnd, "*", Msg);
	  }
	  else goto EndThread; // Otherwise: Unknown command

  } // End of while (1)

 EndThread:
  // Exiting the thread:
  //  Indicate theat the connection is not alive anymore!
  ConnectionAlive=FALSE;
  //  Inform the user about it,
  WriteChatText(P->DlgWnd, "*", "Connection reset by peer.");
  //  and inform the other threads too,
  ConnectionAlive=FALSE;
  //  then clean up this thread and end it!
  WinDestroyMsgQueue(MQ);
  WinTerminate(AB);
  _endthread();
}

//---------------------------------------------------------------------
// StartGame_ClientMode
//
// The NetGame Client calls this when the connection has been made, and
// it's time to start the game. This function is called when this side
// acts as the client in this game (we're the invited one).
//
// In our case, there is no difference, so the *_ServerMode will call
// this function too.
//---------------------------------------------------------------------
BOOL FAR __syscall __export StartGame_ClientMode( ULONG BufSize, pStartClientBuf Buffer ){
  BOOL rc;               // Result code (general)
  HWND DlgWin;           // Handle of out dialog window to be created
  pThreadParams TP;      // Thread parameters
  TID MsgReceiverThread; // Thread ID of the receiver thread
  HPOINTER WindowIcon;

  // Check the parameters!
  if (BufSize!=sizeof(tStartClientBuf)) return FALSE;

  // Get direct pointers to local and remote usernames
  LocalUserName=Buffer->LocalUserName;
  RemoteUserName=Buffer->RemoteUserName;

  // Load Dialog window
  DlgWin=WinLoadDlg(HWND_DESKTOP, HWND_DESKTOP, (PFNWP)MyDlgProc, Buffer->hmod,
             DLG_CHATWINDOW, NULL);
  if (DlgWin) 
  {
    char *TempStr;
    int iMaxSize;
    SWP swp;
    SWCNTRL swctl;

    // Setup controls in dialog window

    // Set test limit of entry field to 256 bytes
    WinSendDlgItemMsg(DlgWin, EF_ENTRY, EM_SETTEXTLIMIT, MPFROMLONG(256L), 0);
    // Show the name of the remote user
    WinSetDlgItemText(DlgWin, EF_PEERNAME, RemoteUserName);

    // Show the name of the local user in the groupbox's caption:
    // Query the actual caption,
    iMaxSize=WinQueryDlgItemTextLength(DlgWin, GB_INFORMATION) + strlen(LocalUserName) + 4;
    TempStr=(char *)malloc(iMaxSize);
    WinQueryDlgItemText(DlgWin, GB_INFORMATION, iMaxSize, TempStr);
    // and append the local username in ()'s.
    strcat(TempStr, " (");
    strcat(TempStr, LocalUserName);
    strcat(TempStr, ")");
    // Now set this new caption!
    WinSetDlgItemText(DlgWin, GB_INFORMATION, TempStr);
    free(TempStr);

    // Set window title (include remote username)

    TempStr=(char *)malloc(512);
    WinQueryWindowText(DlgWin, 512, TempStr);
    strcat(TempStr, " (");
    strcat(TempStr, RemoteUserName);
    strcat(TempStr, ")");
    WinSetWindowText(DlgWin, TempStr);

    // Set the position of the window (let the system give us a position)
    // and show it

    WinQueryTaskSizePos(Buffer->ab, 0, &swp);
    WinSetWindowPos(DlgWin, HWND_TOP,  swp.x, swp.y, 0, 0, 
		    SWP_MOVE | SWP_SHOW);

    WindowIcon=WinLoadPointer(HWND_DESKTOP, Buffer->hmod, DLG_CHATWINDOW);
    WinPostMsg(DlgWin, WM_SETICON, (MPARAM) WindowIcon, 0);

    // Add the window into the window list
    swctl.hwnd=DlgWin;
    swctl.hwndIcon=WindowIcon;
    swctl.hprog=NULLHANDLE;
    swctl.idProcess=0;
    swctl.idSession=0;
    swctl.uchVisibility=SWL_VISIBLE;
    swctl.fbJump=SWL_JUMPABLE;
    strcpy(swctl.szSwtitle,TempStr);
    swctl.bProgType=PROG_PM;
    hSwitchEntry=WinAddSwitchEntry(&swctl);

    free(TempStr);

    // Create thread that will get messages

    // First setup thread parameters
    TP = (pThreadParams) malloc(sizeof(tThreadParams));

    if (!TP)
    {
      WinRemoveSwitchEntry(hSwitchEntry);
      WinDestroyWindow(DlgWin);
      WinDestroyPointer(WindowIcon);
      return FALSE;
    }
    TP->Socket = Socket = Buffer->Socket;
    TP->DlgWnd = DlgWin;

    // then try to create thread:
    MsgReceiverThread = _beginthread( MsgReceiver,
				      NULL,
				      16384,
				      (void *)TP);

    if (MsgReceiverThread==-1)
    {
      free(TP); TP=NULL;
      WinRemoveSwitchEntry(hSwitchEntry);
      WinDestroyWindow(DlgWin);
      WinDestroyPointer(WindowIcon);
      return FALSE;
    }

    // Ok, everything set up!
    WinProcessDlg(DlgWin);

    // So, window closed, kill the thread if it's still running!
    if (ConnectionAlive)
      DosKillThread(MsgReceiverThread);
    // Cleanup!
    free(TP); TP=NULL;
    WinRemoveSwitchEntry(hSwitchEntry);
    WinDestroyWindow(DlgWin);
    WinDestroyPointer(WindowIcon);

  } else return FALSE;
  // All done, return!
  return TRUE;
}

//---------------------------------------------------------------------
// StartGame_ServerMode
//
// The NetGame Client calls this when the connection has been made, and
// it's time to start the game. This function is called when this side
// acts as the server in this game (we're the invitor).
//
// In our case, there is no difference, so we simply call the
// *_ClientMode function!
//---------------------------------------------------------------------
BOOL FAR __syscall __export StartGame_ServerMode( ULONG BufSize, pStartServerBuf Buffer )
{
  return StartGame_ClientMode(BufSize, (pStartClientBuf) Buffer);
}
