/* ------------------------------------------------------------------------
   Definition file for GameDLL creation

   Use #define HAVE_SETUPGAME before including this header file if your DLL
   will have 'SetupGame' function to export!
   ------------------------------------------------------------------------ */

#ifndef GAMEDLLDEF_H
#define GAMEDLLDEF_H

#define INCL_TYPES
#include <os2.h>

#pragma pack(push, 1)

// ----------------------------------------- Game types -------

// Simple two-player game
#define GT_DUEL        0
// Multiplayer game
#define GT_MULTIPLAYER 1

// ----------------------------------------- General contants -

#define MAX_GAMENAME_LENGTH 40
#define MAX_GAMEDESC_LENGTH 512
#define MAX_USERNAME_LENGTH 20

// ----------------------------------------- GameInfoBuf ------

typedef struct _tGameInfoBuf {
  UCHAR bGameVerMaj;                       // Game Major version number
  UCHAR bGameVerMin;                       // Game Minor version number
  USHORT sGameVerRel;                      // Game release/build number
  USHORT sGameNetCommVerMaj;               // Version of network communication protocol
  USHORT sGameNetCommVerMin;               //  (Only the same game with same NetComm version can communcate!)
  CHAR   csGameName[MAX_GAMENAME_LENGTH];  // Short name of game
  CHAR   csGameDesc[MAX_GAMEDESC_LENGTH];  // Description of game
  ULONG  lGameType;                        // Type of game (see GT_* contants)
} tGameInfoBuf;
typedef tGameInfoBuf *pGameInfoBuf;

// ----------------------------------------- InitBuf ----------

// fnSendMessage    = function(Socket:longint; Msg:pchar):LongBool;
// fnReceiveMessage = function(Socket:longint; cbBufSize:longint; MsgBuf:pchar; TimeOut:longint):LongBool;

typedef BOOL __cdecl fnSendMessage(ULONG, PCHAR);
typedef BOOL __cdecl fnReceiveMessage(ULONG, ULONG, PCHAR, LONG);

typedef struct _tInitBuf {
   fnSendMessage    *SendMessage;
   fnReceiveMessage *ReceiveMessage;
} tInitBuf;
typedef tInitBuf *pInitBuf;

// ----------------------------------------- StartClientBuf ---

typedef struct _tStartClientBuf {
  ULONG Socket;
  HAB   ab;
  HMQ   mq;
  HMODULE hmod;
  CHAR  LocalUserName[MAX_USERNAME_LENGTH];
  CHAR  RemoteUserName[MAX_USERNAME_LENGTH];
} tStartClientBuf;
typedef tStartClientBuf *pStartClientBuf;

// ----------------------------------------- StartServerBuf ---

#define tStartServerBuf tStartClientBuf
#define pStartServerBuf pStartClientBuf

// ----------------------------------------- SetupGameBuf -----

typedef struct _tSetupGameBuf {
  HAB   ab;
  HMQ   mq;
  HMODULE hmod;
} tSetupGameBuf;
typedef tSetupGameBuf *pSetupGameBuf;

// ----------------- DLL Function Declarations -----------------------

BOOL FAR __syscall __export GetGameInfo( ULONG BufSize, pGameInfoBuf Buffer );
BOOL FAR __syscall __export InitializeGameDLL( ULONG BufSize, pInitBuf Buffer );
BOOL FAR __syscall __export StartGame_ClientMode( ULONG BufSize, pStartClientBuf Buffer );
BOOL FAR __syscall __export StartGame_ServerMode( ULONG BufSize, pStartServerBuf Buffer );

// ----------------- Optional DLL Function Declarations --------------

#ifdef HAVE_SETUPGAME
BOOL FAR __syscall __export SetupGame( ULONG BufSize, pSetupGameBuf Buffer );
#endif

#pragma pack(pop)

#endif
