unit CommonNetFuncs;
// Shared Network-related functions across Server and Client
interface
uses os2base, inet;

const MAX_SEND_RETRIES=5;

Const CLIENT_UNKNOWN=0;           // Client types
      CLIENT_PLAYER=1;
      CLIENT_INFORMER=2;

      ServerNameVer='NetGameServer v1.0';

      MIN_ERRORCODE=1;
      MAX_ERRORCODE=9;

      MessageArray:array[MIN_ERRORCODE..MAX_ERRORCODE] of string=  // Message strings
                ('Too many clients!',      // 001
                 'Unknown command!',       // 002
                 'Shutting down!',         // 003
                 'Timeout!',               // 004
                 'Wrong protocol!',        // 005
                 'Invalid parameter!',     // 006
                 'No such game here!',     // 007
                 'Not accepted by user!',  // 008
                 'Not supported game type!'// 009
                 );

      ErrorDescription:array[MIN_ERRORCODE..MAX_ERRORCODE] of string=  // Message strings
                ('Too many clients connected to that player!',
                 'The client program of that player is incompatible!',
                 'That player is shutting down!',
                 'Too slow network connection, timeout occured!',
                 'The client program of that player is incompatible!',
                 'The client program of that player is incompatible!',
                 'That player doesn''t have this game!',
                 'That player didn''t accept your invitation.',
                 'The client program of that player doesn''t support this kind of games!'
                 );


Function SendString(var sock:longint;s:string):boolean;
function ReceiveString(var sock:longint):string;  // Blocking!
function ReceiveString_Timeout(var sock:longint; msTimeout:longint):string;

procedure SendErrorString(Socket:longint; ErrorNo:longint);

function StrToLong(s:string):longint;
function LongToStr(l:longint):string;
function StrPas(cs:pchar):string;
function GetSafeString(s:String):string;

// DLL-helper functions:
{$cdecl+}
function SendMessagePChar(Socket:longint; Msg:pchar):LongBool;
function ReceiveMessagePChar(Socket:longint; cbBufSize:longint; MsgBuf:pchar; TimeOut:longint):LongBool;
{$cdecl-}
implementation

{$cdecl+}
function SendMessagePChar(Socket:longint; Msg:pchar):LongBool;
begin
  if Msg<>nil then
  result:=SendString(Socket, StrPas(Msg));
end;

function ReceiveMessagePChar(Socket:longint; cbBufSize:longint; MsgBuf:pchar; TimeOut:longint):LongBool;
var s:string;
begin
  result:=false;
  if MsgBuf=nil then exit;
  s:=ReceiveString_Timeout(Socket, TimeOut)+#0;
  if s='' then exit;
  if length(s)>cbBufSize then setlength(s,cbBufSize);
  move(s[1], MsgBuf^, length(s));
  result:=true;
end;
{$cdecl-}

Function SendString(var sock:longint;s:string):boolean;
var sent:integer;
    Tries:word;
//    error:longint;
//    data,datasize:longint;
//    msock:longint;
begin
  result:=false;

  if sock=-1 then exit;

//  msock:=sock;
//  error:=os2_select(msock,0,1,0,0);
//  if (error<>1) then exit;
//  msock:=sock;
//  error:=os2_select(msock,1,0,0,0);
//  if (error<>1) then exit;
//  error:=os2_select(sock,0,0,1,0);
//  if (error<>1) then exit;

//  datasize:=sizeof(data);
//  error:=getsockopt(sock, SOL_SOCKET, SO_ACCEPTCONN, data, datasize);

  Tries:=0;
  s:=s+#13+#10;
  repeat
    sent:=send(sock,s[1],length(s),0);
    inc(Tries);
  Until (Tries>MAX_SEND_RETRIES) or (Sent=Length(s));
  result:=(Sent=Length(S));

end;

procedure SendErrorString(Socket:longint; ErrorNo:longint);
var sTemp:string;
begin
  str(ErrorNo, sTemp);
  while length(sTemp)<3 do sTemp:='0'+sTemp;
  SendString(Socket,'ERROR '+sTemp+' '+MessageArray[ErrorNo]);
end;

function ReceiveString(var sock:longint):string;
var ch:char;
    readed:longint;
begin
  result:='';
  if sock=-1 then exit;

  repeat
    readed:=recv(sock,ch,1,0);
    if (readed>0) then
    begin
      if (ch<>#10) and (ch<>#13) then
        result:=result+ch
      else
      if result<>'' then exit;
    end else exit;
  until false;
end;

function ReceiveString_Timeout(var sock:longint; msTimeout:longint):string;
var ch:char;
    readed:longint;
    rc:longint;
    SocketToCheck:longint;
begin
  result:='';
  if sock=-1 then exit;

  repeat
    SocketToCheck:=Sock;
    rc:=os2_select(SocketToCheck,1,0,0,msTimeout); // Wait for data to read, max msTimeOut!
    if rc>0 then
    begin
      readed:=recv(sock,ch,1,0);
      if (readed>0) then
      begin
        if (ch<>#10) and (ch<>#13) then
          result:=result+ch
        else
        if result<>'' then exit;
      end else exit;
    end;
  until (rc<=0);
end;

function StrToLong(s:string):longint;
var code:longint;
begin
  val(s,result,code);
end;

function LongToStr(l:longint):string;
begin
  str(l,result);
end;

function StrPas(cs:pchar):string;
begin
  result:='';
  if cs=nil then exit;
  while cs^<>#0 do
  begin
    result:=result+cs^;
    inc(longint(cs));
  end;
end;

// Gets a string between "'s
function GetSafeString(s:String):string;
var l:longint;
    state:byte;
begin
  result:='';
  // Try to find a string like: ["str"]
  l:=1;state:=0;
  while (l<=length(s)) and (State<2) do
  begin
    if s[l]='"' then inc(state) else
    if state=1 then result:=result+s[l];
    inc(l);
  end;
  if result='' then
  begin // Ok, the previous way did not give result, so another round:
    l:=1;state:=0;
    while (l<=length(s)) and (State<1) do
    begin
      if s[l]='"' then inc(state) else
      result:=result+s[l];
      inc(l);
    end;
  end;
end;


end.

