//-------------------------------------------------------------------------------------------------------------
//
// GameList related definitions and functions
//
//-------------------------------------------------------------------------------------------------------------

type pGameList=^tGameList;
     tGameList=record
       FileName : string;
       GameInfo : tGameInfoBuf;
       HaveSetup : boolean;
       Next : pGameList;
     end;

var  GameListHead:pGameList;
     GameListProtector_Sem:HEv;

function CreateGameListElement(FName:string):pGameList;
var DLLHandle:HModule;
    rc:ApiRet;
    OrgFName:String;
    DLL_GetGameInfo : fnGetGameInfo;
    DLL_SetupGame   : fnSetupGame;
begin
  result:=nil;
  OrgFName:=FName;
  FName:=FName+#0;
  DLLHandle := -1;
  rc:=DosLoadModule(nil, 0, @FName[1], DLLHandle);
  if rc<>no_error then exit;
  rc:=DosQueryProcAddr(DLLHandle, 0, 'GetGameInfo', @DLL_GetGameInfo);
  if rc=no_error then
  begin
    new(Result);
    Result^.FileName:=OrgFName;
    LongBool(rc):=DLL_GetGameInfo(Sizeof(Result^.GameInfo), @(Result^.GameInfo));
    if LongBool(rc)=false then
    begin
      dispose(result); result:=nil;
    end else
    begin
      rc:=DosQueryProcAddr(DLLHandle, 0, 'SetupGame', @DLL_SetupGame);
      Result^.HaveSetup:=(rc=no_error);
    end;
  end;
  rc:=DosFreeModule(DLLHandle);
end;

procedure FillGameList;
Var hdirFindHandle : hDir;
    FindBuffer     : FileFindBuf3; // Returned from FindFirst/Next
    ulResultBufLen : ULong;
    ulFindCount    : ULong;        // Look for 1 file at a time
    rc             : ApiRet;       // Return code
    NewGame        : pGameList;
    ItemName       : string;
begin
  if GameListProtector_Sem<>-1 then
    DosRequestMutexSem(GameListProtector_Sem, sem_indefinite_wait);

  WinSendDlgItemMsg( hwndDlgClient, LB_GAMES, LM_DELETEALL, 0, 0);

  hdirFindHandle:=HDIR_CREATE;
  ulResultBufLen := sizeof(FileFindBuf3);
  ulFindCount    := 1;

  rc := DosFindFirst(
    '*.DLL',         // File pattern - all files
    hdirFindHandle,  // Directory search handle
    file_Normal,     // Search attribute
    FindBuffer,      // Result buffer
    ulResultBufLen,  // Result buffer length
    ulFindCount,     // Number of entries to find
    fil_Standard);   // Return level 1 file info

  if rc = No_Error then
  begin
    repeat
      NewGame:=CreateGameListElement(FindBuffer.achName);
      if NewGame<>Nil then
      begin
        NewGame^.Next:=GameListHead;
        GameListHead:=NewGame;
        ItemName:=NewGame^.GameInfo.csGameName+#0;
        WinSendDlgItemMsg( hwndDlgClient, LB_GAMES,
                           LM_INSERTITEM, 0, MPARAM(@ItemName[1]));
      end;

      ulFindCount := 1;                      // Reset find count.

      rc := DosFindNext(
        hdirFindHandle,     // Directory handle
        FindBuffer,         // Result buffer
        ulResultBufLen,     // Result buffer length
        ulFindCount);       // Number of entries to find
    until (rc<>no_error);

    // Close our directory handle
    rc := DosFindClose(hdirFindHandle);
  end;

  if GameListHead=Nil then
  begin
    WinSendDlgItemMsg( hwndDlgClient, MLE_DESCRIPTIONOFTHESELECTEDGAME,
                       MLM_DELETE, 0, MAXLONGINT);
    ItemName:='Could not find any GameDLLs!'+#10+#13+
              'Please get some, and restart the program!'+#10+#13+#0;
    WinSendDlgItemMsg( hwndDlgClient, MLE_DESCRIPTIONOFTHESELECTEDGAME,
                       MLM_INSERT, MPARAM(@ItemName[1]), 0);
  end else
  begin
    WinSendDlgItemMsg( hwndDlgClient, LB_GAMES,
                       LM_SELECTITEM, 0, MPARAM(TRUE) );
  end;

  if GameListProtector_Sem<>-1 then
    DosReleaseMutexSem(GameListProtector_Sem);
end;

function GetCompatibleGame(Name:string; NetCommMaj, NetCommMin:smallword):pGameList;
var pTemp:pGameList;
begin
  if GameListProtector_Sem<>-1 then
    DosRequestMutexSem(GameListProtector_Sem, sem_indefinite_wait);

  pTemp:=GameListHead;
  while (pTemp<>nil) and ((StrPas(pTemp^.GameInfo.csGameName)<>Name) or
                          (pTemp^.GameInfo.sGameNetCommVerMaj<>NetCommMaj) or
                          (pTemp^.GameInfo.sGameNetCommVerMin<>NetCommMin)) do pTemp:=pTemp^.Next;

  result:=pTemp;

  if GameListProtector_Sem<>-1 then
    DosReleaseMutexSem(GameListProtector_Sem);
end;

procedure EraseGameList;
var ToDel:pGameList;
begin
  if GameListProtector_Sem<>-1 then
    DosRequestMutexSem(GameListProtector_Sem, sem_indefinite_wait);

  while GameListHead<>nil do
  begin
    ToDel:=GameListHead;
    GameListHead:=GameListHead^.Next;
    dispose(ToDel);
  end;
  WinSendDlgItemMsg( hwndDlgClient, LB_GAMES, LM_DELETEALL, 0, 0);

  if GameListProtector_Sem<>-1 then
    DosReleaseMutexSem(GameListProtector_Sem);
end;

