unit MemUInfo;
interface
uses os2base;

implementation

type pMBInfo=^MBInfo;
     MBInfo=record
       Size:longint;
       Address:longint;
       TID:longint;
       CallerAddr:longint;
       fVMT:Longint;
       fType:byte;
       Next:pMBInfo;
     end;

var OldMM:TMemoryManager;
    MBHead:pMBInfo=Nil;
    HCMux:hMtx;
    SumAlloc:longint;
    MaxAlloc:longint;

    Log:text;

function LongToStr(l:longint;length:longint):string;
begin
  str(l:length,result);
end;

function hex(l:longint):string;
const chars:array[0..15] of char=('0','1','2','3','4','5','6','7','8','9','A','B','C','D','E','F');
var e:string;
    b,a:byte;
begin
  e:='';
  for b:=0 to 7 do
  begin
    a:=l and $F;
    e:=chars[a]+e;
    l:=l shr 4;
  end;
  hex:=e;
end;


procedure Enter;
begin
  DosRequestMutexSem(HCMux,-1);
end;

Procedure Leave;
begin
  DosReleaseMutexSem(HCMux);
end;

procedure AddLogLine(s:String);
begin
  {$I-}
  append(Log);
  {$I+}
  if ioresult<>0 then
  begin
    {$I-}
    rewrite(Log);
    {$I+}
    if ioresult<>0 then exit;
  end;
  Writeln(Log,s);
  close(Log);
end;

Function GetListElement(Address:Longint; var Prev:pMBInfo):pMBInfo;
var p:pMBInfo;
begin
  p:=MBHead;Prev:=Nil;
  while (p<>Nil) and (p^.Address<>Address) do
  begin
    Prev:=p; p:=p^.Next;
  end;
  Result:=p;
end;

Procedure AddToList(Size,Address,TID,CallerAddr:longint;Kind:longint);
var p,prev:pMBInfo;
    fType:byte;
begin
  if GetListElement(Address,prev)<>Nil then
  begin
    AddLogLine('ERROR: Multiple Allocation for the same address!');
  end;
  if Kind=0 then
    fType:=0
  else if Kind=2 then
  begin
    fType:=vtAnsiString;
    Kind:=0;
  end else if Kind and 1 = 1 then
  begin
    fType:=vtObject;
    Kind:=Kind and $fffffffe;
  end else
    fType:=vtClass;

  p:=pMBInfo(OldMM.GetMem(sizeof(MBInfo)));
  p^.Next:=MBHead;
  p^.Size:=Size;
  SumAlloc:=SumAlloc+Size; if MaxAlloc<SumAlloc then MaxAlloc:=SumAlloc;
  p^.Address:=Address;
  p^.TID:=TID;
  p^.CallerAddr:=CallerAddr;
  p^.fVMT:=Kind;
  p^.fType:=fType;
  MBHead:=p;
end;

Procedure RemoveFromList(Address:Longint);
var p,prev:pMBInfo;
begin
  p:=GetListElement(Address,prev);
  if p=Nil then
  begin
    AddLogLine('WARNING: Freeing a non allocated memory address!');
    exit;
  end;
  if Prev=Nil then
  begin
    MBHead:=p^.Next;
  end else
  begin
    Prev^.Next:=p^.Next;
  end;
  SumAlloc:=SumAlloc-p^.Size;
  OldMM.FreeMem(p);
end;

function IntChkGetMem( Size: Longint; Caller:longint; Kind: longint ): Pointer;
begin
  Enter;
  Result:=OldMM.GetMem(Size);
  AddToList(Size,Longint(Result),GetThreadID,longint(caller), Kind);
  Leave;
end;

// CheckGetMem is at the core of the heap checker.  When it receives control
// from the System unit's _MemNew function, it saves the return address
// and attempts to determine the real origin of the memory allocation
// request, which is passed on to IntCheckMem that does the actual
// allocation.

function ChkGetMem(Size: Longint): Pointer; assembler; {&uses none} {&frame-}
asm
        pop     eax              // Return address

        mov     edx,[esp+$0c]    // Get real caller's address
        cmp     edx,OFFSET System.TObject.NewInstance
        jl      @@NotClass
        cmp     edx,OFFSET System.TObject.FreeInstance
        jge     @@NotClass
        mov     edx,[esp+$44]    // Constructing a class instance
        jmp     @@Go             // So class reference pointer in edx

      @@NotClass:
        cmp     edx,OFFSET System._ObjCtr
        jl      @@NotObject
        cmp     edx,OFFSET System._ObjDtr
        jge     @@NotObject
        mov     edx,[esp+$2c]    // Constructing an object instance
        or      ecx,1            // Mark as Object
        jmp     @@Go

      @@NotObject:
        cmp     edx,OFFSET System._LStrNew
        jl      @@Untyped
        cmp     edx,OFFSET System._LStrPacked
        jge     @@Untyped
//        mov edx,[esp+56]
        mov     ecx,2            // Mark as AnsiString
        jmp     @@Go

      @@Untyped:
        xor     ecx,ecx          // Not constructing a class or an object

      @@Go:
                                 // Param::0 - size - already on stack
        push    edx              // Param::1 - caller
        push    ecx              // Param::2 - TObject, if object
        push    eax              // Return address

        jmp     IntChkGetMem
end;


function ChkFreeMem(p:pointer):longint;
begin
  Enter;
  RemoveFromList(longint(p));
  Result:=OldMM.FreeMem(p);
  Leave;
end;

function IntChkReallocMem(p:pointer;size:longint; caller:longint; Kind:longint):pointer;
var m,prev:pMBInfo;
    OldSize:longint;
begin
  Enter;
  m:=GetListElement(longint(p),prev);
  if m=Nil then
  begin
    AddLogLine('ERROR: Reallocating a non allocated memory address!');
    result:=OldMM.ReAllocMem(p,size);
    Leave;
    exit;
  end;
  OldSize:=m^.Size;
  RemoveFromList(longint(p));
  Result:=OldMM.ReallocMem(p,Size);
  AddToList(Size,Longint(Result),GetThreadID,caller,Kind);
  Leave;
end;

function ChkReallocMem(p:Pointer;Size: Longint): Pointer; assembler; {&uses none} {&frame-}
asm
        pop     eax              // Return address

        mov     edx,[esp+$10]    // Get real caller's address
        cmp     edx,OFFSET System.TObject.NewInstance
        jl      @@NotClass
        cmp     edx,OFFSET System.TObject.FreeInstance
        jge     @@NotClass
        mov     edx,[esp+$48]    // Constructing a class instance
        jmp     @@Go             // So class reference pointer in edx

      @@NotClass:
        cmp     edx,OFFSET System._ObjCtr
        jl      @@NotObject
        cmp     edx,OFFSET System._ObjDtr
        jge     @@NotObject
        mov     edx,[esp+$30]    // Constructing an object instance
        or      ecx,1            // Mark as Object
        jmp     @@Go

      @@NotObject:
        cmp     edx,OFFSET System._LStrNew
        jl      @@Untyped
        cmp     edx,OFFSET System._LStrPacked
        jge     @@Untyped
//        mov edx,[esp+44]
        mov     ecx,2            // Mark as AnsiString
        jmp     @@Go

      @@Untyped:
        xor     ecx,ecx          // Not constructing a class or an object

      @@Go:
                                 // Param::0 - size - already on stack
        push    edx              // Param::1 - caller
        push    ecx              // Param::2 - TObject, if object
        push    eax              // Return address


        jmp     IntChkReallocMem
end;



var
  HeapChecker: TMemoryManager = (
    GetMem: ChkGetMem;
    FreeMem: ChkFreeMem;
    ReallocMem: ChkReallocMem);

function InfoLine( pItem:pMBInfo ): String;
const
  DumpLength = 24;
var
  sContent: String;
  sLoc: String;
  pSource: Pointer;
  FileName: ShortString;
  LineNo: Longint;
  pC: pChar;
  i: Integer;

begin
  // Construct contents of unfreed pointer
  case pItem^.fType of
    vtObject     : sContent := '(Object vmt: '+hex(pItem^.fVMT)+')';
    vtClass      : sContent := '(Class: '+TClass(pItem^.fVmt).ClassName+')';
    vtAnsiString : sContent := '(str: '+AnsiString(pItem^.Address+8)+')';
  else
    begin
      pC := pChar(pItem^.Address);
      sContent := '';
      i := 0;
      while (i < DumpLength) and (i<pItem^.Size) do
        begin
          if pC^ >= ' ' then
            sContent := sContent + pC^
          else
            sContent := sContent + '.';
          inc(pC);
          inc(i);
        end;
    end;
  end;
  while length(sContent)<DumpLength do sContent:=sContent+' ';

  // Determine where Problem originated
  if pItem^.CallerAddr = 0 then
    sLoc := ' [Unknown]'
  else
    begin
      pSource := GetLocationInfo( pointer(pItem^.CallerAddr), FileName, LineNo );
      if pSource = nil then          // Source not found
        sLoc := ' [Addr = '+Hex(pItem^.CallerAddr)+']'
      else                           // Source found
        sLoc := ' ['+FileName+' #'+LongToStr(LineNo,1)+']';
    end;

  Result := sContent + sLoc;
end;

procedure HC_ExitProc;
var p:pMBInfo;
    SumLeaked:longint;
begin
  SetMemoryManager(OldMM);
  DosCloseMutexSem(HCMux);
  AddLogLine('---- [ General Information ] ----');
  AddLogLine('');
  AddLogLine(' Executable: '+paramstr(0));
  AddLogLine(' Maximum used memory: '+LongToStr(MaxAlloc,8)+' byte(s)');
  if MBHead<>Nil then
  begin
    SumLeaked:=0;
    AddLogLine('');
    AddLogLine('---- [ Memory Leakage Information: ] ----');
    AddLogLine('');
    AddLogLine('Leaked memory blocks (first is the youngest...) :');
    AddLogLine('TID  Address    Bytes              Memory Dump     Allocated by    ');
    AddLogLine('--- -------- -------- ------------------------ --------------------');
    while MBHead<>Nil do
    begin
      p:=MBHead;
      MBHead:=MBHead^.Next;
      AddLogLine(LongToStr(p^.TID,3)+' '+Hex(p^.Address)+' '+
                 LongToStr(p^.Size,8)+' '+InfoLine(p));
      OldMM.FreeMem(Pointer(p^.Address));
      SumLeaked:=SumLeaked+p^.Size;
      OldMM.FreeMem(p);
    end;
    AddLogLine('');
    AddLogLine('Summary of leaked memory: '+LongToStr(SumLeaked,8)+' byte(s)');
  end else
  begin
    AddLogLine('');
    AddLogLine(' No memory leakage detected.');
  end;
  AddLogLine('');
  AddLogLine('---- [ END OF FILE ] ----');
end;


var LoggingNeeded:boolean;

function UpStr(s:String):String;
var b:byte;
begin
  for b:=1 to length(s) do s[b]:=upcase(s[b]);
  result:=s;
end;

procedure Process_Parameters;
var b:byte;
begin
  LoggingNeeded:=false;
  for b:=1 to paramcount do
  begin
    if upstr(paramstr(b))='-MEMUINFO' then
      LoggingNeeded:=true;
  end;
end;

begin

  Process_Parameters;
  if LoggingNeeded then
  begin
    MBHead:=Nil;
    SumAlloc:=0;
    MaxAlloc:=0;
    if DosCreateMutexSem(nil, HCMux, 0, false)=no_error then
    begin
      GetMemoryManager(OldMM);
      SetMemoryManager(HeapChecker);

      Assign(Log,'MemUInfo.log');
      {$I-}
      Erase(log);
      {$I+}
      if ioresult=0 then   // clear the flag.
      begin  end;

      AddExitProc(HC_ExitProc);
    end;
  end;
end.
