program NetGameServer;
{$PMTYPE VIO}
{$H-} // DONT use huge string, they are buggy in Virtual Pascal v2.1!
uses os2base, os2def,
     crt,
     vputils,
     inet,
     CommonNetFuncs;

const AppName=ServerNameVer;
      Server_Port:Longint=630;             // Default port for server
      Debug:Boolean=False;

const MAX_CLIENTS=512;

      MSG_WAITTIME =20000; // Send Ping after 20 secs of inactivity!
      PING_TIMEOUT =1;     // Terminate connection if no PONG response in 1*PING_WAITTIME ms!

type pCommandList=^tCommandList;  // List of commands for a client
     tCommandList=record
       CommandStr:String;
       Next:pCommandList;
     end;

type pClientDataList=^tClientDataList; // Description of an active client
     tClientDataList=record
       ThreadID:TID;                 {My ThreadID}
       BufHead,BufTail:pCommandList; {Commands' queue}
       BufProtector_Sem:HMtx;        // Mutex to protect the command queue
       BufEvent:hEv;                 // Event semaphore to signal new command arrival
       ReaderTID:Tid;                {Reader's TID}
       MySocket:longint;
       Remote:SockAddr_In;

       ClientName:String; // Name of player
       ClientType:byte;   // See CLIENT_* contants!
       ClientDirectIP:longint;    // Server address of client, where other clients can
       ClientDirectPort:longint;  // contact it

       Next:pClientDataList;
     end;

var   MainSock:longint;
      ConnectorTID:longint;
      NumOfClients:longint;
      ClientDatasProtector_Sem:hMtx; // Mutex Semaphore
      ClientDatasHead:pClientDataList;

procedure AddNewCmd(Client:pClientDataList; Cmd:String);
var NewCmd:pCommandList;
begin
  if Client=nil then exit;
  with Client^ do
  begin
    DosRequestMutexSem(BufProtector_Sem, sem_indefinite_wait);
    New(NewCmd);
    NewCmd^.Next:=Nil;
    NewCmd^.CommandStr:=Cmd;
    If BufTail<>Nil then
    Begin
      BufTail^.next:=NewCmd;
      BufTail:=NewCmd;
    end else
    begin
      BufHead:=NewCmd;
      BufTail:=NewCmd;
    end;

    DosReleaseMutexSem(BufProtector_Sem);

    DosPostEventSem(BufEvent);
  end;
end;

Function NextCmd(Client:pClientDataList):String;
var Temp:pCommandList;
begin
  result:='';
  if Client=Nil then exit;
  with Client^ do
  begin
    DosRequestMutexSem(BufProtector_Sem, sem_indefinite_wait);
    if BufHead<>nil then
    begin
      temp:=BufHead;
      BufHead:=BufHead^.next;
      if BufHead=Nil then BufTail:=Nil;
      Result:=temp^.CommandStr;
      dispose(temp);
    end;
    DosReleaseMutexSem(BufProtector_Sem);
  end;
end;

Function ClientReaderThread(p:pointer):longint;
var Cmd:pCommandList;
    ch:char;
    Str:string;
    readed:longint;
    sock:longint;
begin
  sock:=pClientDataList(p)^.MySocket;
  str:='';
  repeat
    str:=ReceiveString(sock);
    if str='' then DosSleep(32) else
      AddNewCmd(pClientDataList(p), str);
  until false;
end;

procedure SendClientInfo(Socket:ulong; pClientData:pClientDataList);
var s,msg:string;
begin
  if pClientData=Nil then exit;

  msg:='CLIENTINFO ';
  with pClientData do
  begin
    str(longint(pClientData),s);
    msg:=msg+s+' ';
    str(ClientType,s);
    msg:=msg+s+' "'+ClientName+'"';
    str(longint(ClientDirectIP),s);
    msg:=msg+' '+s;
    str(longint(ClientDirectPort),s);
    msg:=msg+' '+s;
  end;
  SendString(Socket, msg);
end;

procedure SendAllClientInfo(Socket:ulong);
var Temp:pClientDataList;
begin
  if ClientDatasProtector_Sem<>-1 then
    DosRequestMutexSem(ClientDatasProtector_Sem, sem_indefinite_wait);

  Temp:=ClientDatasHead;
  while (Temp<>Nil) do
  begin
    SendClientInfo(Socket, Temp);
    Temp:=Temp^.Next;
  end;

  if ClientDatasProtector_Sem<>-1 then
    DosReleaseMutexSem(ClientDatasProtector_Sem);
end;

procedure SendClientInfoA(pClientData:pClientDataList);
var Temp:pClientDataList;
begin
  if pClientData=Nil then exit;

  if ClientDatasProtector_Sem<>-1 then
    DosRequestMutexSem(ClientDatasProtector_Sem, sem_indefinite_wait);

  Temp:=ClientDatasHead;
  while (Temp<>Nil) do
  begin
    SendClientInfo(Temp^.MySocket, pClientData);
    Temp:=Temp^.Next;
  end;

  if ClientDatasProtector_Sem<>-1 then
    DosReleaseMutexSem(ClientDatasProtector_Sem);
end;

procedure SendStringA(s:String);
var Temp:pClientDataList;
begin

  if ClientDatasProtector_Sem<>-1 then
    DosRequestMutexSem(ClientDatasProtector_Sem, sem_indefinite_wait);

  Temp:=ClientDatasHead;
  while (Temp<>Nil) do
  begin
    SendString(Temp^.MySocket, s);
    Temp:=Temp^.Next;
  end;

  if ClientDatasProtector_Sem<>-1 then
    DosReleaseMutexSem(ClientDatasProtector_Sem);
end;

procedure SendStringAET(Socket:longint; s:String);
var Temp:pClientDataList;
begin

  if ClientDatasProtector_Sem<>-1 then
    DosRequestMutexSem(ClientDatasProtector_Sem, sem_indefinite_wait);

  Temp:=ClientDatasHead;
  while (Temp<>Nil) do
  begin
    if Socket<>Temp^.MySocket then
      SendString(Temp^.MySocket, s);
    Temp:=Temp^.Next;
  end;

  if ClientDatasProtector_Sem<>-1 then
    DosReleaseMutexSem(ClientDatasProtector_Sem);
end;

function GenerateClientName:String;
const Base='Unnamed';
var Temp:pClientDataList;
    idx:longint;
    s:string;
begin
  if ClientDatasProtector_Sem<>-1 then
    DosRequestMutexSem(ClientDatasProtector_Sem, sem_indefinite_wait);

  idx:=1;
  repeat
    str(idx,s);
    s:=Base+s;
    Temp:=ClientDatasHead;
    while (Temp<>Nil) and (Temp^.ClientName<>s) do Temp:=Temp^.Next;
    inc(idx);
  until (Temp=Nil);

  if ClientDatasProtector_Sem<>-1 then
    DosReleaseMutexSem(ClientDatasProtector_Sem);

  Result:=s;
end;


function AddNewClient(data:tClientDataList):pClientDataList;
var Prev,Temp:pClientDataList;
begin
  if ClientDatasProtector_Sem<>-1 then
    DosRequestMutexSem(ClientDatasProtector_Sem, sem_indefinite_wait);

  Prev:=Nil;
  Temp:=ClientDatasHead;
  while Temp<>Nil do
  begin
    Prev:=Temp;
    Temp:=Temp^.Next;
  end;
  Result:=New(pClientDataList);
  Result^:=data;
  Result^ .Next:=Nil;
  if Prev=Nil then ClientDatasHead:=Result else
    Prev^.Next:=Result;

  inc(NumOfClients);

  if ClientDatasProtector_Sem<>-1 then
    DosReleaseMutexSem(ClientDatasProtector_Sem);
end;

procedure DeleteOldClient(data:pClientDataList);
var Prev,Temp:pClientDataList;
    Msg:String;
begin
  if ClientDatasProtector_Sem<>-1 then
    DosRequestMutexSem(ClientDatasProtector_Sem, sem_indefinite_wait);

  Prev:=Nil;
  Temp:=ClientDatasHead;
  while (Temp<>Nil) and (Temp<>data) do
  begin
    Prev:=Temp;
    Temp:=Temp^.Next;
  end;
  if temp<>nil then
  begin // Found, delete it!
    if Prev=Nil then ClientDatasHead:=Temp^.Next else
    Prev^.Next:=Temp^.Next;

    str(longint(Temp), Msg);
    Msg:='REMOVEUSER '+Msg;
    SendStringAET(Temp^.MySocket, Msg);

    Dispose(Temp);

    dec(NumOfClients);
  end;

  if ClientDatasProtector_Sem<>-1 then
    DosReleaseMutexSem(ClientDatasProtector_Sem);
end;

function GetClientByName(name:string):pClientDataList;
var Temp:pClientDataList;
begin
  if ClientDatasProtector_Sem<>-1 then
    DosRequestMutexSem(ClientDatasProtector_Sem, sem_indefinite_wait);

  Temp:=ClientDatasHead;
  while (Temp<>Nil) and (Temp.ClientName<>name) do Temp:=Temp^.Next;

  if ClientDatasProtector_Sem<>-1 then
    DosReleaseMutexSem(ClientDatasProtector_Sem);

  result:=Temp;
end;

function GetClientByID(id:longint):pClientDataList;
var Temp:pClientDataList;
begin
  if ClientDatasProtector_Sem<>-1 then
    DosRequestMutexSem(ClientDatasProtector_Sem, sem_indefinite_wait);

  Temp:=ClientDatasHead;
  while (Temp<>Nil) and (longint(Temp)<>ID) do Temp:=Temp^.Next;

  if ClientDatasProtector_Sem<>-1 then
    DosReleaseMutexSem(ClientDatasProtector_Sem);

  result:=Temp;
end;

function AddressPort2String(lSrvAddress:longint;lSrvPort:longint):string;
var adr:pchar;
begin
  adr:=inet_ntoa(lSrvAddress);
  lSrvPort:=ntohs(lSrvPort);
  result:=StrPas(adr)+':'+LongToStr(lSrvPort);
end;

function ClientServerThread( p:pointer) :longint;
Var sTemp:String;
    lTemp,lTemp2:Longint;
    Code:longint;
    pClientData:pClientDataList;
    InactiveTime:longint;
    PingTime:longint;
    s:string;
    ErrorKill:Boolean;
    WaitRc:ApiRet;
    PostCount:longint;
    SockError:longint;
begin
  pClientData:=P;
  ErrorKill:=False;
  PingTime:=0;
  with pClientData^ do
  begin
    DosCreateMutexSem(nil,BufProtector_Sem,0,false);
    DosCreateEventSem(nil,BufEvent,0,false);
    ReaderTID:=VPBeginThread(ClientReaderThread,65535,p);
    DosSetPriority(prtys_Thread,prtyc_TimeCritical,+15,ReaderTID);
    InactiveTime:=0;
    s:='';
    Str(longint(pClientData),sTemp);
    SendString(MySocket,'YOURID '+sTemp);
    SendStringAET(MySocket,'ADDUSER '+sTemp);
    SendClientInfoA(pClientData);
    SendString(MySocket,'READY');

    repeat
      if BufHead=Nil then
      begin
        DosResetEventSem(BufEvent, PostCount);
        WaitRc:=DosWaitEventSem(BufEvent, MSG_WAITTIME);
      end else WaitRc:=0;
      if (WaitRc=0) and (BufHead<>Nil) then
      begin
        s:=NextCmd(pClientData);
        if s='PING' then
        begin
          If Debug then
            writeln('  - PING from ',ClientName);
          SendString(MySocket,'PONG');
        end else
        if copy(s,1,9)='SET NAME ' then
        begin
          sTemp:=GetSafeString(copy(s,10,length(s)-9));
          if (sTemp=ClientName) or (GetClientByName(sTemp)<>Nil) then
          begin // A client with this new name already exists
                // so no update, and send back the old values as result!
            If Debug then
              writeln('  - NOT setting name of ',ClientName,' to ',sTemp);
            SendClientInfo(MySocket,pClientData);
          end else
          begin
            If Debug then
              writeln('  - Setting name of ',ClientName,' to ',sTemp);
            ClientName:=sTemp;
            SendClientInfoA(pClientData); // Send the updated client information to all users
          end;
        end else
        if copy(s,1,9)='SET TYPE ' then
        begin
          val(copy(s,10, length(s)-9), lTemp, Code);
          If Debug then
            writeln('  - Setting type of ',ClientName,' to ',lTemp);
          ClientType:=lTemp;
          SendClientInfoA(pClientData); // Send the updated client information to all users
        end else
        if copy(s,1,15)='SET DIRECTADDR ' then // SET DIRECTADDR <IP> <PORT>
        begin
          s:=copy(s,16, length(s)-15); // Get only the parameters
          sTemp:=copy(s,1,pos(' ',s)-1);
          val(sTemp, lTemp, Code);
          if (Code<>0) or (lTemp=0) or (lTemp=-1) then
          begin
            if Debug then writeln('  - Invalid IP address from ',ClientName,' for direct address: ',sTemp);
          end else
          begin
            sTemp:=copy(s,pos(' ',s)+1,length(s));
            val(sTemp, lTemp2, Code);
            if (Code<>0) or (lTemp2=0) or (lTemp2=-1) then
            begin
              if Debug then writeln('  - Invalid Port from ',ClientName,' for direct address: ',sTemp);
            end else
            begin
              If Debug then
                writeln('  - Setting direct address of ',ClientName,' to [',AddressPort2String(lTemp,lTemp2),']');
              ClientDirectIP:=lTemp;
              ClientDirectPort:=lTemp2;
              SendClientInfoA(pClientData); // Send the updated client information to all users
            end;
          end;
        end else
        if copy(s,1,15)='GET CLIENTINFO ' then
        begin
          sTemp:=copy(s,16, length(s)-15);
          if sTemp='ALL' then
          begin
            If Debug then
              writeln('  - Get all clientinfo from ',ClientName);
            SendAllClientInfo(MySocket);
          end else
          begin
            val(copy(s,16, length(s)-15), lTemp, Code);
            If Debug then
              writeln('  - Get clientinfo of ',lTemp,' from ',ClientName);
            SendClientInfo(MySocket, GetClientByID(lTemp));
          end;
        end else
        if s='DISCONNECT' then
        begin
          If Debug then
            writeln('  - Disconnect request from ',ClientName);
        end else
        If s='PONG' then
        begin
          If Debug then
            writeln('  - Reply to ping has arrived from ',ClientName);
          PingTime:=0;
        end else
        begin
          If Debug then
            writeln('*** Unknown command from ',clientname,' : ',s);
          SendString(MySocket,'ERROR 002 '+MessageArray[002]+' : '+s);
        end;
        InactiveTime:=0;
      end else
      Begin
        inc(InactiveTime);
      end;

      If (WaitRc=Error_TimeOut) then // We have waiten too much, see if the connection is still alive
      begin
        if PingTime=0 then
        begin
          If Debug then
            writeln('  - Pinging ',ClientName);
            SendString(MySocket,'PING');
            PingTime:=PING_TIMEOUT; InactiveTime:=0;
        end else
        begin
          dec(PingTime);
          if PingTime=0 then
          begin
            If Debug then
              writeln('  - ',ClientName,' is not responding for ',PING_TIMEOUT*MSG_WAITTIME,' ms, forcing Disconnect...');
            SendErrorString(MySocket, 004); // Timeout
            ErrorKill:=True;
          end;
        end;
      end;
      DosSleep(0);
    until (s='DISCONNECT') or (ErrorKill);
    SendString(MySocket,'DISCONNECT');
    KillThread(ReaderTID);
    DosWaitThread(ReaderTID, dcww_Wait);
    soClose(MySocket);
    If Debug then
      writeln('Client Disconnected from ',AddressPort2String(Remote.sin_addr,Remote.sin_port),
              ' (',ClientName,')');
  end;
  DeleteOldClient(pClientData);
end;

function ConnectorThread(p:pointer):longint;
var
  aSock    : longint;
  Server,aClient  :SockAddr_in;
  SockAddr_Size:longint;
  i,j      : Longint;
  Serv     : ServEnt;
  NewClData: tClientDataList;
  pNewClData:pClientDataList;
begin
  Writeln('Starting Service...');
  aSock:=0;
  while (listen(MainSock,5)=0) and (aSock>-1) do
  begin   // Listen for anything interesting
    SockAddr_Size:=sizeof(SockAddr);
    aSock:=accept(MainSock,@(SockAddr(aClient)), @SockAddr_Size);
    if (aSock>-1) then
    begin
      If Debug then
        Writeln('New client from ',AddressPort2String(aClient.sin_addr,aClient.sin_port));

      SendString(aSock,AppName);

      if (NumOfClients>MAX_CLIENTS) then
      begin
        If Debug then
          writeln('  * Disconnecting, too many clients! (',NumOfClients,')');
        SendErrorString(aSock, 001); // Too many clients!
        SendString(aSock,'DISCONNECT');
        soClose(aSock);
      end else
      begin
        If Debug then
          Writeln('  * Registering new client');
        with NewClData do
        begin
          MySocket:=aSock;
          Remote:=aClient;
          BufHead:=Nil;
          BufTail:=Nil;
          ClientName:=GenerateClientName;
          ClientType:=CLIENT_UNKNOWN;
          ClientDirectIP:=-1;
          ClientDirectPort:=-1;
        end;

        pNewClData:=AddNewClient(NewClData);

        pNewClData^.ThreadID:=vpbeginthread(ClientServerThread,16384,pNewClData);
        DosSleep(50);
        DosSetPriority(prtys_Thread,prtyc_TimeCritical,+15,
           pNewClData^.ThreadID);
      end;
    end;
  end;
  writeln('Stopping service...');
end;

Function Open_Socket:longint;
var
  Server : SockAddr_in;
  i,j    : Integer;
  rc:apiret;
begin
  result:=0; // Default: no problem

  rc:=Sock_Init;
  if rc<>0 then
  begin
    if Debug then
      Writeln('Open_Socket: Unable to initilize sockets, Sock_Init, rc=',rc);
    Result:=1; exit;
  end;

  MainSock:=Socket(PF_INET,SOCK_STREAM,0);  // Open a socket
  if MainSock<0 then
  begin
    if Debug then
      writeln('Open_Socket: Error creating new socket, Socket, rc=',MainSock);
    Result:=2; exit;
  end;

  Server.sin_family  := AF_INET;
  Server.sin_port  := htons(Server_Port);
  Server.sin_addr := INADDR_ANY;

  rc:=Bind(MainSock,SockAddr(Server), sizeof(SockAddr));      // Bind the socket to the port
  if rc<0 then
  begin
    if Debug then
      writeln('Open_Socket: Error binding socket, Bind, rc=',rc);
    soclose(MainSock);
    Result:=3; exit;
  end;
  if Debug then
    Writeln('Open_Socket: Socket opened on port ',server_port)
  else
    Writeln('Using port ',server_port);
end;

Procedure CleanUp_Clients;
var Temp:pClientDataList;
begin

  if ClientDatasProtector_Sem<>-1 then
    DosRequestMutexSem(ClientDatasProtector_Sem, sem_indefinite_wait);

  Temp:=ClientDatasHead;
  while (Temp<>Nil) do
  begin
    Writeln('Cleanup_Clients: Cleaning up ',Temp^.ClientName);
    SendErrorString(Temp^.MySocket, 003); // Shutting down
    SendString(Temp^.MySocket,'DISCONNECT');
    DosKillThread(Temp^.ReaderTID);
    DosKillThread(Temp^.ThreadID);
    DosSleep(32);
    soclose(Temp^.MySocket);
    Temp:=Temp^.Next;
  end;

  if ClientDatasProtector_Sem<>-1 then
    DosReleaseMutexSem(ClientDatasProtector_Sem);

  while ClientDatasHead<>Nil do DeleteOldCLient(ClientDatasHead);

end;

Procedure Close_Socket;
begin
  soclose(MainSock);
  MainSock:=0;
  if Debug then
    Writeln('Close_Socket: Service closed');
end;

procedure DuelServer_ExitProc;
begin
  Cleanup_Clients; // Make sure all clients are disconnected
  if MainSock<>0 then
  begin
    if Debug then
      Writeln('ExitProc: Closing Main Socket');
    Close_Socket;
  end;
end;

procedure Initialize_Variables;
begin
  MainSock:=0;
  ConnectorTID:=-1;
  NumOfClients:=0;
  ClientDatasHead:=nil;
  ClientDatasProtector_Sem:=-1;
end;

function MyUpperCase(s:string):string;
var l:longint;
begin
  for l:=1 to length(s) do s[l]:=upcase(s[l]);
  result:=s;
end;

procedure Console;
var s:string;
begin
  Writeln('Service is running.');
  Writeln('Enter HELP to get a list of usable commands!');
  Writeln;
  repeat
    write('Console>');
    readln(s);
    s:=MyUpperCase(s);
    if s='HELP' then
    begin
      Writeln('The following commands can be used in the console:');
      writeln('  HELP               : This help');
      writeln('  QUIT / BYE / EXIT  : Stop the service and quit');
      writeln('  DEBUG <ON>|<OFF>   : Turn debug messages on/off');
    end else
    if copy(s,1,6)='DEBUG ' then
    begin
      if copy(s,7,2)='ON' then
      begin
        Debug:=True;
        Writeln('Debug messages ON');
      end else
      begin
        Debug:=false;
        Writeln('Debug messages OFF');
      end;
    end else
    if (s='QUIT') or (s='BYE') or (s='EXIT') then
    begin
      Writeln('Exiting...');
    end else
    writeln('Unknown console command! Enter HELP to get a list of usable commands!');
  until (s='QUIT') or (s='BYE') or (s='EXIT');
end;

procedure Process_Parameters;
var i:longint;
    s:string;
    code:longint;
begin
  for i:=1 to paramcount do
  begin
    s:=MyUpperCase(paramstr(i));
    if s='/DEBUG' then Debug:=true;
    if s[1]+s[2]='/P' then
    begin
      val(copy(s,3,length(s)),Server_Port,code);
    end;
  end;
end;


begin
  clrscr;
  writeln(AppName);
  Initialize_Variables;
  Process_Parameters;

  AddExitProc(DuelServer_ExitProc);

  if sock_init<>0 then
  begin
    Writeln('Could not initialize sockets!');
    Writeln('Make sure you have TCP/IP networking installed!');
    halt(-1);
  end;

  if Open_Socket<>0 then
  begin
    Writeln('Initialization failed! Exiting.');
    Writeln('Are you sure that nothing is running on port ',Server_Port,' ?');
    halt(1);
  end;

  DosCreateMutexSem(nil,ClientDatasProtector_Sem,0,false);

  ConnectorTID:=VPBeginThread(ConnectorThread,16384,nil);
  if ConnectorTID<=0 then
  begin
    Writeln('Could not create Connector thread! Exiting.');
    Close_Socket;
    halt(2);
  end;

  DosSleep(128); // Wait some

  Console; // Start console, process local user commands

  Cleanup_Clients;
  Close_Socket; // This will also timeout the Connector Thread's accept()!
  DosSleep(128); // Wait some

  DosKillThread(ConnectorTID);
  DosWaitThread(ConnectorTID, dcww_wait);

  DosCloseMutexSem(ClientDatasProtector_Sem); ClientDatasProtector_Sem:=-1;

  Writeln('Bye!');
end.

